package NORID::EppClient;

use strict;
use warnings;
use Catalyst::Runtime '5.90';
use YAML;
use Log::Log4perl;

our $VERSION='1.10';

=pod

=head1 NAME

EppClient - Catalyst based EPP Client application for .NO 

package NORID::EppClient - Catalyst EppClient application

Norid EPP Client


=head1 SYNOPSIS

    script/eppclient_server.pl

See PODs otherwise.

=head1 DEPENDENCIES

Needs the following modules:

Net::DRI for .NO

=head1 DESCRIPTION

The client supports creation and maintainance of host, contact and domain 
objects for .NO. It supports various transfer operations, as well as poll 
operation for the message queue.

It is intended as a simple and manual web client for low volume registrars.

=head1 SEE ALSO

L<NORID::EppClient::Controller::Root>, L<Catalyst>

=head1 AUTHOR

Trond Haugen

=head1 LICENSE

This library is free software, you can redistribute it and/or modify
it under the same terms as Perl itself.

=cut


# Set flags and add plugins for the application
#
#         -Debug: activates the debug mode for very useful log messages
#   ConfigLoader: will load the configuration from a YAML file in the
#                 application's home directory
# Static::Simple: will serve static files from the application's root 
#                 directory
# Session and cookies: For cookie bases authentication and session 
#                 storage
# Add RequireSSL when real deployment to an SSL enabled apache is to be done
#
require 5.14.0;

my ($storage_type, $config, $cat_version);

BEGIN {
  $config = YAML::LoadFile($ENV{'CF_EPPCLIENT_CONF'});
  $storage_type = $config->{session}->{storage_type} || 'File';
  $cat_version  = $Catalyst::Runtime::VERSION;
}
use if ($cat_version == 5.90007), 
    # Ubuntu 12.04

    Catalyst => 

	'ConfigLoader',

	'Static::Simple',
	  
	'Unicode::Encoding',
	  
	'StackTrace',
    
	'Authentication',
	'Authorization::Roles',
	'Authorization::ACL',
	
	'Session',
	"Session::Store::$storage_type",
	'Session::State::Cookie',
    ;

use if ($cat_version >=  5.90053),
    # Ubuntu 14.04: 5.90053
    # Ubuntu 15.04: 5.90097

    Catalyst =>

	'ConfigLoader',

	'Static::Simple',

	'StackTrace',

	'Authentication',
	'Authorization::Roles',
	'Authorization::ACL',

	'Session',
	"Session::Store::$storage_type",
	'Session::State::Cookie',
    ;


# Configure the application. 
#
# Note that settings in eppclient.yaml (or other external
# configuration file that you set up manually) take precedence
# over this when using ConfigLoader. Thus configuration
# details given here can function as a default configuration,
# with a external configuration file acting as an override for
# local deployment.

__PACKAGE__->config($config);

__PACKAGE__->config->{'Plugin::Static::Simple'}->{ignore_extensions} 
        = [ qw/html/ ];

__PACKAGE__->config->{'Plugin::Static::Simple'}->{ignore_dirs} 
        = [ qw/.svn/ ];

__PACKAGE__->config('Controller::Changepw' => {
        session_name => 'pw_token',
        request_name => 'pw_token',
    });

__PACKAGE__->config('Controller::Login' => {
        session_name => 'l_token',
        request_name => 'l_token',
    });

__PACKAGE__->config('Controller::Host' => {
        session_name => 'h_token',
        request_name => 'h_token',
    });

__PACKAGE__->config('Controller::Contact' => {
        session_name => 'c_token',
        request_name => 'c_token',
    });

__PACKAGE__->config('Controller::Domain' => {
        session_name => 'd_token',
        request_name => 'd_token',
    });

__PACKAGE__->config('Controller::DomainBulk' => {
        session_name => 'db_token',
        request_name => 'db_token',
    });

__PACKAGE__->config('Controller::Message' => {
        session_name => 'm_token',
        request_name => 'm_token',
    });

if ($cat_version >=  5.90053) {
    __PACKAGE__->config( encoding => 'UTF-8');
}

__PACKAGE__->setup;

#
# Verify that the session and log directories and files are accessible

# The session plugin shall create the session directory when needed
# Verify that the XML log directory exists and is accessible,
# create it if not.
#
my $xmld = __PACKAGE__->config->{xmllog}->{xmllogdir};
unless (-d $xmld) {
    mkdir $xmld or die ("Cannot make log directory $xmld");
}
unless (-d $xmld) {
    die ("No $xmld directory");
}
unless (-w $xmld) {
    die ("$xmld directory is not writable");
}

###
# ACL protect the changepw command if not activated in the config
__PACKAGE__->acl_add_rule(
                          "/changepw",
			  sub {
			      my ( $c, $action ) = @_;

			      if ( $c->config->{client}->{changepw_cmd}) {
				  $c->acl_access_allowed;
			      } else {
				  $c->acl_access_denied;
			      }           
			  }
			  );




###
# Init the logging subsystem

Log::Log4perl->init_once(__PACKAGE__->config->{log}->{logconf});

if (exists(__PACKAGE__->config->{catalystlog})) {
    # config as defined in config file
    if (__PACKAGE__->config->{catalystlog}->{enabled}) {
	if (__PACKAGE__->config->{catalystlog}->{syslog}) {
	    #print STDERR "Logging to syslog\n";
	    __PACKAGE__->log(Log::Log4perl->get_logger('EppClientSyslog'));
	} else {
	    #print STDERR "Logging to file\n";
	    __PACKAGE__->log(Log::Log4perl->get_logger('EppClientFile'));
	}
    } else {
	#print STDERR "Logging is disabled. Turn off all Catalyst default logging\n";
	__PACKAGE__->log->disable('debug', 'info', 'warn','error','fatal');

    }

} else {
    # Default setup as before
    #print STDERR "default log config setup, to file\n";
    __PACKAGE__->log(Log::Log4perl->get_logger('EppClientFile'));
}

1;
