package NORID::EppClient::Controller::Contact;

use strict;
use warnings;

use Net::DRI;

use NORID::EppClient::Model::NORID::DRI;
use NORID::EppClient::Controller::Util;
use URI::Escape qw /uri_escape/;
use base qw(Catalyst::Controller::RequestToken);
use Email::Valid;
use Storable qw(dclone);

# For Debug purposes:
use Data::Dumper;
$Data::Dumper::Indent=1;

my $TT_CONTACT = 'contact.tt2';
my $TT_LOGIN = 'login.tt2';
my $BRREG = Encode::decode('ISO8859-1', "Brnnysund");

=pod

=head1 NAME

EppClient - Contact object operations

=head1 DESCRIPTION

The module supports creation and maintainance of contact objects for .NO. 

=head1 SYNOPSIS

=head1 DESCRIPTION

package NORID::EppClient::Controller::Contact - Catalyst Contact Controller

=head1 METHODS

=cut

# Contact scalar values
my @cs = ('srid', 'name', 'type', 'org', 'cc', 'pc', 'city', 
	  'voice', 'mobilephone', 'email');

# Contact array values
my @ca = ('street', 'organization', 'xemail', 'rolecontact');


# Disclose
my @discloses = ('voicedisclose', 'emaildisclose');
my @xdiscloses = ('mobilePhonedisclose');

# To hash from object, ('sp', 'org' and 'auth' is not used in .NO)
my @ho = ('srid', 'name', 'org', 'type', 'identity',
	  'pc', 'cc', 'city', 'email', 
	  'voice', 'mobilephone',
	  'disclose', 'xdisclose');

# To hash from object or session
my @hos   = ('status', 'statuslist','roid', 'clID', 'crID', 'upID', 'crDate', 'upDate');

# To hash from object arrays
my @hoa_nsort = ('street');
my @hoa_sort = ('xemail', 'organization', 'rolecontact');


# @cs varaibles can all be trimmed
my %trimmable = map { $_ => 1 } @cs, @hoa_sort, @hoa_nsort;


=head2 menu 

menu method.

=cut

sub menu : Local {
    my ( $self, $c ) = @_;

    $c->stash->{pc} = $c->session->{pc} if ($c->session->{pc} && !defined($c->stash->{pc}));

    # Recreate token when menu is pressed, this will cleanup any errors
    $c->session->{'c_token'} = $self->create_token($c);

    $c->stash->{template} = $TT_CONTACT;
}

=head2 onemore_streetline

Add another street address line

=cut

sub onemore_streetline : Private {
    my ( $self, $c ) = @_;

    # Avoid that the user adds more than 3 address lines.
    # So, just count the number of lines and stop on three

    # We expect at least an empty street array
    if (!defined($c->stash->{pc}->{street}) || 
	defined($c->stash->{pc}->{street}) && (scalar(@{$c->stash->{pc}->{street}}) < 3)) {
	push @{$c->stash->{pc}->{street}}, "";
    } else {
	# Explain the limitation
	push @{$c->stash->{opsummary}}, "Sorry, EPP has a maximum of three address lines.";
    }
}

=head2 onemore_xemail

Add another xemail line

=cut

sub onemore_xemail : Private {
    my ( $self, $c ) = @_;
   
    push @{$c->stash->{pc}->{xemail}}, "";
}

=head2 onemore_rolecontact

Add another role contact line

=cut

sub onemore_rolecontact : Private {
    my ( $self, $c ) = @_;
   
    push @{$c->stash->{pc}->{rolecontact}}, "";
}

=head2 onemore_orghandle

Add another related org. handle line line

=cut

sub onemore_orghandle : Private {
    my ( $self, $c ) = @_;
   
    push @{$c->stash->{pc}->{organization}}, "";
}


=head2 action

action method.

This method catches all the EPP contact actions coming from the contact form 
submission and dispatches the action to the real action method.

=cut

sub action : Local {
    my ( $self, $c) = @_;
 
    $c->session->{lastnamespace} = $c->namespace;

    $c->stash->{template} = $TT_CONTACT;

    $c->forward('validate_and_create_csrf_token');

    my $set = $c->forward('set_stash_from_req_param');
    #$c->log->debug("set_stash_from_req_param returned: $set");

    unless ($set) {
	push @{$c->stash->{opsummary}}, "Nothing to do";
	return;
    }

    my $op = $c->stash->{op};
    my $obj = $c->stash->{obj};
    my $cmd  = $c->stash->{cmd};
    my $srid = $c->stash->{pc}->{srid} || undef;
    my $name = $c->stash->{pc}->{name} || undef;

    if ($op) {
	if ($op eq 'xincrement') {
	    $c->forward('onemore_xemail');
	} elsif ($op eq 'rincrement') {
	    $c->forward('onemore_rolecontact');
	} elsif ($op eq 'sincrement') {
	    $c->forward('onemore_streetline');
	} elsif ($op eq 'oincrement') {
	    $c->forward('onemore_orghandle');
	} elsif ($op eq 'reset') {
	    $c->forward('reset');
	} elsif ($cmd) {
	    ####
	    # All parameters are now stashed. Validate them locally from the stash
	    # before any DRI stuff is called.

	    $c->stash->{cmd_show_status} = 1;

	    my ($valok, @verr) = validate_from_stash ($c);
	    unless ($valok) {
		#$c->log->debug('Validate failed!');
		# $s should contain an array with some validation error descriptions
		push @{$c->stash->{opsummary}}, @verr;
		$c->stash->{cmd_success} = $valok;
		return;
	    }

	    #$c->log->debug("Validated OK, valok: $valok !");

	    # Do a Brreg check for create and update if requested
	    if ($c->config->{brreg}->{org_check} && ($op eq 'create' || $op eq 'update')) {
		my $pc = $c->stash->{pc};
		#print STDERR "pc: ", Dumper $pc;
		if ($pc->{type} eq 'organization' && $pc->{identity}->{type} eq 'organizationNumber' && $c->req->param('brchk')) {

		    #print STDERR "brregchk: looking up identity: ",  $pc->{identity}->{value}, "\n";
		    my ($resok, $orgnm, $errs)  = &NORID::EppClient::Controller::Util::online_brreg_check( $pc->{identity}->{value});
		    #print STDERR "brregchk: resok: $resok, orgnm: $orgnm, errs: @$errs\n";

		    if ($orgnm) {
			# If an org is given, use that, else use name
			my $key = 'name';
			if ($pc->{org}) {
			    $key = 'org';
			}

			if (uc($pc->{$key}) ne uc($orgnm)) {
			    # Mismatch on number/name, tell the user the correct name
			    push @{$c->stash->{opsummary}}, "$BRREG result:";
			    push @{$c->stash->{opsummary}}, " - The organization number was found.";
			    push @{$c->stash->{opsummary}}, " - Mismatch on organization name ['$pc->{$key}' vs. '$orgnm']";
			    if ($set) {
				push @{$c->stash->{opsummary}}, " - Name corrected to '$orgnm' before the $op was performed.";
				$pc->{$key} = $pc->{"n" . $key} = $orgnm;
				$name = $orgnm if ($key eq 'name');
				# Add any warnings from the Brreg lookup
				push @{$c->stash->{opsummary}}, @$errs if ($errs);
                                
				# continue with rest of operation below
			    } else {
				# No changes in data, brreg check only, and mismatch.
				$c->stash->{cmd_success} = 0;
				return;
			    }
			} else {
			    # Match and all was OK.
			    # If no changes, only a brreg check was wanted.
			    unless ($set) {
				$c->stash->{cmd_success} = 1;
				push @{$c->stash->{opsummary}}, "$BRREG check only: Match on organization number and name.";
				return;
			    }
			}
		    } else {
			# lookup did not find the orgname, present the error
			$c->stash->{cmd_success} = 0;
			push @{$c->stash->{opsummary}}, "The organization was NOT found in $BRREG, please provide a valid organization number.";
			push @{$c->stash->{opsummary}}, @$errs if ($errs);
			return;
		    }
		}
	    }

	    # Syntax OK, do the DRI stuff
	    #$c->log->debug('Login before EPP operation needed!');

	    unless ($c->forward('/login/_login_is_ok')) {
		$c->log->info('EppClient: login before EPP operation failed!');
		# Send to the login page
		$c->stash->{template} = $TT_LOGIN;
		return;
	    }

	    $c->log->info('Login before EPP operation SUCCEEDED!');

	    ######
	    # Do the EPP operation, send it to Net::DRI
	    $c->stash->{p} = $c->stash->{pc};
	    if ($c->stash->{facets} && $c->stash->{facets}->{vfyfacets} && $c->stash->{facets}->{eppf}) {
		$c->stash->{p}->{facets} = $c->stash->{facets}->{eppf};
	    }

	    my ($resok, $dri, $rc, $rx) = NORID::EppClient::Model::NORID::DRI::do_epp_command($c, 'hash');

	    $c->log->info("do_epp_command: op: $op, res: $resok");

	    $c->stash->{cmd_success} = $resok;

	    if ($resok) {
		# Success, map the various operation results to stash, summary and session 
		$c->stash->{cmd_pending} = 1 if ($rc->is_pending);
		#$c->log->debug('EppClient: resok!');

		if ($name && $op eq 'create') {
		    push @{$c->stash->{opsummary}},  "$op '$name' succeeded. ";
		} elsif ($srid) {
		    push @{$c->stash->{opsummary}},  "$op '$srid' succeeded. ";
		} else {
		    push @{$c->stash->{opsummary}},  
		    "$op failed, too little data specified, don't know what to do. ";	
		}

		if ($op eq 'info') {
		    $c->stash->{infoobj} = 
			&NORID::EppClient::Model::NORID::DRI::contact_object_as_fmt($dri, 'hash');
		}

		NORID::EppClient::Controller::Util::set_common_results($c, $resok, $dri, $rc, $rx);

		$c->forward("set_".$op."_result");

	    } else {
		# some EPP error, like DRI exception with some error
		$c->log->info('EppClient: NOT resok, some EPP error!');

		#####
		# Collect results from the DRI-operation.
		if ($name && $op eq 'create') {
		    push @{$c->stash->{opsummary}},  "$op '$name' failed. ";
		} elsif ($srid) {
		    push @{$c->stash->{opsummary}},  "$op '$srid' failed. ";
		} else {
		    push @{$c->stash->{opsummary}},  "$op failed, too little data specified, don't know what to do. ";
		}
		NORID::EppClient::Controller::Util::set_common_results($c, $resok, $dri, $rc, $rx);
	    }

	} else {
	    push @{$c->stash->{opsummary}}, "Nothing to do";
	    return;
	}
    } else {
	# No operation
	$c->forward('menu');
    }
}


=head2 get_contact_info

Do an info on an existing contact to fetch the current values.

=cut

sub get_contact_info : Private {
    my ( $self, $c ) = @_;

    $c->stash->{cmd} = 'info';

    my ($resok, $dri, $rc, $rx) = NORID::EppClient::Model::NORID::DRI::do_epp_command($c, 'hash');
    
#    $c->log->debug("EppClient: get_info: srid: " .  $c->stash->{pc}->{srid} . " resok: $resok");
	    
    if ($resok) {
	# Success, map the various results to stash, summary and session 
	# depending of the operation
	$c->stash->{infoobj} = 
	    &NORID::EppClient::Model::NORID::DRI::contact_object_as_fmt($dri, 'hash');
	$c->forward("set_info_result");
    } else {
	# some EPP error, like already exists etc.
	$c->log->info('EppClient: NOT resok, some EPP error!');
	push @{$c->stash->{opsummary}},  "Info failed";
    }
}


=head2 set_check_result

Set result from check operation.

=cut

sub set_check_result : Private {
    my ( $self, $c ) = @_;

    my $dri = $c->stash->{dri};
    my $srid = $c->stash->{pc}->{srid};

    push @{$c->stash->{opsummary}},  "Check '$srid', " . ($dri->get_info('exist')?"contact exists. ":"contact does not exist. ");

    $c->forward('null_data');
#    $c->forward('set_session_from_stash');
    $c->stash->{pc}->{srid} = $srid;
}

=head2 set_info_result

Set result and object data from info operation.

=cut

sub set_info_result : Private {
    my ( $self, $c ) = @_;
    my $dri = $c->stash->{dri};

    my $s = $c->stash->{infoobj};

#    my $s = NORID::EppClient::Model::NORID::DRI::contact_object_as_fmt($dri, 'hash');
    
    $c->forward('set_stash_from_contact_object', [ $s ]);
    $c->forward('set_session_from_stash');
 
}

=head2 set_create_result

Set result from create operation.

When a create is succesful, the object is created. 
A handle is created as well, and we do an info to fetch it, unless pending.

=cut

sub set_create_result : Private {
    my ( $self, $c ) = @_;

    my $dri  = $c->stash->{dri};

    # Info on the created handle if any was returned (not the case for a pending create)
    if (my $srid = $dri->get_info('id')) {
	$c->stash->{pc}->{srid} = $srid;
	# Do a get info to fetch the created data and to get the handle.
	$c->forward('get_contact_info');
    }
}


=head2 set_update_result

Set result from the update operation.

=cut

sub set_update_result : Private {
    my ( $self, $c ) = @_;

    my $dri = $c->stash->{dri};
    my $rc = $c->stash->{rc};
    my $srid = $c->stash->{pc}->{srid};

    unless ($rc && $rc->can('is_pending') && $rc->is_pending) {
	# The new name has been effectuated
	$c->stash->{pc}->{name} = $c->stash->{pc}->{nname} if ($c->stash->{pc}->{nname});
	$c->stash->{pc}->{org}  = $c->stash->{pc}->{norg}  if ($c->stash->{pc}->{norg});
    }
    # Do a get info to fetch the updated data.
    $c->forward('get_contact_info');
}

=head2 set_delete_result

Set result from the delete operation.

When a delete is done successfully, all form fields are cleared except the name.
When a delete is tried, but not succeeded, the current values are still valid, thus all are kept.

=cut

sub set_delete_result : Private {
    my ( $self, $c ) = @_;

    my $dri = $c->stash->{dri};
    my $srid = $c->stash->{pc}->{srid};

    $c->forward('null_data');
#    $c->forward('set_session_from_stash');
    $c->stash->{pc}->{srid} = $srid;
}

=head2 set_stash_from_req_param

Map map the request parameters to the stash.
Also do the detection of what needs to be updated and set the stash 
according to it.

=cut

sub set_stash_from_req_param: Private {
    my ( $self, $c) = @_;

    my $ret = 1;

    my $op = lc($c->req->param('contactop') || "") ||	lc($c->req->param('reset') || "");

    unless ($op) {
	$op = 'xincrement' if $c->req->param('xincrement');
	$op = 'rincrement' if $c->req->param('rincrement');
	$op = 'sincrement' if $c->req->param('sincrement');
	$op = 'oincrement' if $c->req->param('oincrement');
    } 

    $c->stash->{obj} = 'contact';
    $c->stash->{op} = $op;
    $c->stash->{cmd} = lc($c->req->param('contactop') || "");

    #$c->log->debug("op: $op");

    $c->session->{objdetails} = $c->req->param('objdetails');
    $c->session->{cmddetails} = $c->req->param('cmddetails');
    $c->session->{showstatus} = $c->req->param('showstatus');
    $c->session->{showdisclose} = $c->req->param('showdisclose');
    $c->session->{facetmenu}  = $c->req->param('facetmenu');

    # Store any facet values
    &NORID::EppClient::Controller::Util::set_stash_facets_from_req_param($c);

    ####
    # Pick the parameters from the request form

    # map all single scalar form parameters
    # srid is the lookup key which is used for everything but the create
    # but we don't use it for a create.
    # trim trimmable values as they are processed.

    foreach my $e (@cs) {
  	if ($e eq 'name') {
	    $c->req->parameters->{'name'} = $c->req->param('contactname');
	}
  	if ($e eq 'org') {
	    $c->req->parameters->{'org'} = $c->req->param('contactorg');
	}
	if (my $v = $c->req->param($e)) {
	    if ($e eq 'srid') {
		$v = uc($v);
	    }
	    if ($e eq 'pc') {
		if ($v =~ m/^\d\d\d\d$/mx) {
		   my $cc = $c->req->param('cc');
		    # Add 'NO-' 
		   $v = 'NO-' . $v if ($cc && $cc eq 'NO');
		}
	    } elsif ($e eq 'voice' || $e eq 'mobilephone') {
		# strip any spaces from the number
		$v =~ s/\s//g;
	    }
	    $c->stash->{pc}->{$e} = $trimmable{$e} ? NORID::EppClient::Controller::Util::trim($v) : $v;
	}
    }
    # Use disclose parameters only when user wants to use it
    if ($c->session->{showdisclose}) {
	foreach my $e (@discloses, @xdiscloses) {

	    if ($e =~ m/^(voice|email|mobilePhone)disclose$/) {
		my $d = $1;
		# discloses only needed if value exist for its parameter
		if ($c->stash->{pc}->{lc($d)}) {

		    #print STDERR "mphone set: $d:'", $c->stash->{pc}->{lc($d)}, "\n";

		    my $p;
		    if ($d eq 'mobilePhone') {
			$p = "x";
		    } else {
			$p = "";
		    }
		    #print STDERR "disclose: $d is set to: $dis\n";
		    my $dis;
		    $dis = $c->req->param($e) || "0";

		    if ($op eq 'create') {
			# Because default disclose is that full disclose is set,
			# a create only need to set non-disclosed info.
			# Also, this is needed because EPP (and DRI) does not handle 
			# a combination of 0 and 1 in a single EPP-transaction,
			# we need only 0 or 1.
			$c->stash->{pc}->{$p . "disclose"}->{$d} = $dis if ($dis eq "0");
		    } else {
			$c->stash->{pc}->{$p . "disclose"}->{$d} = $dis;
		    }
		}
	    }
	}
    }
    # Set identity type and value
    if (my $v = $c->req->param('identitytype')) {
	$c->stash->{pc}->{identity}->{type} = $v;
    }
    if (my $v = $c->req->param('identityvalue')) {
	$c->stash->{pc}->{identity}->{value} = NORID::EppClient::Controller::Util::trim($v);
    }

    # Grab all arrays, sort all except street
    foreach my $a (@hoa_sort, @hoa_nsort) {

	if (my @v = $c->req->param($a)) {
	    if ($a eq 'organization' || $a eq 'rolecontact') {
		@v = map { uc() } @v;
	    }
	    # Trim unneeded spaces
	    @v = NORID::EppClient::Controller::Util::trim(@v);

	    # Empty arrays may be returned. Wash them so only real values are left.
	    @v = NORID::EppClient::Controller::Util::wash_array(@v) ;

	    if (@v) {
		@v = sort @v unless ($a eq 'street');
		push @{$c->stash->{pc}->{$a}}, @v;
	    }
	}
    }

#print STDERR "set stash from req par: ",  Dumper $c->stash;

    if ($op eq 'update') {
	$ret = $c->forward('detect_and_stash_contact_object_changes');
    }
    return $ret;
}


=head2 validate_from_stash

Validate the parameters not checked locally by the DRI lib.

=cut

sub validate_from_stash {
    my ( $c ) = @_;

    my $op   = $c->stash->{op};
    my $cmd  = $c->stash->{cmd};
    my $name = $c->stash->{pc}->{name};
    my $srid = $c->stash->{pc}->{srid};
    my $type = $c->stash->{pc}->{type} || "";
    my $idt  = $c->stash->{pc}->{identity}->{type} || "";
    my $idv  = $c->stash->{pc}->{identity}->{value} || "";

    my $valok = 1;
    my @e;

    my ($facok, @facerrs) = &NORID::EppClient::Controller::Util::is_valid_facets($c);
    unless ($facok) {
	$valok = 0;
	push @e, "Invalid facets", @facerrs;
    }

    if ($cmd eq 'check' || $cmd eq 'info' ||  $cmd eq 'update' ||  $cmd eq 'delete') {
#	$c->log->debug("Validate not creates: op: $op, cmd: $cmd, srid: '$srid', name: $name");
	unless ($srid) {
	    $valok = 0;
	    push @e, "'$op' requires a handle";
	    $srid = "";   # to awoid warnings
	}
	unless (NORID::EppClient::Controller::Util::isOrgContactID($srid)) {
	    $valok = 0;
	    push @e, "'$op': Invalid Contact ID: $srid";
	}

	# The srid is a sufficient parameter except for update which need further validation.
	unless ($cmd eq 'update') {
	    return $valok, @e;
	}
    } 

    if ($cmd eq 'create' && $srid) {
	$valok = 0;
	push @e, "$op: Contact ID must be empty to verify the create";
    }

    # Check that the mandatory fields has values
    foreach my $el ('type', 'name', 'email', 'cc', 'pc', 'city') {
	unless ( defined($c->stash->{pc}->{$el}) && $c->stash->{pc}->{$el}) {
	    $valok = 0;
	    push @e, "$op: mandatory parameter '$el' not specified";
	}
    }

    # Check that when CC=NO, the postal code value is prefixed with 'NO-' and consists of 4 digits
    my $cc   = $c->stash->{pc}->{cc} || "";
    my $pc   = $c->stash->{pc}->{pc} || "";
    my $city = $c->stash->{pc}->{city} || "";

    if ( $cc eq 'NO' ) {
	 unless ($pc =~ m/^NO-\d\d\d\d$/mx) {
	    $valok = 0;
	    push @e, "$op: Norwegian postal code must be prefixed with 'NO-', followed by 4 digits";
	 }
    }

    # Check that types and values are not in conflict.
    # - Roles must not have a id. value
    # - 

    if ($type eq 'role' || ($type eq 'person' && $idt && $idt ne 'nationalIdentityNumber' && $idt ne 'anonymousPersonIdentifier')) {
	if ($idt) {
	    $valok = 0;
	    push @e, "Illegal '$type' data, cannot have identity type";
	}
	if ($idv) {
	    $valok = 0;
	    push @e, "Illegal '$type' data, cannot have identity value";
	}
    }

    # Organizations must have some id
    if ($type eq 'organization') {

	unless ($idt) {
	    $valok = 0;
	    push @e, "$op: mandatory parameter 'identitytype' not specified";
	}
	unless ($idv) {
	    $valok = 0;
	    push @e, "$op: mandatory parameter 'identityvalue' not specified";
	}
	if ($idt eq 'organizationNumber') {
	    unless ($idv && $idv =~ m/^\d{9}$/mx) {
		$valok = 0;
		push @e, "Illegal organization $BRREG number, must be nine digits";
	    }
	} elsif ($idt eq 'localIdentity') {
	    unless ($idv && $idv =~ m/^N\.ORG\.AUTO$|N\.ORG\.\d+$|N\.LEG\.\d+$/mx) {
		$valok = 0;
		push @e, "Illegal organization local id. number. Syntax: 'N.ORG.AUTO' or 'N.ORG.x..x' or N.LEG.x..x, where 'x..x' are digits only";
	    }
	}

    } elsif ($type eq 'person' && $idt) {
	if ($idt eq 'nationalIdentityNumber') {
	    # 11 digits, assume folkeregister person number until we have anything else
	    unless ($idv && $idv =~ m/^\d{11}$/mx) {
		$valok = 0;
		push @e, "Illegal person 'Folkeregister' identification, must be 11 digits";
	    }
	}
	if ($idt eq 'anonymousPersonIdentifier') {
	    # N.PRI.xxxx syntax assumed 
	    unless ($idv && $idv =~ m/^N\.PRI\.\d+|^-$/mx) {
		$valok = 0;
		push @e, "Illegal person private (anonymous) person identification, must be 'N.PRI.xx..', 'xx..' being digits";
	    }
	} 
    }

    # Max 3 street lines
    my $noa=0;
    my @adr;
    if (defined($c->stash->{pc}->{street}) && @{$c->stash->{pc}->{street}}) {
	@adr = @{$c->stash->{pc}->{street}};
	$noa = scalar(@adr);
	if ($noa > 3) {
	    $valok = 0;
	    push @{$c->stash->{opsummary}}, "Too many address lines, EPP has a maximum of three";
	}
    }

    # Validate email addresses
    if (my $em = $c->stash->{pc}->{email}) {
	unless (Email::Valid->address($em)) {
	    $valok = 0;
	    push @e, "Illegal primary email address '$em'";
	}
    }
    foreach my $em (@{$c->stash->{pc}->{xemail}}) {
	unless (Email::Valid->address($em)) {
	    $valok = 0;
	    push @e, "Illegal secondary email address '$em'";
	}
    }

    #
    # To consider: Do a brreg lookup on the org.number to fetch or verify the name
    #

    # Telephone numbers have their special e164 syntaxes when set
    foreach my $el ('voice', 'mobilephone') {
	if (my $v = $c->stash->{pc}->{$el}) {
	    #$c->log->debug("Validate $el: $v");
	    unless ($v =~ m/^\+[0-9]{1,3}\.[0-9]{1,14}(?:x\d+)?$/) {
		$valok = 0;
		push @e, "Illegal telephone number for '$el': $v (syntax: +47.73551000)";
	    }
	}
    }

    # Validate contact handle syntaxes
    foreach my $o (@{$c->stash->{pc}->{organization}}) {
	unless (NORID::EppClient::Controller::Util::isOrgID($o)) {
	    $valok = 0;
	    push @e, "'$op': Invalid Related org. ID: $o";
	}
    }
    foreach my $r (@{$c->stash->{pc}->{rolecontact}}) {
	unless (NORID::EppClient::Controller::Util::isPersonID($r)) {
	    $valok = 0;
	    push @e, "'$op': Invalid Role contact ID: $r";
	}
    }

    #$c->log->debug("Validate returns: valok: $valok, errs: " . join(", ", @e));
    return $valok, @e;
}

sub detect_and_stash_contact_object_changes: Private {
    my ( $self, $c ) = @_;
  
    # Update always requires a handle
    # check which parameters are changed by comparing to the session values.
    # If a scalar parameter 'xxx' is changed, mark is in pc as name 'nxxx' and 
    # the DRI will pick the 'nxxx'.

    my $update = 0;

    # Detect address change, because the addr-block must be updated then
    my $addr = 0;

    foreach my $e (@cs) {
	my $o = $c->session->{pc}->{$e} || "";
	my $n = $c->stash->{pc}->{$e} || "";

	if (($o || $n) && ($n ne $o)) {
	    # A new value found in this field
	    $c->stash->{pc}->{"n$e" } = $n;
	    
	    if ($e eq 'cc' || $e eq 'pc' || $e eq 'city') {
		++$addr;
	    }
	    ++$update;
	    $c->log->debug("EppClient: action: UPDATE: new $e: $o -> $n");
	}
    }

    # detect changes to identitytype and/or value

    my $idch;
    foreach my $e ('type', 'value') {
	my $o = $c->session->{pc}->{identity}->{$e} || "";
	my $n = $c->stash->{pc}->{identity}->{$e} || "";
	
	if (($o || $n) && ($n ne $o)) {
	    # A new value found in this field
	    $c->stash->{pc}->{nidentity}->{$e} = $n;
	    ++$update;
	    ++$idch;
	    $c->log->debug("EppClient: action: UPDATE: new $e: $o -> $n");
	}
    }
    if ($idch && $idch == 1) {
	# If only type or value is changed, insert the other value to make it 
	# a valid pair
	if (defined($c->stash->{pc}->{nidentity}->{type})) {
	    $c->stash->{pc}->{nidentity}->{value} = $c->stash->{pc}->{identity}->{value};
	}
	if (defined($c->stash->{pc}->{nidentity}->{value})) {
	    $c->stash->{pc}->{nidentity}->{type} = $c->stash->{pc}->{identity}->{type};
	}
    }

    if ($c->session->{showdisclose}) {
	foreach my $e (@discloses, @xdiscloses) {
	    if ($e =~ m/^(voice|email|mobilePhone)disclose$/) {
		my $d = $1;
		
		my $p = "";
		if ($d eq 'mobilePhone') {
		    $p = 'x';
		}
		my $o = $c->session->{pc}->{$p . "disclose"}->{$d} || "0";
		my $n = $c->stash->{pc}->{$p . "disclose"}->{$d} || "0";
		
		if ((defined($o) || defined($n)) && ($n ne $o)) {
		    # A new value found in this field
		    $c->stash->{pc}->{"n".$p."disclose"}->{$d} = $n;
		    ++$update;
		    $c->log->debug("EppClient: action: UPDATE: new $e: $o -> $n");
		}
	    }
	}
    }

    # All array changes except street need to be done as add/delete arrays.
    foreach my $e (@ca) {
	my $n = $c->stash->{pc}->{$e};
	my $o = $c->session->{pc}->{$e};

	if ($n || $o) {
	    my ($ar, $dr) = NORID::EppClient::Controller::Util::make_add_del_list($n, $o);

	    if (($e eq 'street') && (($ar && @$ar) || ($dr && @$dr))) {
		# Some change in the street, so change it to the new value
		delete($c->stash->{pc}->{nstreet}); 
		if ($c->stash->{pc}->{street} && @{$c->stash->{pc}->{street}}) {
		    push @{$c->stash->{pc}->{nstreet}}, @{$c->stash->{pc}->{street}};
		} else {
		    push @{$c->stash->{pc}->{nstreet}}, "";
		}
		++$update;
		++$addr;
	    } else {
		if ($ar && @$ar) {
		    $c->stash->{pc}->{"n$e"}->{add} = $ar;
		    $c->log->debug("UPDATE add: $e: " . (join ",", @$ar));
		    ++$update;
		}
		if ($dr && @$dr) {
		    $c->log->debug("UPDATE del: $e: " . (join ",", @$dr));
		    $c->stash->{pc}->{"n$e"}->{del} = $dr;
		    ++$update;
		}
	    }
	}
    }
    # Incremental addr block is not permitted, if any addr info is changed,
    # the whole block must be set.
    if ($addr) {
	foreach my $e ('cc', 'city', 'pc') {
	    unless ($c->stash->{pc}->{"n$e"}) {
		$c->stash->{pc}->{"n$e"} = $c->stash->{pc}->{$e};
	    }
	}
	unless ($c->stash->{pc}->{nstreet}) {
	    if ($c->stash->{pc}->{street} && @{$c->stash->{pc}->{street}}) {
		push @{$c->stash->{pc}->{nstreet}}, @{$c->stash->{pc}->{street}};
	    } else {
		push @{$c->stash->{pc}->{nstreet}}, "";
	    }
	}
    }
    
    #$c->log->debug("UPDATE: update: $update, addr: $addr");
    return $update;
}

=head2 set_session_from_stash

Put the data from the stash into the session data so we can present 
the previous values in the next page.

=cut

sub set_session_from_stash: Private {
    my ( $self, $c ) = @_;

    $c->session->{pc} = dclone($c->stash->{pc});

}

=head2 set_stash_from_contact_object

We have contact object data in $s
Put the interesting data from it into the stash.

=cut

sub set_stash_from_contact_object: Private {
    my ( $self, $c, $s) = @_;

    $c->forward('null_data');
    
    foreach my $e (@ho) {
	$c->stash->{pc}->{$e} = $s->{$e} if ($s->{$e});
    }
    foreach my $e (@hoa_nsort) {
	my @v = @{$s->{$e}};
	push @{$c->stash->{pc}->{$e}}, @v if @v;
    }
    foreach my $e (@hoa_sort) {
	push @{$c->stash->{pc}->{$e}}, sort @{$s->{$e}} if $s->{$e};
    }
    foreach my $e (@hos) {
	$c->stash->{pc}->{$e} = $s->{$e} || "";
    }
    if (defined($s->{disclose})) {
	$c->stash->{pc}->{disclose} = $s->{disclose};
    }
    if (defined($s->{xdisclose})) {
	$c->stash->{pc}->{xdisclose} = $s->{xdisclose};
    }
}

=head2 reset

Reset form, zap session and stash, and display empty values.

=cut

sub reset : Private {
    my ( $self, $c ) = @_;

    $c->forward('null_data');
    delete ($c->session->{objdetails});
    delete ($c->session->{cmddetails});
    delete ($c->session->{showstatus});
    delete ($c->session->{showdisclose});
    delete ($c->session->{facetmenu});
    delete $c->stash->{facets};
    delete $c->session->{facets};
}

sub null_data : Private {
    my ( $self, $c ) = @_;

    delete $c->stash->{pc};
    delete $c->session->{pc};
}

=head2 index

Show the contact page.

=cut

sub index : Private {
    my ( $self, $c ) = @_;

    $c->forward('menu');
}

=head2 validate_and_create_csrf_token

CSRF validation.

=cut

sub validate_and_create_csrf_token: Private {
    my ( $self, $c) = @_;

    &NORID::EppClient::Controller::Util::validate_and_create_csrf_token($self, $c, 'c');

}


=head1 AUTHOR

Trond Haugen

=head1 LICENSE

This library is free software, you can redistribute it and/or modify
it under the same terms as Perl itself.

=cut

1;
