package NORID::EppClient::Controller::Domain;

use strict;
use warnings;

use Net::DRI;
use NORID::EppClient::Model::NORID::DRI;
use NORID::EppClient::Controller::Util;
use Log::Log4perl qw(get_logger :levels);
use base qw(Catalyst::Controller::RequestToken);
use Email::Valid;
use Storable qw(dclone);
use Net::LibIDN  ':all';
use Encode qw / is_utf8 /; 
use Archive::Zip qw(:ERROR_CODES);
use DateTime;
use DateTime::TimeZone;
use Data::Validate qw /is_between is_integer is_hex/;
use MIME::Base64;

# For Debug purposes:
use Data::Dumper;
$Data::Dumper::Indent=1;

my $TT_DOMAIN = 'domain.tt2';
my $TT_LOGIN = 'login.tt2';

my $MIN_NAMESERVERS = 2;

=pod

=head1 NAME

EppClient - Domain object operations

=head1 DESCRIPTION

The module supports creation and maintainance of domain objects for .NO. 

=head1 SYNOPSIS

=head1 DESCRIPTION

package NORID::EppClient::Controller::Domain - Catalyst Contact Controller

=head1 METHODS

=cut

# Domain scalar values
my @ds = ('name', 'registrant', 'pw', 'token', 'duration', 
	  'vfydelete', 'vfyrenew', 'vfywithdraw', 
	  'deletefromdns', 'deletefromregistry', 
	  );

my @dsall = ('name', 'ace', 'registrant', 'pw', 'pwsetbuthidden', 'token', 'duration', 
	  'vfydelete', 'vfyrenew', 'vfywithdraw', 'curexpiry', 
           'vfytransferwauthinfo', 'vfytransferwtoken', 
           'vfyapplicantdatasetupdate', 'vfydnssecupdate',
	  'deletefromdns', 'deletefromregistry', 'status'
	   );

my @dsapplicantdataset = ('versionnumber', 'acceptdate', 'acceptname' );


# Domain array values
my @da = ('nsset', 'tech', 'statuslist');


#################
# DNSSEC 
#
my $MAX_DSSET_KEY_SIZE = 5000;

# dsData
my @dnsseca  = ('keyTag', 'alg', 'digestType', 'digest');

# keyData (optional)
my @dnssecao = ('key_flags', 'key_protocol', 'key_alg', 'key_pubKey');

# variables that can be trimmed
my %trimmable = map { $_ => 1 } @da, @ds, @dsapplicantdataset, @dnsseca, @dnssecao;

# To hash from object or session
my @dos   = ('roid', 'clID', 'crID', 'upID', 'crDate', 'upDate', 'exDate');

##
# Mapping from the Applicant dataset file from the Norid EE-automaton
# to internal attribute values used for local checks. 
my %ds_file_map = (
    'Domain name'        => 'domainname',
    'Domain name ACE'    => 'domainname_ace',
    'Applicant name'     => 'applicant_name',
    'Applicant id'       => 'applicant_identity',
    );

##
# Mapping from the Applicant dataset file from the Norid EE-automaton
# to attribute values in Net::DRI.
my %ds_dri_map = (
    'Version number'     => 'versionnumber',
    'Accept date'        => 'acceptdate',
    'Accept name'        => 'acceptname',
    );


#### DNSSEC stuff:

# Test values until registry returns registered values in the domain-info:
my %dnssec_default =  ();

my %xdnssec_default = (

    # tags must be equal, alg and key_alg must be equal to form 
    # a unique DS set.

    'ds' => {
	'111-3-1' => {
	    keyTag     => '111',
	    alg        => '3',
	    digestType => '1',
	    digest     => '111333AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA',
	},

	'333-8-2' => {
	    keyTag     => '333',
	    alg        => '8',
	    digestType => '2',
	    digest     => '22CCCBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBB',
	},

	'222-5-2' => {
	    keyTag     => '222',
	    alg        => '5',
	    digestType => '2',
	    digest     => '22CCCBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBB',
	},
    },

    # two keys corresponding to two of the three DS posts over
    'key' => {
	'111-3-257' => {
	    key_keyTag   => '111',
	    key_alg      => '3',
	    key_flags    => '257',
	    key_protocol => '3',
	    key_pubKey   => 'cGxlYXN1cmUu',
	},
   	'222-3-257' => {
	    key_keyTag   => '222',
	    key_alg      => '3',
	    key_flags    => '257',
	    key_protocol => '3',
	    key_pubKey   => 'YXN1cmUu',
	}
    }
    );

my %dnssec_add_ds = (

    '11199-3-1' => {
	keyTag     => '11199',
	alg        => '3',
	digestType => '1',
	digest     => 'DADDAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA',
    },
    
    '22222-8-1' => {
	keyTag     => '22222',
	alg        => '8',
	digestType => '1',
	digest     => 'DADDAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA',
    },  
    );

 my %dnssec_add_key = (
     
     '22222-8-257' => {
	 key_keyTag   => '22222',
	 key_alg      => '8',
	 key_flags    => '257',
	 key_protocol => '3',
	 key_pubKey   => 'ZWFzdXJlLg==',
     },
    '333-8-257' => {
	 key_keyTag   => '22222',
	 key_alg      => '8',
	 key_flags    => '257',
	 key_protocol => '3',
	 key_pubKey   => 'cGxlYXN1cmUu',
     },
    );


=head2 menu 

menu method.

=cut

sub menu : Local {
    my ( $self, $c ) = @_;

    $c->stash->{pd} = $c->session->{pd} if ($c->session->{pd} && !defined($c->stash->{pd}));
    $c->stash->{pdc} = $c->session->{pdc} if ($c->session->{pdc} && !defined($c->stash->{pdc}));

    # Recreate token when menu is pressed, this will cleanup any errors
    $c->session->{'d_token'} = $self->create_token($c);

    $c->stash->{template} = $TT_DOMAIN;

    #print STDERR Dumper $c->config->{dnssec};
}

=head2 onemore_ns

Add another name server line

=cut

sub onemore_ns : Private {
    my ( $self, $c ) = @_;
   
    if (NORID::EppClient::Controller::Util::wash_array($c->req->param('nsset'))) {
	push @{$c->stash->{pd}->{nsset}}, "";
    }
}

=head2 onemore_tech

Add another tech contact line

=cut

sub onemore_tech : Private {
    my ( $self, $c ) = @_;
   
    push @{$c->stash->{pd}->{coset}->{tech}}, "";
}

=head2 onemore_dnssec_ds

Add another DNSSEC DS record line

=cut

sub onemore_dnssec_ds : Private {
    my ( $self, $c ) = @_;

    #print STDERR "one more dnssec DS before: ", Dumper $c->stash->{pd}->{secdns}->{ds};
    
    %{$c->stash->{pd}->{secdns}->{ds}->{'99999-?-?'}} = (keyTag => "99999", alg => "?", digestType => "?", digest => "?");

    #print STDERR "one more dnssec DS after: ", Dumper $c->stash->{pd}->{secdns}->{ds};
}


=head2 oneless_dnssec_ds

Remove a DNSSEC DS record, usng the keyTag as reference

=cut

sub oneless_dnssec_ds : Private {
    my ( $self, $c ) = @_;

    # Remove the selected DS record from stash.
    # The entry is identified by the checkbox value, with syntax 'keyTag - alg - digestType'

    my @del = $c->req->param('dnssec_ds_todelete');

    #print STDERR "one less dnssec DS before: ", Dumper $c->stash->{pd}->{secdns}->{ds};

    foreach my $dsp (@del) {
	#print STDERR "Delete DS $dsp\n";
	delete($c->stash->{pd}->{secdns}->{ds}->{$dsp}) if ($c->stash->{pd}->{secdns}->{ds}->{$dsp});
    }

    push @{$c->stash->{opsummary}}, 
    "Selected DSs are removed from screen.", 
    "Check 'verify update' and then click 'Update' to complete operation!";
    $c->stash->{cmd_pending} = 1;
    $c->stash->{cmd_show_status} = 1;

    #print STDERR "one less dnssec DS after: ", Dumper $c->stash->{pd}->{secdns}->{ds};
}

=head2 onemore_dnssec_key

Add another DNSSEC key line

=cut

sub onemore_dnssec_key : Private {
    my ( $self, $c ) = @_;

    #print STDERR "one more dnssec KEY before: ", Dumper $c->stash->{pd}->{secdns}->{key};

    %{$c->stash->{pd}->{secdns}->{key}->{'99999-?-257'}} = (key_keyTag => "99999", key_alg => "?", key_flags => 257, key_protocol => 3, key_pubKey => "?");

    #print STDERR "one more dnssec KEY after: ", Dumper $c->stash->{pd}->{secdns}->{key};
}


=head2 oneless_dnssec_key

Remove a DNSSEC key line, usng the keyTag as reference

=cut

sub oneless_dnssec_key : Private {
    my ( $self, $c ) = @_;

    # Remove the selected KEY from stash.
    # The entry is identified by the checkbox value, with syntax 'keyTag - alg - flags'

    my @del = $c->req->param('dnssec_key_todelete');

    #print STDERR "one less dnssec KEY before: ", Dumper $c->stash->{pd}->{secdns}->{key};

    foreach my $dsp (@del) {
	#print STDERR "Delete KEY $dsp\n";
	delete($c->stash->{pd}->{secdns}->{key}->{$dsp}) if ($c->stash->{pd}->{secdns}->{key}->{$dsp});
    }
    push @{$c->stash->{opsummary}}, 
    "Selected keys are removed from screen.", 
    "Check 'verify update' and then click 'Update' to complete operation!";
    $c->stash->{cmd_pending} = 1;
    $c->stash->{cmd_show_status} = 1;

    #print STDERR "one less dnssec KEY after: ", Dumper $c->stash->{pd}->{secdns}->{key};
}

=head2 secdatapaste

Add another DS or DNSKEY from a text box.

=cut

sub secdatapaste : Private {
    my ( $self, $c ) = @_;

    #print STDERR "Paste DS/DNSSKEY data ", Dumper $c->stash->{pd}->{secdns}->{secdatabox};

    $c->stash->{pd}->{secdns}->{ds_key_paste} = 1;

}

=head2 asetshow_toggle

Toggle explode/collapse display of applicant dataset

=cut

sub asetshow_toggle : Private {
    my ( $self, $c ) = @_;

    #print STDERR "Toggle show of aset data\n";
    $c->session->{appdataset_show} = !$c->session->{appdataset_show};

}

=head2 secdatashow_toggle

Toggle explode/collapse display of DNSSEC data

=cut

sub secdatashow_toggle : Private {
    my ( $self, $c ) = @_;

    #print STDERR "Toggle show of DNSSEC data\n";
    $c->session->{secdata_show} = !$c->session->{secdata_show};
}


=head2 asetfileload

Load data from application dataset file.

=cut

sub asetfileload : Private {
    my ( $self, $c ) = @_;

    my $zfn = $c->stash->{pd}->{applicantdataset_fn};

    unless ($zfn) {
        push @{$c->stash->{opsummary}}, "No dataset file specified";
        $c->stash->{cmd_success} = 0;
        return;
    }

    #print STDERR "FILE LOAD: $zfn\n";                                                                                                                           

    my $upload = $c->req->upload('asetfilename');

    #print STDERR "Upload: ", Dumper $upload;                                                                                                                    

    $c->stash->{cmd_show_status} = 1;
    $c->stash->{cmd_success}     = 1;

    # Validate file existence, format and syntax                                                                                                                 
    unless ($upload) {
        push @{$c->stash->{opsummary}},  "Load dataset from file failed";
        $c->stash->{cmd_success} = 0;
        return;
    }
    # Check that some kind of a zip file MIME-type has been loaded.                                                                                              
    # Only log the problem if not, as the unpack_dataset_file()                                                                                                  
    # should trap real format problems anyway.                                                                                                                   
    unless ($upload->type =~ m/zip/) {
        my $err = "Dataset file is not a zip file: " . $upload->type;
        $c->log->debug($err);
    }

    # Unpack the zip file and parse it to get the ds and fs data                                                                                                 
    my ($resok, $errs, $ds, $fs) = unpack_dataset_file($upload->tempname);
    unless ($resok) {
        push @{$c->stash->{opsummary}}, @$errs if ($errs && @$errs);
        $c->stash->{cmd_success} = 0;
        return;
    }
    #print STDERR "unpack called ok\n";                                                                                                                          
    #print STDERR "ds: ", Dumper $ds;                                                                                                                            
    #print STDERR "fs: ", Dumper $fs;                                                                                                                            

    ###
    # Verify that:
    # - all the domains in the dataset file are in correct syntax
    # - the domainname in question, if stated, is among the domain names in the dataset file

    my $idn = $c->stash->{pd}->{name} || "";
    my $ace = $c->stash->{pd}->{ace} || "";
    my ($dsidn, $dsace);

    if ( $fs && exists($fs->{domainname} ) ) {
	
	my $match;
	
	foreach my $domainname (@{$fs->{domainname}}) {
	    #print STDERR "dname: '$domainname', idn: $idn, ace: $ace\n";
	    
	    $dsidn = lc($domainname);
	    die "Cannot lower case domain name: $dsidn" unless ($dsidn);
	    
	    $dsace = idn_to_ascii( $dsidn, 'utf-8', IDNA_USE_STD3_ASCII_RULES );
	    die "Cannot convert domain to ace" unless ($dsace);
	    
	    $dsidn = idn_to_unicode( $dsace, 'utf-8', IDNA_USE_STD3_ASCII_RULES );
	    die "Cannot convert domain to ace" unless ($dsidn);
	    
	    # The domain name to be created/updated must be found
	    if ($ace eq $dsace) {
		#print STDERR "MATCH ON $ace and '$domainname'\n";
		++$match;
	    }
	}
	if ($ace && !$match) {
	    push @{$c->stash->{opsummary}}, "Domain name '$idn' is not found among the domain names in the applicant dataset in the zip file.";
	    push @{$c->stash->{opsummary}}, "The zip file contains the following domain names: " .	(join ", ", @{$fs->{domainname}});
	    $c->stash->{cmd_success} = 0;
	    return;
	}
    }
    
    # Populate the stash with the three ds values since all is well
    $c->stash->{pd}->{applicantdataset} = $ds;
    $c->stash->{pd}->{applicantdataset_fs}->{fs} = $fs;


}

=head2 secfileload

Load DNSSEC DS or DNSKEY data from data file.

=cut

sub secfileload : Private {
    my ( $self, $c ) = @_;
   
    my $kfn = $c->stash->{pd}->{key_fn};

    unless ($kfn) {
	push @{$c->stash->{opsummary}}, "No DS or key file specified";
	$c->stash->{cmd_success} = 0;
	return;
    }

    my $upload = $c->req->upload('secfilename');

    $c->stash->{cmd_show_status} = 1;
    $c->stash->{cmd_success}     = 1;

    # Validate file existence, format and syntax
    unless ($upload) {
	push @{$c->stash->{opsummary}},  "Load key from file failed";
	$c->stash->{cmd_success} = 0;
	return;
    }

    # Read the DS or key file and parse it to get the ds or key data
    my ($resok, $errs, $secd) = parse_dsset_key_file($c, $upload);
    unless ($resok) {
        push @{$c->stash->{opsummary}}, @$errs if ($errs && @$errs);
        $c->stash->{cmd_success} = 0;
        return;
    }

    # Verify that the domainname in the key is the same as the one entered for the form.
    # Require the domain name parameter to be specified.
    my $ace = $c->stash->{pd}->{ace};
    unless ($ace) {
	push @{$c->stash->{opsummary}},  "Please specify a domain name!";
	$c->stash->{cmd_success} = 0;
        return;
    }
    unless ($ace eq $secd->{keyids}->{domainname}) {
	push @{$c->stash->{opsummary}}, "The domain name entered '$ace' does not match the domain name specified in the DS or key file '$secd->{keyids}->{domainname}'!\n";
	$c->stash->{cmd_success} = 0;
        return;
    }

    $c->forward('map_secd', [$secd]);

}


=head2 secdataload

Load DNSSEC DS or DNSKEY data from the text box

=cut

sub secdataload : Private {
    my ( $self, $c ) = @_;
   
    $c->stash->{cmd_show_status} = 1;
    $c->stash->{cmd_success}     = 1;

    my $d = $c->req->param('secdatabox');
    unless ($d) {
	push @{$c->stash->{opsummary}}, "No DS/DNSKEY data specified";
	$c->stash->{cmd_success} = 0;
	$c->stash->{pd}->{secdns}->{ds_key_paste} = 1;
	return;
    }

    if (length($d) > $MAX_DSSET_KEY_SIZE) {
	push @{$c->stash->{opsummary}}, "DS/DNSKEY text size too big, maximum size accepted is $MAX_DSSET_KEY_SIZE.";
	$c->stash->{cmd_success} = 0;
	$c->stash->{pd}->{secdns}->{ds_key_paste} = 1;
        return;
    }

    my ($resok, $errs, $secd) = parse_dsset_key_text($c, $d);
    unless ($resok) {
        push @{$c->stash->{opsummary}}, @$errs if ($errs && @$errs);
        $c->stash->{cmd_success} = 0;
        return;
    }

    # Verify that the domainname in the key is the same as the one entered for the form.
    # Require the domain name parameter to be specified.
    my $ace = $c->stash->{pd}->{ace};
    unless ($ace) {
	push @{$c->stash->{opsummary}},  "Please specify a domain name!";
	$c->stash->{cmd_success} = 0;
        return;
    }
    unless ($ace eq $secd->{keyids}->{domainname}) {
	push @{$c->stash->{opsummary}}, "The domain name entered '$ace' does not match the domain name specified in the input data '$secd->{keyids}->{domainname}'!\n";
	$c->stash->{cmd_success} = 0;
        return;
    }
    $c->forward('map_secd', [$secd]);

    # $errs may be filled with warning, present if so
    if ($errs && @$errs) {
       push @{$c->stash->{opsummary}}, @$errs if ($errs && @$errs);
       $c->stash->{cmd_success} = 0;
       return;
    }

}


=head1 map_secd

Verify that the domain name matches the one for the DS or key.
ACE format is already assumed.

=cut

sub map_secd : Private {
    my ( $self, $c, $secd ) = @_;


    ###
    # Map the loaded DS or key data into the secdns structure for display
    # Each time we get here, either a dsset entry OR key entry has been loaded
    #

    #print STDERR "map loaded keys into stash: ", Dumper $secd;
    #print STDERR "secfileload, stash before is           : ", Dumper $c->stash->{pd}->{secdns};

    my $stdsa   = $c->stash->{pd}->{secdns}->{ds} if  $c->stash->{pd}->{secdns}->{ds};
    my $fidsa   = $secd->{keyids}->{ds} if $secd->{keyids}->{ds};

    my $stkeya   = $c->stash->{pd}->{secdns}->{key} if $c->stash->{pd}->{secdns}->{key};
    my $fikeya   = $secd->{keyids}->{key} if $secd->{keyids}->{key};

    if (defined($fidsa) && $fidsa) {
	foreach my $fids (keys %$fidsa) {

	    #print STDERR "loading DS file entry on id: $fids: ", Dumper($fidsa->{$fids});

	    # replace any DS entry in stash ($stdsa) if a DS entry with same keyTag/alg/digestType exists, 
	    # else add it as a new
	    my %updated;
	    foreach my $sids (keys %$stdsa) {
		if ($fids eq $sids) {
		    # same id, replace stashed entry with the one loaded
		    #print STDERR "Updated DS: ", Dumper $stdsa->{$sids} , " to ", Dumper $fidsa->{$sids};
		    $c->stash->{pd}->{secdns}->{ds}->{$sids} = $fidsa->{$sids};
	
		    # log the ones updated, any loaded DSs not updated must be added instead.
		    $updated{$sids} = 1;
		}
	    }
	    foreach my $fids (keys %$fidsa) {
		unless ($updated{$fids}) {
		    # new DSs, add them
		    $c->stash->{pd}->{secdns}->{ds}->{$fids} = $fidsa->{$fids};
		}
	    }
	}

    } elsif (defined($fikeya) && $fikeya) {
	
	foreach my $fids (keys %$fikeya) {
	    
	    #print STDERR "loading KEY file entry on id: $fids: ", Dumper($fikeya->{$fids});
	    
	    # replace any KEY entry in stash ($stkeya) if a KEY entry with same keyTag/alg/flags combo exists, else add it as a new
	    
	    my %updated;
	    foreach my $sids (keys %$stkeya) {
		if ($fids eq $sids) {
		    # same id, replace stashed entry with the one loaded
		    #print STDERR "Updated KEY: ", Dumper $stkeya->{$sids} , " to ", Dumper $fikeya->{$sids};
		    $c->stash->{pd}->{secdns}->{key}->{$sids} = $fikeya->{$sids};
		    
		    # log the ones updated, any loaded KEYs not updated must be added instead.
		    $updated{$sids} = 1;
		}
	    }
	    foreach my $fids (keys %$fikeya) {
		unless ($updated{$fids}) {
		    # new KEY, add it
		    $c->stash->{pd}->{secdns}->{key}->{$fids} = $fikeya->{$fids};
		}
	    }
	}

    }
    #print STDERR "secfileload, stash after is           : ", Dumper $c->stash->{pd}->{secdns};

}


=head2 action

action method.

This method catches all the EPP contact actions coming from the contact form 
submission and dispatches the action to the real action method.

=cut

sub action : Local {
    my ( $self, $c, $done) = @_;

    $c->session->{lastnamespace} = $c->namespace;

    $c->stash->{template} = $TT_DOMAIN;

    # Perform a CSRF validation if not already done by the caller
    unless ($done) {
	$c->forward('validate_and_create_csrf_token');
    }

    my $set = $c->forward('set_stash_from_req_param');
    #$c->log->debug("set_stash_from_req_param returned: $set");
    unless ($set) {
	push @{$c->stash->{opsummary}}, "Nothing to do";
	return;
    }

    my $op = $c->stash->{op};
    my $obj = $c->stash->{obj};
    my $cmd  = $c->stash->{cmd};
    my $name = $c->stash->{pd}->{name} || undef;

    #print STDERR "op: $op\n";

    delete $c->stash->{pdc};
    delete $c->session->{pdc};

    if ($op) {

	if ($op eq 'tincrement') {
	    $c->forward('onemore_tech');

	} elsif ($op eq 'nsincrement') {
	    $c->forward('onemore_ns');

	} elsif ($op eq 'dnssecincrement_ds') {
	    $c->forward('onemore_dnssec_ds');

	} elsif ($op eq 'dnssecincrement_key') {
	    $c->forward('onemore_dnssec_key');

	} elsif ($op eq 'dnssecdecrement_ds') {
	    $c->forward('oneless_dnssec_ds');

	} elsif ($op eq 'dnssecdecrement_key') {
	    $c->forward('oneless_dnssec_key');

	} elsif ($op eq 'asetfileload') {
	    $c->forward('asetfileload');

	} elsif ($op eq 'secfileload') {
	    $c->forward('secfileload');

	} elsif ($op eq 'secdatapaste') {
	    $c->forward('secdatapaste');

	} elsif ($op eq 'secdataload') {
	    $c->forward('secdataload');

	} elsif ($op eq 'asetshow_toggle') {
	    $c->forward('asetshow_toggle');

	} elsif ($op eq 'secdatashow_toggle') {
	    $c->forward('secdatashow_toggle');

	} elsif ($op eq 'reset') {
	    $c->forward('reset');

	} elsif ($cmd) {

	    ####
	    # All parameters are now stashed. Validate them locally from the stash
	    # before any DRI stuff is called.
	    my ($valok, @verr) = validate_from_stash ($c);
	    unless ($valok) {
		#$c->log->debug('Validate failed!');
		# $s should contain an array with some validation error descriptions
		push @{$c->stash->{opsummary}}, @verr;
		$c->stash->{cmd_show_status} = 1;
		$c->stash->{cmd_success} = $valok;
		return;
	    }
	    #$c->log->debug("Validated OK, valok: $valok !");
	    ###
	    # Syntax OK, do the DRI stuff

	    # Perform a login if not already done by the caller
	    unless ($done) {
		$c->log->info("cmd: $cmd, op: $op. Login before EPP operation needed.");
		unless ($c->forward('/login/_login_is_ok')) {
		    $c->log->info('EppClient: login before EPP operation failed!');
		    # If either of above don't work out, send to the login page
		    $c->stash->{template} = $TT_LOGIN;
		    return;
		}
		$c->log->info('Login before EPP operation SUCCEEDED!');
	    }

	    my ($resok, $dri, $rc, $rx);

	    if ($op eq 'lookupnotify') {
		# LookupNotify is special, is attacks the contact objects instead
		# Use a dedicated get method for it to fetch the  possible contact notify addresses

		my $res = $c->forward('lookup_notify_addresses');

		$resok = $c->stash->{lnaret}->{resok};
		$dri   = $c->stash->{lnaret}->{dri};
		$rc    = $c->stash->{lnaret}->{rc};
		$rx    = $c->stash->{lnaret}->{rx};

	    } else {

		######
		# Do the normal domain EPP operations
		$c->stash->{p} = $c->stash->{pd};
		if ($c->stash->{facets} && $c->stash->{facets}->{vfyfacets} && $c->stash->{facets}->{eppf}) {
		    $c->stash->{p}->{facets} = $c->stash->{facets}->{eppf};
		}

		#print STDERR "p: ", Dumper  $c->stash->{p};
	
		my $n = "";
		$n = "n" if ($cmd eq 'update');
		# DRI need decoded UTF-8 parameters to fit into its XML.
		#&NORID::EppClient::Controller::Util::decode_p($c->stash->{p}, $n);

		($resok, $dri, $rc, $rx) = NORID::EppClient::Model::NORID::DRI::do_epp_command($c, 'hash');
		
		$c->log->debug(&NORID::EppClient::Controller::Util::encode_unicode
			       ("do_epp_command: op: $op, name: $name, res: $resok")
			       );
	    }

	    #print STDERR "op: $op, resok: $resok\n";

	    $c->stash->{cmd_show_status} = 1;
	    $c->stash->{cmd_success} = $resok;

	    if ($resok) {
		# Success, map the various operation results to stash, summary and session 
		$c->stash->{cmd_pending} = 1 if ($rc->is_pending);
		$c->log->info('EppClient: resok!');

		if ($name) {
		    push @{$c->stash->{opsummary}},  "$op '$name' succeeded. ";
		} else {
		    push @{$c->stash->{opsummary}},  "$op failed, too little data specified, don't know what to do. ";
		}
		if ($op eq 'info') {
		    $c->stash->{infoobj} = 
			&NORID::EppClient::Model::NORID::DRI::domain_object_as_fmt($dri, 'hash');

		    #print STDERR "stash infoobj: ", Dumper $c->stash->{infoobj}, "\n";

		}

		NORID::EppClient::Controller::Util::set_common_results($c, $resok, $dri, $rc, $rx);

		$c->forward("set_".$op."_result") unless ($op eq 'lookupnotify');

	    } else {
		# some EPP error, like DRI exception with some error
		$c->log->info('EppClient: NOT resok, some EPP error!');
	
		#####
		# Collect results from the DRI-operation.
		if ($name) {
		    push @{$c->stash->{opsummary}},  "$op '$name' failed. ";
		} else {
		    push @{$c->stash->{opsummary}},  "$op failed, too little data specified, don't know what to do. ";
		}
		NORID::EppClient::Controller::Util::set_common_results($c, $resok, $dri, $rc, $rx);
	    }
	} else {
	    push @{$c->stash->{opsummary}}, "Nothing to do";
	    return;
	}

    } else {
	$c->forward('menu');
    }
}

=head2 get_domain_info

Do an info on an existing domain to fetch the current values.

=cut

sub get_domain_info : Private {
    my ( $self, $c ) = @_;

    #$c->log->debug("EppClient: get_domain_info: op: " .  $c->stash->{op});

    $c->stash->{cmd} = 'info';

    my ($resok, $dri, $rc, $s) = NORID::EppClient::Model::NORID::DRI::do_epp_command($c, 'hash');
    
    if ($resok) {
	#$c->log->debug("EppClient: get_domain_info, resok=OK");
	# Success, map the various results to stash, summary and session 
	# depending of the operation
	$c->stash->{infoobj} = 
	    &NORID::EppClient::Model::NORID::DRI::domain_object_as_fmt($dri, 'hash');
	$c->forward("set_info_result");
    } else {
	#$c->log->debug("EppClient: get_domain_info, resok=FAILED");
	# some EPP error, like already exists etc.
	#$c->log->debug('EppClient: NOT resok, some EPP error!');
	push @{$c->stash->{opsummary}},  "Info failed";
    }
}


=head2 set_check_result

Set result from check operation.

=cut

sub set_check_result : Private {
    my ( $self, $c ) = @_;

    my $dri = $c->stash->{dri};
    my $name = $c->stash->{pd}->{name};

    # If domain exists, it may actually be one of the prohibited etc. which cannot be registered.
    # Present the more detailed exist_reason if so
    if ($dri->get_info('exist')) {
	if (my $exreason = $dri->get_info('exist_reason')) {
	    # Also present reason
	    push @{$c->stash->{opsummary}},  "$exreason.";
	}
    } else {
	push @{$c->stash->{opsummary}}, "Domain does not exist.";
    }
    $c->forward('null_data');
#    $c->forward('set_session_from_stash');
    $c->stash->{pd}->{name} = $name;
}

=head2 set_info_result

Set result and object data from info operation.

=cut

sub set_info_result : Private {
    my ( $self, $c ) = @_;
    my $dri = $c->stash->{dri};

    my $srid = $c->stash->{pd}->{srid};

    my $s = NORID::EppClient::Model::NORID::DRI::domain_object_as_fmt($dri, 'hash');

    #    print STDERR "exDate in info: ", $s->{exDate}, "\n";

    #print STDERR "set info res: s:", Dumper $s, "\n";

    $c->forward('set_stash_from_domain_object', [ $s ]);
    $c->forward('set_session_from_stash');
 
}

=head2 set_create_result

Set result from create operation.

When a create is succesful, the object is created. 
A handle is created as well, and we do an info to fetch it.

=cut

sub set_create_result : Private {
    my ( $self, $c ) = @_;

    my $dri  = $c->stash->{dri};
    # Get the created handle 
    my $srid   = $dri->get_info('id') if $dri->get_info('id');
    $c->stash->{pd}->{srid} = $srid if ($srid);

    # Do a get info to fetch the created data and to get the handle.
    # If the create is pending, info will fail, so don't do that.
    $c->forward('get_domain_info') unless ($c->stash->{cmd_pending});
}


=head2 set_update_result

Set result from the update operation.

=cut

sub set_update_result : Private {
    my ( $self, $c ) = @_;

    my $dri = $c->stash->{dri};
    my $rc = $c->stash->{rc};
    my $srid = $c->stash->{pd}->{srid};

    unless ($rc && $rc->can('is_pending') && $rc->is_pending) {
	# The new name has been effectuated
	$c->stash->{pd}->{name} = $c->stash->{pd}->{nname} if ($c->stash->{pd}->{nname});
    }

    # Do a get info to fetch the updated or at least, when pending, latest data.
    $c->forward('get_domain_info');
}

=head2 set_delete_result

Set result from the delete operation.

When a delete is done successfully, all form fields are cleared except the name.
When a delete is tried, but not succeeded, the current values are still valid, thus all are kept.

=cut

sub set_delete_result : Private {
    my ( $self, $c ) = @_;

    my $dri = $c->stash->{dri};
    my $srid = $c->stash->{pd}->{srid};

    #$c->forward('null_data');
    $c->stash->{pd}->{srid} = $srid;

}


=head2 set_withdraw_result

Set result from withdraw operation.

When a withdraw is succesful, only the rcode is interesting,
and we only need dummy here

=cut

sub set_withdraw_result : Private {
    my ( $self, $c ) = @_;

    # Do a get info to fetch the updated data.
    $c->forward('get_domain_info');
}

=head2 set_transfer_result

Set result from transfer init (request token) operation.

When a transfer init (request token) is succesful, only the rcode
is interesting, no data are changed.

=cut

sub set_transfer_result : Private {
    my ( $self, $c ) = @_;

    return 1;
}

=head2 set_transfer_query_result

Set result from transfer_query operation.

A query does not chenge any data, so no need to reread stuff.

=cut

sub set_transfer_query_result : Private {
    my ( $self, $c ) = @_;

    return 1;
}

=head2 set_transfer_cancel_result

Set result from trancfer_cancel operation.

When a transfer_cancel is succesful, statuses etc. are changed,
so we reread the object to get updated object data.

=cut

sub set_transfer_cancel_result : Private {
    my ( $self, $c ) = @_;

    # Do a get info to fetch the updated data.
    $c->forward('get_domain_info');
}

=head2 set_transfer_execute_result

Set result from transfer execute operation.

When a transfer execute is succesful, the sponsor and statuses are
updated so we reread the object to get updated object data.

=cut

sub set_transfer_execute_result : Private {
    my ( $self, $c ) = @_;

    # Do a get info to fetch the updated data.
    $c->forward('get_domain_info');
}

=head2 set_renew_result

Set result from renew operation.

When a renew is succesful, the expiry date is updated so we 
reread the object to get updated object data.

=cut

sub set_renew_result : Private {
    my ( $self, $c ) = @_;

    # Do a get info to fetch the updated data.
    $c->forward('get_domain_info');

    # The renew parameters in case the renew failed

}

=head2 lookup_notify_addresses

Lookup all notify addresses.

Returning arguments from the method is done via the stash, sice it seems as if
only a single scalar can be returned in a catalyst forward situation

=cut

sub lookup_notify_addresses: Private {
    my ( $self, $c ) = @_;

    # A LookupNotify is done
    # Lookup registrant contacts which are the possible targets
    # for notify address selections

    my $op = $c->stash->{op};

    delete($c->stash->{pdc});
    delete($c->session->{pdc});

    # Lookup the possible contact notify addresses
    $c->stash->{obj} = 'contact';
    $c->stash->{cmd} = 'info'; 

    #print STDERR "pd: ", Dumper  $c->stash->{pd};

    # get the registrant
    $c->stash->{p}->{srid} = $c->stash->{pd}->{registrant};

    my ($resok, $dri, $rc, $rx) = NORID::EppClient::Model::NORID::DRI::do_epp_command($c, 'hash');
    #$c->log->debug("EppClient: get_info: srid: " .  $c->stash->{p}->{srid} . " resok: $resok");

    unless ($resok) {
	# return error
	$c->stash->{lnaret}->{resok} = $resok;
	$c->stash->{lnaret}->{dri} = $dri;
	$c->stash->{lnaret}->{rc} = $rc;
	$c->stash->{lnaret}->{rx} = $rx;
	return 0;
    }

    my $e = 'registrant';
    $c->forward('set_lookup_notify_result', [$e]);

    ###
    # massage the results a bit:

    # remove duplicates and store them
    foreach my $v ('email', 'xemail', 'mobilephone') {
	   
	if (exists($c->stash->{pdc}->{$e}->{$v})) {
	    my @vs = @{$c->stash->{pdc}->{$e}->{$v}};
	    next unless @vs; 
	    #print STDERR "Before wash: $e $v: '@vs'\n";
	    @vs = NORID::EppClient::Controller::Util::wash_array(@vs);
	    delete $c->stash->{pdc}->{$e}->{$v};
	    push @{$c->stash->{pdc}->{$e}->{$v}}, sort @vs;
	    #print STDERR "After wash: $e $v: '@vs'\n";
	}
    }

    $c->session->{pdc} = $c->stash->{pdc};
    # print STDERR "After wash: ", Dumper  $c->stash->{pdc};
    #print STDERR "Returning ressok: $resok\n";
    $c->stash->{lnaret}->{resok} = $resok;
    $c->stash->{lnaret}->{dri} = $dri;
    $c->stash->{lnaret}->{rc} = $rc;
    $c->stash->{lnaret}->{rx} = $rx;

    return 1;
}

=head2 set_lookup_notify_result

We have fetched the contact object, put the relevant email addresses on
a dedicated pdc-hash.

=cut

sub set_lookup_notify_result : Private {
    my ( $self, $c, $what ) = @_;

    # A contact info is done
    my $dri = $c->stash->{dri};
    my $s = NORID::EppClient::Model::NORID::DRI::contact_object_as_fmt($dri, 'hash');
 
    foreach my $e ('mobilephone', 'email') {
#	print STDERR "what: '$what': s $e: ", $s->{$e}, "\n";
	push @{$c->stash->{pdc}->{$what}->{$e}}, $s->{$e} if ($s->{$e});
    }
    push @{$c->stash->{pdc}->{$what}->{xemail}}, sort @{$s->{xemail}} if $s->{xemail};
}


=head2 set_stash_from_req_param

Map map the request parameters to the stash.
Also do the detection of what needs to be updated and set the stash 
according to it.

=cut

sub set_stash_from_req_param: Private {
    my ( $self, $c) = @_;

    my $ret = 1;

    my $op = lc($c->req->param('domainop') || "") || lc($c->req->param('reset') || "");

    unless ($op) {
	$op = 'tincrement'     if $c->req->param('tincrement');
	$op = 'aincrement'     if $c->req->param('aincrement');
	$op = 'nsincrement'    if $c->req->param('nsincrement');

	$op = 'dnssecincrement_ds'  if $c->req->param('dnssecincrement_ds');
	$op = 'dnssecincrement_key' if $c->req->param('dnssecincrement_key');

	$op = 'dnssecdecrement_ds'   if $c->req->param('dnssecdecrement_ds');
	$op = 'dnssecdecrement_key'  if $c->req->param('dnssecdecrement_key');

	$op = 'asetfileload'   if $c->req->param('asetfileload');
	$op = 'secfileload'    if $c->req->param('secfileload');

	$op = 'secdatapaste'   if $c->req->param('secdatapaste');
	$op = 'secdataload'    if $c->req->param('secdataload');

	$op = 'asetshow_toggle' if $c->req->param('asetshow_toggle');

	$op = 'secdatashow_toggle' if $c->req->param('secdatashow_toggle');
    } 

    $c->stash->{cmd} = lc($c->req->param('domainop') || "");
    $c->stash->{obj} = 'domain';

    # The transfer cmds are special
    if ($op eq 'transfer_wauthinfo' || $op eq 'transfer_wtoken') {
        # If a transfer_execute is ordered, we must indicate to DRI whether to use 
	# the authInfo or the token, do that via the reqcmd
	$c->stash->{pd}->{reqcmd} = $op;
	$op = 'transfer_execute';
	$c->stash->{cmd} = $op;

    } elsif ($op eq 'transfercancel') {
	$op = 'transfer_cancel';
	$c->stash->{cmd} = $op;

    } elsif ($op eq 'transferquery') {
	$op = 'transfer_query';
	$c->stash->{cmd} = $op;

    } elsif ($op eq 'requesttoken') {
	# transfer_init, need to request a token 
	$op = 'transfer';
	$c->stash->{cmd} = $op;
	if (my $n = $c->req->param('notify')) {
	    # set the notify parameter as requested
	    my $what = 'email';
	    # determine whether the value is an email or a mobilephone
	    # if it starts with a '+', it's a phone number
	    if ($n =~ m/^\+/) {
		$what = 'mobilephone';
	    }
	    $c->stash->{pd}->{notify} = { $what => $n };
	}
    } elsif ($op eq 'asetfileload') {
	$c->stash->{pd}->{applicantdataset_fn} = $c->req->param('asetfilename');

	#print STDERR "Aset filename : ",  $c->req->param('asetfilename'), "\n";

    } elsif ($op eq 'secfileload') {
	$c->stash->{pd}->{key_fn} = $c->req->param('secfilename');
    }
    $c->stash->{op} = $op;

    #$c->log->debug("op: $op");

    $c->session->{objdetails}    = $c->req->param('objdetails');
    $c->session->{cmddetails}    = $c->req->param('cmddetails');
    $c->session->{showstatus}    = $c->req->param('showstatus');
    $c->session->{dnssecdetails} = $c->req->param('dnssecdetails');
    $c->session->{transfercmds}  = $c->req->param('transfercmds');
    $c->session->{editcmds}      = $c->req->param('editcmds');
    $c->session->{facetmenu}     = $c->req->param('facetmenu');

    # Store any facet values
    my $forceupdate = &NORID::EppClient::Controller::Util::set_stash_facets_from_req_param($c);

    ####
    # Pick the parameters from the request form

    # map all single scalar form parameters
    # srid is the lookup key which is used for everything but the create
    # but we don't use it for a create.
    # trim trimmable values as they are processed.

    foreach my $e (@dsall) {
        # domainname and name is the same field, but we want the browser
        # to remember the field-name and need a unique form-name.
  	if ($e eq 'name') {
	    my $idn = lc($c->req->param('domainname') || "");

	    if ($idn) {
		# Add .no in case the name is witten without it
		$idn = NORID::EppClient::Controller::Util::trim($idn);

		# Allow 'no' to be able to look up '.no' zone.
		# Otherwise add '.no' to the end of domain name if missing
		if ($idn eq '.no' || $idn eq 'no.') {
		    $idn = 'no';
		}
		unless ($idn eq 'no') {
		    unless ($idn =~ m/.+\.no$/) {
			$idn .= '.no';
		    }
		}
	    }
	    
	    my $oidn = $idn;
	    
	    if (my $ace = idn_to_ascii($idn, 'utf-8', IDNA_USE_STD3_ASCII_RULES)) {
		if (my $ridn = idn_to_unicode($ace, 'utf-8', IDNA_USE_STD3_ASCII_RULES)) {
		    $idn = $ridn;
		    unless (is_utf8($idn)) {
			$idn = Encode::decode('UTF-8', $idn, 1);
		    }
		}
		$c->stash->{pd}->{ace} = $ace;
	    }
	    $c->req->parameters->{'name'} = $oidn;
	}
	if (my $v = $c->req->param($e)) {
	    $v = "" unless $v;
	    if ($e eq 'registrant') {
		$v = uc($v);
	    }
	    $v = $trimmable{$e} ? NORID::EppClient::Controller::Util::trim($v) : $v;

	    if ($e eq 'duration') {
		$c->stash->{pd}->{$e} = {months => $v};
	    } else {
		$c->stash->{pd}->{$e} = $v;
	    }
	    #print STDERR "$e is '$v'\n";
	}
    }

    foreach my $e (@dsapplicantdataset) {
	if (my $v = $c->req->param($e)) {
	    $v = "" unless $v;
	    $v = $trimmable{$e} ? NORID::EppClient::Controller::Util::trim($v) : $v;
	    $c->stash->{pd}->{applicantdataset}->{$e} = $v;
	}
    }

    # Grab all arrays, and sort them
    foreach my $a (@da) {
	if (my @v = $c->req->param($a)) {

	    if ($a eq 'tech') {
		@v = map { uc() } @v;
	    } elsif ($a eq 'nsset') {
		@v = map { lc() } @v;
	    }

	    # Trim unneeded spaces
	    @v = NORID::EppClient::Controller::Util::trim(@v);
	    @v = $trimmable{$a} ? NORID::EppClient::Controller::Util::trim(@v) : @v;

	    # Empty arrays may be returned. Wash them so only real values are left.
	    @v = NORID::EppClient::Controller::Util::wash_array(@v) ;
	    
	    if (@v) {
		@v = sort @v;

		#print STDERR "a: $a, v: ", Dumper \@v, "\n";
		
		if ($a eq 'nsset' || $a eq 'statuslist' ) {
		    push @{$c->stash->{pd}->{$a}}, @v;
		    #print STDERR "stashed status: $a: ", scalar @v, " ", Dumper($c->stash->{pd}->{$a}), "\n";
		    
		} else {
		    push @{$c->stash->{pd}->{coset}->{$a}}, @v;
		}
	    }
	    #print STDERR "stashed: $a array: @v\n";
	}
    }


    # Grab all DNSSEC arrays, the array indexes need to be kept, so handle one 
    # offset at the time.
    my @ktag      = $c->req->param('keyTag');
    my @key_ktag  = $c->req->param('key_keyTag');

    #print STDERR "before stashed DNSSEC DS and key: ", Dumper $c->stash->{pd}->{secdns}, "\n";
 
    if (@ktag) {

	# If delete all is selected, that is the only operation that will be performed
	if ($c->req->param('dnssecdeleteall')) {
	    $c->stash->{pd}->{secdns_delete_all} = 1;
	}
	    
	# the number of entries in ktag, which also has alg set, determines the number of DS-entries
	@ktag = $trimmable{'keyTag'} ? NORID::EppClient::Controller::Util::trim(@ktag) : @ktag;

	my @alg  = $c->req->param('alg');
	@alg     = $trimmable{'alg'} ? NORID::EppClient::Controller::Util::trim(@alg) : @alg;

	my @digt = $c->req->param('digestType');
	@digt    = $trimmable{'digestType'} ? NORID::EppClient::Controller::Util::trim(@digt) : @digt;

	my @dig  = $c->req->param('digest');
	@dig     = $trimmable{'digest'} ? NORID::EppClient::Controller::Util::trim(@dig) : @dig;

	# Strip any spaces from the digest and convert to upper case.
	# This because a dsset is observed to contain some spaces from some sources.

	# Also convert to upper case since EPP requires all upper in digest:
        # ref. http://www.w3.org/TR/2001/REC-xmlschema-2-20010502/#hexBinary
	@dig = map { uc($_) } @dig;
	@dig = map { s/\s+//g; $_ } @dig;

	#print STDERR "ktag     : ", Dumper \@ktag, "\n";
	#print STDERR "alg      : ", Dumper \@alg, "\n";

	my $ix = 0;
	my $d;
	foreach my $tag (@ktag) {
	    #print STDERR "adding DS tag $tag\n";
	    
	    if ($alg[$ix]) {
		# DS data
		#print STDERR "adding DS  $tag for ix: $ix, alg: $alg[$ix]\n";

		# A DS is unique by it's keyTag, alg and digestType.
		# A DS can be generated with different digestTypes, both with same keyTag and alg,
		# so we need to handle those as separate DSs.
		# Several DSs with different algs must be handled as well, those should have different tags, though.

		my $a =  $tag . '-' . $alg [$ix] . '-' . $digt[$ix];
		
		$d->{$a}->{keyTag}     = $tag;
		$d->{$a}->{alg}        = $alg [$ix];
		$d->{$a}->{digestType} = $digt[$ix];
		$d->{$a}->{digest}     = $dig [$ix];
	    }
	    ++$ix;
	}
	#print STDERR "DS d after: ", Dumper $d, "\n";

	if (defined($d) && $d) {
	    $c->stash->{pd}->{secdns}->{ds} = $d;
	}
    }
 
    if (@key_ktag) {
	# the number of entries in ktag, which also has key_alg set, determines the number of Key-entries

	@key_ktag = $trimmable{'keyTag'} ? NORID::EppClient::Controller::Util::trim(@key_ktag) : @key_ktag;

	my @key_flags    = $c->req->param('key_flags');
	@key_flags       = $trimmable{'key_flags'} ? NORID::EppClient::Controller::Util::trim(@key_flags) : @key_flags;

	my @key_alg      = $c->req->param('key_alg');
	@key_alg         = $trimmable{'key_alg'} ? NORID::EppClient::Controller::Util::trim(@key_alg) : @key_alg;

	my @key_protocol = $c->req->param('key_protocol');
	@key_protocol    = $trimmable{'key_protocol'} ? NORID::EppClient::Controller::Util::trim(@key_protocol) : @key_protocol;

	my @key_pubKey   = $c->req->param('key_pubKey');
	@key_pubKey      = $trimmable{'pubKey'} ? NORID::EppClient::Controller::Util::trim(@key_pubKey) : @key_pubKey;
	# remove with space since such might exist for readability, but can be discarded.
	@key_pubKey      = map { s/\s+//g; $_ } @key_pubKey;

	#print STDERR "key_flags: ", Dumper \@key_flags, "\n";
	#print STDERR "key_alg  : ", Dumper \@key_alg, "\n";

	my $ix = 0;
	my $d;
	foreach my $tag (@key_ktag) {
	    #print STDERR "adding KEY tag $tag\n";

	    if ($key_alg[$ix]) {
		# key data on this tag
		# A Key is unique by it's keyTag, alg. and flags,
		# so we can handle keys with flags 256 and 257 if such should be handled.

		#print STDERR "adding KEY $tag for ix: $ix, key_alg: $key_alg[$ix]\n";

		my $a =  $tag . '-' . $key_alg [$ix] . '-' . $key_flags[$ix];

		$d->{$a}->{key_keyTag}   = $tag;
		$d->{$a}->{key_alg}      = $key_alg[$ix];
		$d->{$a}->{key_flags}    = $key_flags[$ix];
		$d->{$a}->{key_protocol} = $key_protocol[$ix];
		$d->{$a}->{key_pubKey}   = $key_pubKey[$ix];
	    }
	    ++$ix;
	}
	#print STDERR "KEY d after: ", Dumper $d, "\n";
	if (defined($d) && $d) {
	    $c->stash->{pd}->{secdns}->{key} = $d;
	}
    }
    #print STDERR "stashed DNSSEC DS and key after: ", Dumper $c->stash->{pd}->{secdns}, "\n";

    push @{$c->stash->{pd}->{secdns_merged}}  , make_dnssec_merged($c->stash->{pd}->{secdns});

    #print STDERR "op: $op, pd stash before set update       : ", Dumper $c->stash->{pd}->{secdns}, "\n--\n"; 
    #print STDERR "op: $op, pd stash   merged before set update: ", Dumper $c->stash->{pd}->{secdns_merged}, "\n--\n"; 
    #print STDERR "op: $op, pd session merged before set update: ", Dumper $c->session->{pd}->{secdns_merged}, "\n--\n"; 

    if ($op eq 'update') {
	$ret = $c->forward('detect_and_stash_domain_object_changes');
	$ret ||= $forceupdate;
	#print STDERR "op: $op, pd stash after set update: ", Dumper $c->stash->{pd}, "\n"; 
    }
    
    #print STDERR "op: $op, pd stash etter set update: ", Dumper $c->stash->{pd}->{secdns}, "\n"; 

    return $ret;
}

=head2 make_dnssec_merged

Map the secdns key and ds to a merged secdns_merged entry for use by net-dri operations
Merging means that keys that matchs a DS should be added to that ds.
Merge shall be done only if the key_keyTag AND the key_alg is the same as in a DS

=cut

sub make_dnssec_merged {
    my $s = shift;

    my @dkeys = keys(%{$s->{ds}})  if ($s->{ds});
    my @kkeys = keys(%{$s->{key}}) if ($s->{key});

    my $ds  = dclone($s->{ds})  if ($s->{ds});
    my $key = dclone($s->{key}) if ($s->{key});

    my @merged;
    foreach my $d (@dkeys) {
	# Add DS, as it shall always be there,
	# insert key info if any key exists on tag and alg.
	#print STDERR "merging: ", Dumper $ds->{$d};
	foreach my $k (@kkeys) {
	    if ($key->{$k}->{key_keyTag} == $ds->{$d}->{keyTag} &&
		$key->{$k}->{key_alg}    == $ds->{$d}->{alg}) {
		# copy the key values into DS
		foreach my $e ('key_keyTag', 'key_flags', 'key_alg', 'key_protocol', 'key_pubKey') {
		    $ds->{$d}->{$e} = $key->{$k}->{$e};
		}
	    }
	}
	push @merged, $ds->{$d};
    }
    #print STDERR "merged result ", Dumper \@merged;

    return @merged;
}



=head2 validate_from_stash

Validate the parameters needed for each operation.
Some are also checked locally by the DRI lib.
The registry server will validate also.

=cut

sub validate_from_stash {
    my ( $c ) = @_;

    my $op   = $c->stash->{op};
    my $cmd  = $c->stash->{cmd};
    my $reqcmd  = $c->stash->{pd}->{reqcmd};
    my $name = $c->stash->{pd}->{name} || "";
 
    #print STDERR "op: $op, pd: ", Dumper $c->stash->{pd};

    my $priv_domain = 0;
    my $private_registrant = $c->stash->{pd}->{registrant} && 
	NORID::EppClient::Controller::Util::isPersonID($c->stash->{pd}->{registrant});

    my $valok = 1;
    my @e;

    unless ($name) {
	$valok = 0;
	push @e, "'$op' requires a name";
    }
    unless (NORID::EppClient::Controller::Util::is_a_valid_domainname($name)) {
	$valok = 0;
	push @e, "$op: illegal domainname: '$name'";
    }

    if ($name =~ m/^.+\.priv.no$/) {
	$priv_domain = 1;

	unless ($c->config->{client}->{use_anonymousid}) {
	    $valok = 0;
	    push @e, "$op: private domain names not supported: '$name'";
	}
    }

    my ($facok, @facerrs) = &NORID::EppClient::Controller::Util::is_valid_facets($c);
    unless ($facok) {
	$valok = 0;
	push @e, "Invalid facets", @facerrs;
    }

    #print STDERR "cmd: $cmd, op: $op\n";

    # The name is a sufficient parameter for info, check, transfer_query and transfer_cancel
    if ($cmd eq 'info' || $cmd eq 'check' || $cmd eq 'transfer_query' || $cmd eq 'transfer_cancel' || $cmd eq 'transfer') {
	return $valok, @e;
    }

    if ($op eq 'renew') {
	unless ( $c->stash->{pd}->{vfyrenew} ) {
	    $valok = 0;
	    push @e, "Renew requires a checked 'verify' field";
	}
	unless ($c->stash->{pd}->{curexpiry}) {
	    push @e, "'$op' requires curexpiry";
	    $valok=0;
	}
	return $valok, @e;

    } elsif ($op eq 'withdraw') {
	unless ( $c->stash->{pd}->{vfywithdraw} ) {
	    $valok = 0;
	    push @e, "Withdraw requires a checked 'verify' field";
	}
	return $valok, @e;

    } elsif ($op eq 'delete') {
	unless ( $c->stash->{pd}->{vfydelete} ) {
	    $valok = 0;
	    push @e, "Delete requires a checked 'verify' field";
	}

	# When specified, the dates must both be specified and valid!
	my $dd = $c->stash->{pd}->{deletefromdns};
	my $dr = $c->stash->{pd}->{deletefromregistry};

	my $dvalid = 1;
	if (defined($dd) || defined($dr)) {
	    unless ($dd && $dr) {
		$valok = 0;
		push @e, "Delete with dates requires both of the dates";
	    }
	    unless ($dd && NORID::EppClient::Controller::Util::is_a_valid_date($dd)) {
		$valok = 0;
		$dvalid = 0;
		push @e, "Delete requires a valid 'Delete from DNS' date";
	    }
	    unless ($dr && NORID::EppClient::Controller::Util::is_a_valid_date($dr)) {
		$valok = 0;
		$dvalid = 0;
		push @e, "Delete requires a valid 'Delete from Registry' date";
	    }
	    if ($dd && $dr && $dvalid) {
		# Also verify that $dd <= $dr
		my $dde = NORID::EppClient::Controller::Util::date_as_timestamp($dd);
		my $dre = NORID::EppClient::Controller::Util::date_as_timestamp($dr);
		unless ($dde <= $dre) {
		    $valok = 0;
		    push @e, "'Delete from DNS' must be <= 'Delete from Registry'";
		}
		# require delete from DNS not in the past compared to todays date
		my $todayd = NORID::EppClient::Controller::Util::timestamp_as_date;
		my $todays = NORID::EppClient::Controller::Util::date_as_timestamp($todayd);
		unless ($dde >= $todays) {
		    $valok = 0;
		    push @e, "'Delete from DNS' must be >= todays date '$todayd'";
		}
	    }
	}
	return $valok, @e;

    } elsif ($op eq 'transfer_execute') {
	# only name is required for 'requesttoken' which is the 'transfer' operation,
	# so nothing to validate.
	# Also the transfer cancel and query only needs a name.

	# A transfer may optionally be done with a renew period set.
	# So, if a duration is set, we need a renew verify as well.

	if (defined($c->stash->{pd}->{duration})  && $c->stash->{pd}->{duration}) {
	    unless ( $c->stash->{pd}->{vfyrenew} ) {
		$valok = 0;
		push @e, "A transfer with renew requires a checked renew 'verify' field";
	    }
	}
	if ($reqcmd eq 'transfer_wauthinfo') {
	    unless ( $c->stash->{pd}->{pw} ) {
		# accept empty password for domains without a pw.
		# $valok = 0;
		# push @e, "Transfer with authInfo requires an auth info password";
	    }
	    unless ( $c->stash->{pd}->{vfytransferwauthinfo} ) {
		$valok = 0;
		push @e, "Transfer with authInfo requires a checked 'verify' field";
	    }
	    # If vfyrenew is checked, the user wants to renew during transfer. 
	    # An explicit duration period has to be specified then.
	    if ( $c->stash->{pd}->{vfyrenew} && (!defined($c->stash->{pd}->{duration})) ) {
		$valok = 0;
		push @e, "Transfer with authInfo and renewal requires a ' Renew for' period";
	    }
	} elsif ($reqcmd eq 'transfer_wtoken') {
	    unless ( $c->stash->{pd}->{token} ) {
		$valok = 0;
		push @e, "Transfer with token requires a token";
	    }
	    unless ( $c->stash->{pd}->{vfytransferwtoken} ) {
		$valok = 0;
		push @e, "Transfer with token requires a checked 'verify' field";
	    }
	}
	return $valok, @e;

    } elsif ($op eq 'lookupnotify') {
	# Requires a  registrant contact
	unless ( $c->stash->{pd}->{registrant} ) {
	    $valok = 0;
	    push @e, "Lookup notify addresses requires a registrant";
	}
	return $valok, @e;
    }


    #####
    # For all other commands (eg. create/update), we require a valid parameter set.
    # 

    #####
    # mandatory parameters check

    # Check that the mandatory fields has values
    foreach my $el ('name', 'registrant') {
	unless ( $c->stash->{pd}->{$el}) {
	    $valok = 0;
	    push @e, "$op: mandatory parameter '$el' not specified";
	}
	#
	# A domain name change is not allowed during update
	if ($el eq 'name' && $op eq 'update') {
	    # https://git.intern.norid.no/root/Draupne/issues/22, use {ace} not {name}
	    my $o = $c->session->{pd}->{ace} || "";
	    my $n = $c->stash->{pd}->{ace} || "";
	    my $on = $c->session->{pd}->{name} || "";
	    my $nn = $c->stash->{pd}->{name} || "";
	    #print STDERR "sess        : $o\n";
	    #print STDERR "pd          : $n\n";
	    #print STDERR "stash pdace : ", $c->stash->{pd}->{ace}, "\n";
	    #print STDERR "sess  pdace : ", $c->session->{pd}->{ace}, "\n";
	    #print STDERR "stash pdname: ", $c->stash->{pd}->{name}, "\n";
	    #print STDERR "sess  pdname: ", $c->session->{pd}->{name}, "\n";

	    if ($o && $n && ($o ne $n)) {
		$valok = 0;
		push @e, "$op: domain name change is not allowed: $on -> $nn";
       		#print STDERR "  $op: domain name change is not allowed: $on -> $nn\n";
	    }
	}
    }
    unless ( $c->stash->{pd}->{nsset} && @{$c->stash->{pd}->{nsset}} ) {
	$valok = 0;
	push @e, "$op: mandatory parameter 'name server' not specified";
    } else {
	my @v = @{$c->stash->{pd}->{nsset}};
	if (@v < $MIN_NAMESERVERS) {
	    $valok = 0;
	    push @e, "$op: minimum 2 name servers are required for .no";
	}
	foreach my $h (@v) {
	    unless (NORID::EppClient::Controller::Util::is_a_valid_hostname($h)) {
		$valok = 0;
		push @e, "$op: illegal hostname: '$h'";
	    }
	}
    }

    # registrant
    if ($c->config->{client}->{use_anonymousid}) {
	# Must be set to allow for private registrants
	if ($priv_domain) {
	    unless ($private_registrant) {
		# Only a private person can be registrant under priv.no
		$valok = 0;
		push @e, "$op: Invalid Registrant ID, must be a private person contact: '" . ($c->stash->{pd}->{registrant} || "") . "'"; 
	    }
	} else {
	    # normal domain
	    if ($private_registrant) {
		unless ($c->config->{client}->{allow_private_no_registrant}) {
		    # Normal domain and a private registrant, this require that the 'allow_private_no_registrant' is configured to
		    # enable and allow for this variant.
		    $valok = 0;
		    push @e, "$op: Private person as a registrant is not permitted: '" . ($c->stash->{pd}->{registrant} || "") . "'"; 
		}
	    } else {
		# Only an organization can be registrant
		unless (NORID::EppClient::Controller::Util::isOrgID($c->stash->{pd}->{registrant})) {
		    $valok = 0;
		    push @e, "$op: Invalid Registrant ID, must be an organization contact: '" . ($c->stash->{pd}->{registrant} || "") . "'";
		}
	    }
	}

    } else {
	# Only an organization can be registrant
	unless (NORID::EppClient::Controller::Util::isOrgID($c->stash->{pd}->{registrant})) {
	    $valok = 0;
	    push @e, "$op: Invalid Registrant ID, must be an organization contact: '" . ($c->stash->{pd}->{registrant} || "") . "'";
	}
    }

    # Tech-c is mandatory and must be a role.
    
    unless ( $c->stash->{pd}->{coset}->{tech} && 
	     @{$c->stash->{pd}->{coset}->{tech}} ) {
	$valok = 0;
	push @e, "$op: mandatory parameter 'tech' contact not specified";
    }

    foreach my $o (@{$c->stash->{pd}->{coset}->{tech}}) {
	unless (NORID::EppClient::Controller::Util::isRoleID($o)) {
	    $valok = 0;
	    push @e, "$op: Invalid Tech contact ID: '$o', must be a role";
	}
    }
    
    #
    # Applicant dataset validation:
    #
    # 1) If any of the three parameters in the dataset is set, all must be set and be valid.
    #
    # 2) if show_applicant_dataset = 0, the parameters are optional
    #
    # 3) if show_applicant_dataset = 1, the parameters are required for a domain-create,
    #    and also for a domain-update in case of a registrar change
    #

    if ($c->stash->{pd}->{applicantdataset}) {

	# 1) the three mandatory fields need to be checked
	foreach my $el (  @dsapplicantdataset ) {
	    unless ( $c->stash->{pd}->{applicantdataset}->{$el} ) {
		$valok = 0;
		push @e, "Applicant dataset is missing a mandatory element: $el";
	    }
	}

	my $at = $c->stash->{pd}->{applicantdataset}->{acceptdate} || "" ;
	unless (NORID::EppClient::Controller::Util::is_a_valid_epp_date_time($at)) {
	    $valok = 0;
	    push @e, "Accept date/time must be set to a valid EPP datetime, syntax: '2011-10-12T14:02:37Z'";
	}
	
	my $vn = $c->stash->{pd}->{applicantdataset}->{versionnumber} || "" ;
	unless ($vn =~ m/^\d+\.\d+$/mx) {
	    $valok = 0;
	    push @e, "Version number '$vn' is invalid (syntax is 'x.y', where x and y are numbers)";
	}

    } else {
	# no dataset specified
	if ($c->config->{client}->{show_applicant_dataset}) {
	    if ($op eq 'create') {
		# dataset MUST ALWAYS be specified
		$valok = 0;
		push @e, "Applicant dataset is missing, it is required for a 'domain-create'";

	    } elsif ($op eq 'update') {
		if ($c->stash->{pd}->{nregistrant}) {

		    #print STDERR "pd: ", Dumper $c->stash->{pd};
		    #print STDERR "pd: ", Dumper $c->session->{pd};
			
		    # A dataset MUST be specified for domains created after 2013-01-02,
		    # ref. https://www.norid.no/no/registrar/nytt/erklaring111212
		    # which was when electronic EE became a requirement.

		    my $dom_create      = substr($c->session->{pd}->{crDate}, 0, 10);
		    my $tstamp_create   = NORID::EppClient::Controller::Util::date_as_timestamp($dom_create);
		    my $tstamp_ee_start = NORID::EppClient::Controller::Util::date_as_timestamp('2013-01-03');
		    #print STDERR "$dom_create = $tstamp_create, $tstamp_ee_start\n";
		    
		    if ($tstamp_create >= $tstamp_ee_start) {
			$valok = 0;
			push @e, "Applicant dataset is missing, it is required for a 'domain-update' and change of registrant";
		    }
		}
	    }
	}
    }

    if ($c->stash->{pd}->{napplicantdataset}) {
	# Implies update operation
	# An update operation has been trigged and a change in the dataset has been detected
	unless ( $c->stash->{pd}->{vfyapplicantdatasetupdate} ) {
	    $valok = 0;
	    push @e, "Update of Applicant dataset requires a checked 'verify update' field";
	}
    } else {
	# if vfyapplicantdatasetupdate is checked, the dataset should really be 
	# changed as well, reject update attempt if not to avoid unintended updates.
	if ($c->stash->{pd}->{vfyapplicantdatasetupdate}         &&
	    $c->stash->{pd}->{applicantdataset}->{versionnumber} &&
	    $op eq 'update') {
	    $valok = 0;
	    push @e, "Update of Applicant dataset requires a changed dataset";
	}
    }

    # secdns validation
    #  - keyTag    : 16 bit integer,  i.e. 0<=keyTag<2^16
    #  - alg       : 8 bit integer, and one of config.dnssec.algorithms
    #  - digestType: 8 bit integer, and one of config.dnssec.digestTypes
    #  - digest    : hexnumbers only and bytelength as in config.dnssec.digestTypesLength
    #
    if ($c->config->{dnssec}->{enabled} && $c->stash->{pd}->{secdns}) {

	# DS record validation
	my $ds = $c->stash->{pd}->{secdns}->{ds};

	#print STDERR "Validate DS elements: ", Dumper $ds, "\n";

	foreach my $ds_elem (keys %$ds) {
	    my $ds_el =  $ds->{$ds_elem};
	    #print STDERR "Validate DS element: ", Dumper $ds_el, "\n";
	    
	    my $ktag = $ds_el->{keyTag};
	    unless (defined($ktag)) {
		$valok = 0;
		push @e, "$op: mandatory parameter 'keyTag' not specified";
	    } else {
		unless (defined(is_integer($ktag))) {
		    $valok = 0;
		    push @e, "keyTag must be an integer: '$ktag'";
		} else {
		    unless (defined(is_between($ktag, 0, 65535))) {
			$valok = 0;
			push @e, "keyTag must be an integer in the range 0-65535: '$ktag'";
		    }
		}
	    }
	    return $valok, @e unless ($valok);

	    my $tag = "keyTag $ktag: ";

	    if (defined($ds_el->{alg})) {
		# DS record to validate
		my $alg  = $ds_el->{alg};
		my $digt = $ds_el->{digestType};
		my $dig  = $ds_el->{digest};
		
		# alg:
		unless (defined($alg)) {
		    $valok = 0;
		    push @e, "$op: $tag DS mandatory parameter 'alg' not specified";
		} else {
		    unless (defined(is_integer($alg))) {
			$valok = 0;
			push @e, "$tag DS alg must be an integer: '$alg'";
		    } else {
			unless ($c->config->{dnssec}->{algorithms}->{$alg}) {
			    $valok = 0;
			    push @e, "$tag DS alg is not among the permitted algorithm values: $alg, permitted are: " . 
				join(", ", sort( {$a <=> $b} keys(%{$c->config->{dnssec}->{algorithms}})));
			}
		    }
		}
		return $valok, @e unless ($valok);

		# digestType:
		unless (defined($digt)) {
		    $valok = 0;
		    push @e, "$op: $tag DS mandatory parameter 'digestType' not specified";
		} else {
		    unless (defined(is_integer($digt))) {
			$valok = 0;
			push @e, "$tag DS alg must be an integer: '$alg'";
		    } else {
			unless ($c->config->{dnssec}->{digestTypes}->{$digt}) {
			    $valok = 0;
			    push @e, "$tag DS digestType is not among the permitted types: $digt, permitted are: " . 
				join(", ", sort( {$a <=> $b} keys(%{$c->config->{dnssec}->{digestTypes}})));
			}
		    }
		}
		return $valok, @e unless ($valok);
			
		# digest:
		unless (defined($dig)) {
		    $valok = 0;
		    push @e, "$op: $tag DS mandatory parameter 'digest' not specified";
		} else {
		    my $dlen = length($dig);
		    unless ($dlen == $c->config->{dnssec}->{digestTypesLength}->{$digt}) {
			$valok = 0;
			push @e, "$tag DS invalid digest length $dlen\n($dig),\ndigestType $digt must have digest length of ". $c->config->{dnssec}->{digestTypesLength}->{$digt} . " bytes";
		    } else {
			unless ($dig =~ /^\p{IsXDigit}+$/) {
			    $valok = 0;
			    push @e, "$tag DS invalid digest ($dig), must be hex numbers only";
			}
		    }
		}
	    }
	}

	#
	# KEYs
	#

	my $key = $c->stash->{pd}->{secdns}->{key};
	#print STDERR "Validate KEY elements: ", Dumper $key, "\n";

	foreach my $key_elem (keys %$key) {
	    my $key_el =  $key->{$key_elem};
	    #print STDERR "Validate KEY element: ", Dumper $key_el, "\n";

	    my $ktag = $key_el->{key_keyTag};
	    unless (defined($ktag)) {
		$valok = 0;
		push @e, "$op: mandatory KEY parameter 'keyTag' not specified";
	    } else {
		unless (defined(is_integer($ktag))) {
		    $valok = 0;
		    push @e, "keyTag must be an integer: '$ktag'";
		} else {
		    unless (defined(is_between($ktag, 0, 65535))) {
			$valok = 0;
			push @e, "KEY keyTag must be an integer in the range 0-65535: '$ktag'";
		    }
		}
	    }
	    return $valok, @e unless ($valok);

	    my $tag = "keyTag $ktag: ";

	    if (defined($key_el->{key_alg})) {
		# KEY record to validate
		my $key_alg      = $key_el->{key_alg};
		my $key_protocol = $key_el->{key_protocol};
		my $key_flags    = $key_el->{key_flags};
		my $key_pubKey   = $key_el->{key_pubKey};
		
		unless (defined($key_alg)) {
		    $valok = 0;
		    push @e, "$op: $tag KEY mandatory parameter 'alg' not specified";
		} else {
		    unless (defined(is_integer($key_alg))) {
			$valok = 0;
			push @e, "$tag KEY alg must be an integer: '$key_alg'";
		    } else {
			unless ($c->config->{dnssec}->{algorithms}->{$key_alg}) {
			    $valok = 0;
			    push @e, "$tag KEY alg is not among the permitted algorithm values: $key_alg, permitted are: " . 
				join(", ", sort( {$a <=> $b} keys(%{$c->config->{dnssec}->{algorithms}})));
			}
		    }
		}
		return $valok, @e unless ($valok);

		# protocol, must always be 3
		unless (defined($key_protocol)) {
		    $valok = 0;
		    push @e, "$op: $tag KEY mandatory parameter 'prot' (protocol) not specified";
		} else {
		    unless (defined(is_integer($key_protocol))) {
			$valok = 0;
			push @e, "$tag KEY prot must be an integer: '$key_protocol'";
		    } else {
			unless ($key_protocol == 3) {
			    $valok = 0;
			    push @e, "$tag KEY prot must always have the value '3' for EPP, is '$key_protocol'";
			}
		    }
		}
		return $valok, @e unless ($valok);

		# flags, must always be 257 (TBD if also 256)
		unless (defined($key_flags)) {
		    $valok = 0;
		    push @e, "$op: $tag KEY mandatory parameter 'flags' not specified";
		} else {
		    unless (defined(is_integer($key_flags))) {
			$valok = 0;
			push @e, "$tag KEY flags must be an integer: '$key_flags'";
		    } else {
			unless ($key_flags == 257) {
			    $valok = 0;
			    push @e, "$tag KEY flags must always have the value '257' for EPP, is '$key_flags'!";
			}
		    }
		}
		return $valok, @e unless ($valok);
		
		# pubKey:
		unless (defined($key_pubKey)) {
		    $valok = 0;
		    push @e, "$op: $tag KEY mandatory parameter 'pubKey' not specified";
		} else {
		    #print STDERR "VALIDATING: '$key_pubKey'\n\n";
		    my $dec = decode_base64($key_pubKey);
		    my $enc = encode_base64($dec, "");

		    #print STDERR " decoded: '$dec'\n\n";
		    #print STDERR " encoded: '$enc'\n\n";
		    unless ($key_pubKey eq $enc) {
			$valok = 0;
			push @e, "$tag KEY invalid pubKey, must be base64 numbers only";
		    }
		}

		# Finally, verify consistency between keys and DS.
		# For each key, check if a DS is specified for the same keytag, if so, 
		# the algorithms must match, if not, we reject.
		# Also, if a key is defined, and we have list of DSs, but the keyTag is not in the DS,
		# we have a key which will silently be discardedd by the client because it cannot be used,
		# so we warn about that as well.

		if (defined($ds) && $ds) {

		    #print STDERR "XXX Validate KEY/DS relation: ", Dumper($s), "vs.", Dumper($k), "\n----\n";

		    my ($ds_key_match, $ds_key_alg_match);

		    foreach my $ds_elem (keys %$ds) {
			
			my $ds_el = $ds->{$ds_elem};
			
			#print STDERR "keyTag: $ktag, Validate KEY/DS relation on DS: ", Dumper($ds_el), "\n";

			my $ds_ktag = $ds_el->{keyTag};
			if ($ktag == $ds_ktag) {
			    #print STDERR "$ktag MATCH: Validate KEY/DS relation: ", Dumper($ds_el), "vs.", Dumper($key_el), "\n-----\n";
			    ++$ds_key_match;
			    # OK, we have found matching keyTags, check that the alg's are identical
			    if ($key_alg == $ds_el->{alg}) {
				++$ds_key_alg_match;
			    }
			}
		    }
		    unless ($ds_key_match) {
			$valok = 0;
			push @e, "$tag KEY cannot be used because there are no matching DS record for tag: $ktag";
		    }
		    unless ($ds_key_alg_match) {
			$valok = 0;
			push @e, "$tag KEY cannot be used because there are no matching DS record for tag: $ktag and alg: $key_alg";
		    }
		}
	    }
	}

	if ($c->stash->{pd}->{nsecdns}) {
	    # Implies update operation
	    # An update operation has been trigged and a change in DNSSEC data has been detected
	    unless ( $c->stash->{pd}->{vfydnssecupdate} ) {
		$valok = 0;
		push @e, "Update of DS or key data requires a checked 'verify update' field";
	    }
	}
    }
   
    #$c->log->debug("Validate returns: valok: $valok, errs: " . join(", ", @e));
    return $valok, @e;
}

sub detect_and_stash_domain_object_changes: Private {
    my ( $self, $c ) = @_;
  
    # Update always requires a handle

    # check which parameters are changed by comparing to the session values.
    # If a scalar parameter 'xxx' is changed, mark it in pd as name 'nxxx' and 
    # the DRI will pick the 'nxxx'.

    my $update = 0;
 
    foreach my $e (@ds) {
	# skip all status elements because they are readonly
	next if ($e =~ m/^status/mx);

	# 'name' on session is always idn form. Compare the ace names
	# for 'name' in case the user uses ace-form in input field, in
	# which case we will detect different names even if the name
	# is the same
	$e = 'ace' if ($e eq 'name');
	
	my $o = $c->session->{pd}->{$e} || "";
	my $n = $c->stash->{pd}->{$e} || "";
	
	if (($o || $n) && ($n ne $o)) {
	    # A new value found in this field
	    $c->stash->{pd}->{"n$e" } = $n;
	    ++$update;
	    $c->log->debug("EppClient: action: UPDATE: new $e: '$o' -> '$n'");
	}
	# unset to name for domainname
	$e = 'name' if ($e eq 'ace');
    }

    if ($c->stash->{pd}->{applicantdataset}) {
	# A dataset value is specified, so detect any changes
	foreach my $e (@dsapplicantdataset) {
	    # $o and $n must be scalars, so we use concat to stringify the acceptdate from the date-object
	    my $o = "";
	    my $n = "";

	    if ($c->session->{pd}->{applicantdataset} &&
		$c->session->{pd}->{applicantdataset}->{$e}) {
		$o = "" . ($c->session->{pd}->{applicantdataset}->{$e});
	    }
	    if ($c->stash->{pd}->{applicantdataset} &&
		$c->stash->{pd}->{applicantdataset}->{$e}) {
		$n = "" . ($c->stash->{pd}->{applicantdataset}->{$e});
	    }
	    
	    if (($o || $n) && ($n ne $o)) {
		# A new value found in this field
		$c->stash->{pd}->{napplicantdataset}->{$e} = $n;
		++$update;
		
		# disable logging 
		#$c->log->debug("EppClient: action: UPDATE: new $e: '$o' -> '$n'");
	    }
	}
	# If napplicantdataset is populated, some change is detected.
	# However, we need to make sure all three values are populated because
	# the whole group is required by the schema and Net::DRI
	# so we need to also clone any unchanged values.
	if ($c->stash->{pd}->{napplicantdataset}) {
	    foreach my $e (@dsapplicantdataset) {
		unless ($c->stash->{pd}->{napplicantdataset}->{$e}) {
		    $c->stash->{pd}->{napplicantdataset}->{$e} = $c->stash->{pd}->{applicantdataset}->{$e} || "";
		}
	    }
	}
    }

    # All array changes need to be done as add/delete arrays.
    my $n = $c->stash->{pd}->{nsset};
    my $o = $c->session->{pd}->{nsset};

    #print STDERR "\n-- pd changes: session: ", Dumper $o, "\nnew pd", Dumper $n, "\n";

    if ($n || $o) {
	my ($ar, $dr) = NORID::EppClient::Controller::Util::make_add_del_list($n, $o);
	
	if ($ar && @$ar) {
	    $c->stash->{pd}->{"nnsset"}->{add} = $ar;
	    $c->log->debug("UPDATE add: nsset: " . (join ",", @$ar));
	    ++$update;
	}
	if ($dr && @$dr) {
	    $c->log->debug("UPDATE del: nsset: " . (join ",", @$dr));
	    $c->stash->{pd}->{"nnsset"}->{del} = $dr;
	    ++$update;
	}
	#print STDERR "\n-", Dumper  $c->stash->{pd}->{"nnsset"}, "\n";
	 

    }

    if ( $c->session->{showstatus} &&
	 $c->config->{client}->{domain_status_support}->{enabled}) {
    
	$n = $c->stash->{pd}->{statuslist};
	$o = $c->session->{pd}->{statuslist};

	#print STDERR "\n-- pd changes: session: ", Dumper $o, "\n new statuslist pd", Dumper $n, "\n";
	
	if ($n || $o) {
	    my ($ar, $dr) = NORID::EppClient::Controller::Util::make_add_del_list($n, $o);
	    
	    if ($ar && @$ar) {
		$c->stash->{pd}->{"nstatuslist"}->{add} = $ar;
		$c->log->debug("UPDATE add: nstatuslist: " . (join ",", @$ar));
		++$update;
	    }
	    if ($dr && @$dr) {
		my $dstatusl = join ",", @$dr;
		if ($dstatusl ne 'ok') {
		    $c->log->debug("UPDATE del: nstatuslist: $dstatusl");
		    $c->stash->{pd}->{"nstatuslist"}->{del} = $dr;
		    ++$update;
		}
	    }
	    #print STDERR "\nnstatuslist: ", Dumper  $c->stash->{pd}->{"nstatuslist"}, "\n";
	}
    }
    
    my $a = 'tech';
    $n = $c->stash->{pd}->{coset}->{$a};
    $o = $c->session->{pd}->{coset}->{$a};

    # print STDERR "\n-- $a: old pd: ", Dumper $o, "\nnew pd", Dumper $n, "\n";

    if ($n || $o) {
	my ($ar, $dr) = NORID::EppClient::Controller::Util::make_add_del_list($n, $o);
	
	if ($ar && @$ar) {
	    $c->stash->{pd}->{ncoset}->{add}->{$a} = $ar;
	    $c->log->debug("UPDATE add: coset $a: " . (join ",", @$ar));
	    ++$update;
	}
	if ($dr && @$dr) {
	    $c->log->debug("UPDATE del: coset $a: " . (join ",", @$dr));
	    $c->stash->{pd}->{ncoset}->{del}->{$a} = $dr;
	    ++$update;
	}
    }

    # secdns change detection
    if ($c->config->{dnssec}->{enabled}) {
	my $n = $c->stash->{pd}->{secdns_merged};
	my $o = $c->session->{pd}->{secdns_merged};
	#print STDERR " n: ", Dumper $n, "\n\n";
	#print STDERR " o: ", Dumper $o, "\n\n";

	my (@nl, @ol);
	
	if ($n || $o) {
	    # make a list of all old and new ds hashes, then compare those lists,
	    # the character '_' is used as separator, and is not allowed as a value,
	    # so should be safe.

	    foreach my $el (@{$n}) {
		#print STDERR " n el: ", Dumper $el, "\n";
		my $s = "";
		if ($el->{alg}) {
		    $s .= $el->{keyTag}     . "_";
		    $s .= $el->{alg}        . "_";
		    $s .= $el->{digestType} . "_";
		    $s .= $el->{digest};
		}
		if ($el->{key_alg}) {
		    $s .= '_';
		    $s .= $el->{key_keyTag}   . "_";
		    $s .= $el->{key_alg}      . "_";
		    $s .= $el->{key_flags}    . "_";
		    $s .= $el->{key_protocol} . "_";
		    $s .= $el->{key_pubKey};
		}
		push @nl, $s if ($s);
	    }
	    foreach my $el (@{$o}) {
		my $s = "";
		#print STDERR " o el: ", Dumper $el, "\n";
		if ($el->{alg}) {
		    $s .= $el->{keyTag}     . "_";
		    $s .= $el->{alg}        . "_";
		    $s .= $el->{digestType} . "_";
		    $s .= $el->{digest};
		}
		if ($el->{key_alg}) {
		    $s .= '_';
		    $s .= $el->{key_keyTag}   . "_";
		    $s .= $el->{key_alg}      . "_";
		    $s .= $el->{key_flags}    . "_";
		    $s .= $el->{key_protocol} . "_";
		    $s .= $el->{key_pubKey};

		}
		push @ol, $s if $s;
	    }
	    #print STDERR "nl: ", Dumper \@nl;
	    #print STDERR "ol: ", Dumper \@ol;
	    #print STDERR "detect changes: n: ", Dumper $n, "\n, o: ", Dumper $o, "\n";
	    my ($aar, $dar) = NORID::EppClient::Controller::Util::make_add_del_list(\@nl, \@ol);

	    #print STDERR "add: ", Dumper $aar;
	    #print STDERR "del: ", Dumper $dar;

	    # Remap the data to add/delete arrays
	    foreach my $op ('add', 'del') {
		my $ar = $aar;
		$ar = $dar if ($op eq 'del'); 
		
		if ($ar && @$ar) {
		    foreach my $el (@$ar) {
			my @d = split '_', $el;
			my %ds = ( 
			    keyTag     => $d[0],
			    alg        => $d[1],
			    digestType => $d[2],
			    digest     => $d[3]
			    );
			$ds{key_alg}      = $d[5] if ($d[5]);
			$ds{key_flags}    = $d[6] if ($d[6]);
			$ds{key_protocol} = $d[7] if ($d[7]);
			$ds{key_pubKey}   = $d[8] if ($d[8]);
			
			push @{$c->stash->{pd}->{nsecdns}->{$op}}, \%ds;
		    }
		    ++$update;
		}			
	    }
	    # Check if delete all is selected. If so, we replace any 'del' arrays with 'all'
	    if ($c->stash->{pd}->{secdns_delete_all}) {
		if (defined($c->stash->{pd}->{nsecdns}->{del}) && @{$c->stash->{pd}->{nsecdns}->{del}}) {
		    delete($c->stash->{pd}->{nsecdns}->{del});
		}
		$c->stash->{pd}->{nsecdns}->{del} = 'all';
		++$update;
	    }
	    #print STDERR "add and del hashes in nsecdns: ", Dumper $c->stash->{pd}->{nsecdns};
	}
    }
    #$c->log->debug("UPDATE: update: $update");
    return $update;
}

=head2 set_session_from_stash

Put the data from the stash into the session data so we can present 
the previous values in the next page.

=cut

sub set_session_from_stash: Private {
    my ( $self, $c ) = @_;

    $c->session->{pd} = dclone($c->stash->{pd});

    #print STDERR "PD : ", Dumper $c->stash->{pd}->{secdns};
    #print STDERR "SPD: ", Dumper $c->session->{pd};

}

=head2 set_stash_from_domain_object

We have contact object data in $s
Put the interesting data from it into the stash.

=cut

sub set_stash_from_domain_object: Private {
    my ( $self, $c, $s) = @_;

    #$c->forward('null_data');
    delete $c->stash->{pd};
    delete $c->session->{pd};

    #print STDERR "s: ", Dumper $s;

    foreach my $e (@dsall) {
	#print STDERR "stashing: $e?\n";
	#print STDERR "stashing: $e: ",  $s->{$e}, "\n" if $s->{$e};
	$c->stash->{pd}->{$e} = $s->{$e} if ($s->{$e});
    }
    if ($s->{applicantdataset}) {
	$c->stash->{pd}->{applicantdataset} = $s->{applicantdataset};
    }

    if ($s->{secdns}) {
	$c->stash->{pd}->{secdns} = $s->{secdns};
    }
    if ($s->{secdns_merged}) {
	$c->stash->{pd}->{secdns_merged} = $s->{secdns_merged};
    }


    foreach my $e (@da) {

	#print STDERR "$e\n";

	if ($e eq 'nsset' || $e eq 'statuslist') {
	    my @v = @{$s->{$e}} if (exists($s->{$e}) && $s->{$e} && @{$s->{$e}});
	    #print STDERR "stashing: $e: ",  @v, "\n" if (@v);
	    push @{$c->stash->{pd}->{$e}}, @v if @v;

	} else {
	    my @v = @{$s->{coset}->{$e}} if ($s->{coset}->{$e} && @{$s->{coset}->{$e}});
	    #print STDERR "stashing: $e: ",  @v, "\n" if (@v);
	    push @{$c->stash->{pd}->{coset}->{$e}}, @v if @v;
	}
    }
    foreach my $e (@dos) {
#	print STDERR "stashing: $e: ",  $s->{$e}, "\n" if $s->{$e};

	if ($s->{$e}) {
	    if ($e eq 'exDate') {
		my $ce = $s->{$e};
		if ($ce =~ m/^(\d\d\d\d-\d\d-\d\d).+$/mx) {
		    my $date = $1;
#		    print STDERR "date: $date\n";
		    $c->stash->{pd}->{curexpiry} = $date;
		}
	    }
	    $c->stash->{pd}->{$e} = $s->{$e};
	}
    }

    #print STDERR "pd: ", Dumper $c->stash->{pd};
    #print STDERR "pd : ", Dumper $c->stash->{pd}->{secdns};
    #print STDERR "pds: ", Dumper $c->session->{pd}->{secdns};
}


=head2 reset

Null all data and also control data.

Reset form to initialized state.

=cut

sub reset : Private {
    my ( $self, $c ) = @_;

    $c->forward('null_data');

    delete ($c->session->{objdetails});
    delete ($c->session->{cmddetails});
    delete ($c->session->{showstatus});
    delete ($c->session->{dnssecdetails});
    delete ($c->session->{transfercmds});
    delete ($c->session->{editcmds});
    delete ($c->session->{facetmenu});
    delete $c->stash->{facets};
    delete $c->session->{facets};

}

=head2 null_data

Zap session and stash, to display empty values.

=cut
sub null_data : Private {
    my ( $self, $c ) = @_;

    delete $c->stash->{pd};
    delete $c->session->{pd};
    delete $c->stash->{pdc};
    delete $c->session->{pdc};

}

=head2 index

Show the domain page.

=cut

sub index : Private {
    my ( $self, $c ) = @_;

    $c->forward('menu');
}


=head2 validate_and_create_csrf_token

CSRF validation.

=cut

sub validate_and_create_csrf_token: Private {
    my ( $self, $c) = @_;

    &NORID::EppClient::Controller::Util::validate_and_create_csrf_token($self, $c, 'd');

}


=begin unpack_dataset_file

The applicant data set file shall reside as a txt file inside the
zip file along with the pdf of the applicant declaration.

 - Unpack the zip
    - Two files should be contained, one .pdf and one .txt
    - If this is not the caser, abort.

 - Fetch the contents of the .txt file
   - parse it, if wrong syntax, abort
   - extract the applicant data set entries

 - Return the applicant dataset

=end unpack_dataset_file

=cut

sub unpack_dataset_file {
    my ( $zf ) = @_;
    
    #print STDERR "unpack: $zf\n";

    my $txtfile;
    my ($ds, $fs, @errs);

    unless ( -f $zf ) {
	push @errs, "No zip file";
	return 0, \@errs;
    }

    my $zip = Archive::Zip->new();

    # Set an empty ErrorHandler to avoid STDERR being 
    # polluted with error printouts from the module
    Archive::Zip::setErrorHandler( sub{} );

    my $status = $zip->read( $zf );

    if ($status != AZ_OK) {
	push @errs, "Read of zip file failed";
	if ($status == AZ_FORMAT_ERROR) {
	    push @errs, "Format error in zip file";
	}
	return 0, \@errs;
    }
    
    for my $ft ('pdf', 'txt') {
	my @filemembers = $zip->membersMatching( ".*\.$ft" );
	unless (@filemembers) {
	    push @errs, "No '*.$ft' file found inside zip file";
	    return 0, \@errs;
	}
	if (@filemembers > 1) {
	    push @errs, "More than one '*.$ft' file found inside zip file";
	    return 0, \@errs;
	}
	$txtfile = $filemembers[0];
    }
    my $txtfname = $txtfile->fileName;

    my @adc = split "\n", $zip->contents($txtfname);

    unless (@adc) {
	push @errs, "Cannot extract contents of $txtfname from zip file";
	return 0, \@errs;
    }
    
    foreach my $l (@adc) {

	next unless ($l);
	next if ($l =~ /\#/);

	$l = Encode::decode('UTF-8', $l);

	#print STDERR "l in: ", Encode::encode('UTF-8', $l), "\n";

	# Split on first colon (Limit=2):
	my ($h, $v) = split ':', $l, 2;

	# collapse all white spaces
	$h =~ s/\s+/ /g;
	$h =~ s/^\s+//g;
	$h =~ s/\s+$//g;
	$v =~ s/\s+/ /g;
	$v =~ s/^\s+//g;
	$v =~ s/\s+$//g;
	unless ($ds_dri_map{$h} || $ds_file_map{$h}) {
	    push @errs, "Invalid file key: '$h'";
	    return 0, \@errs;
	}
	#print STDERR "h: $h, v: ",  Encode::encode('UTF-8', $v), "\n";

	if ($ds_dri_map{$h}) {
	    # Map the field names from the file tags to the Net::DRI tags
	    $ds->{$ds_dri_map{$h}} = $v;

	} elsif ($ds_file_map{$h}) {

	    if ($h =~ m/Domain name/) {
		push @{$fs->{$ds_file_map{$h}}}, $v;
	    } elsif ($ds_file_map{$h}) {
		$fs->{$ds_file_map{$h}} = $v;
	    }
	}
    }
    return 1, undef, $ds, $fs;
}

=begin parse_dsset_key_file

Parse file

=end parse_dsset_key_file

=cut

sub parse_dsset_key_file {
    my ( $c, $upload ) = @_;

    my ($kd, @errs);

    my $fn = $upload->tempname;

    unless ( -f $fn ) {
	push @errs, "No file found";
	return 0, \@errs;
    }
    # Check that some kind of a txt or pgp file MIME-type has been loaded.
    # Only log the problem, as the parsing should trap real format problems anyway.
    unless ($upload->type =~ m/txt|pgp-keys|octet-stream/x ) {
	$c->log->debug("Key file is not a 'txt' or 'pgp-keys' or 'octet-stream' MIME type file: " . $upload->type);
    }

    # Parse the key file and parse it to get the various parameters
    my $size = $upload->size;
    if ($size > $MAX_DSSET_KEY_SIZE) {
	push @errs,  "File size too big, probably not a proper DS or key file. Size is $size, whilst maximum size accepted is $MAX_DSSET_KEY_SIZE.";
	return 0, \@errs;
    }

    my $dkfn = $upload->filename;
    my $dkdt = $upload->slurp;

    # strip any \r
    $dkdt =~ s/\r*//g;
    my @lines = split "\n", $dkdt;

    unless (@lines >= 1) {
	push @errs,  "No lines in file.";
	return 0, \@errs;
    }

   if ($dkdt =~ m/\sDS\s/mx) {
	# dsset file and syntax expected
       return parse_dsset_file($c, $upload, \@lines);

   } elsif ($dkdt =~ m/\sDNSKEY\s/mx) {
	# key file and syntax expected
	return parse_key_file($c, $upload, \@lines);

   } else {
       # illegal stuff, reject
	push @errs,  "Illegal file, no DS or DNSKEY data found.";
	return 0, \@errs;
   }

}

=begin parse_dsset_file

A dsset-file is expected to ontain at least one DS entry for each public KSK .key file.

Expected syntax:
 <domainname>            IN DS <keyTag> <alg> <digestType> <digest>

Example with two DS records, one calculated with digestType=1 (SHA1), 
the other with digestType=2 (SHA256), and this is the default output from dnssec-signzone -S.

File name:
dsset-example.com.

Contents:
example.com.            IN DS 12109 8 1 4138F324BC2DBE11E200C6F712E1DEB1ED11157E
example.com.            IN DS 12109 8 2 75800089C6AE280ADE5E89B5EDAFDD3EA07A4E41AD715D06F666C41F C83FD2A6

=end parse_dsset_file

=cut

sub parse_dsset_file {
    my ( $c, $upload, $dl ) = @_;

    my ($kd, @errs);

    my $fn = $upload->filename;
    my $df_domname;

    if ($fn =~ m/^dsset-(.+\..+)\.$/x) {
	# Parse file name as well, since it starts with dsset-.
	# Expected syntax is then "dsset-example.no."
	#print STDERR "dsset with domain name '$1' is loaded\n";
	$df_domname = $1;
    }

    return parse_dsset_text($c, $dl, $df_domname);

}

=begin parse_key_file

Key file is optional, but may be added for each DS in a dsset file.

If added, the keytag/keyId will identify it.

To be able to parse and extract the keyTag properly, the key file name and 
contents MUST be according to a standard DNS KEY file as made by dnssec-keygen.

- The keyid must be found in the filename, as well in a comment line inside the file.
- The key must be the KSK-key, which means the flag must always be 257 (@@TBC if ever 256??)

File name:
  Kexample.com.+003+41209.key

Contents:
; This is a key-signing key, keyid 41209, for example.com.
; Created: Fri Aug  9 09:52:38 2013
; Publish: Fri Aug  9 09:52:38 2013
; Activate: Fri Aug  9 09:52:38 2013
example.com. IN DNSKEY 257 3 3 BMQV+cK4gPgJg5ZwFkLusLXWffjXqTFKslbAC+TfEiVSUK9TkEU3ocoC 0A6u9dXuGO/U0dhBVitZsQp2ybYHf/1vXxnZFldO/CXTEQ2pav0aBaTn 7jMpANSiq46PhVtwAkNQyHHAsjHpMCPUF4U62MCGZ8UNA56cBQE37OOO 1VwNeOmLOhmgJrk5S4QFiLy6LvgiU/NP/QeX67osj/zef2pIlqoiFPhz F+yIzL43WEGEsjz1LZqOBm9PDhx3s6hRKGl7fMf3z7zVqwYOp/kHlgLC SpBuJfdYsZphRIeK/tx97HuCXj8pSCSLK+CHFyHvtfApNiLpZQ4aQNK2 Y8e4ibKrsdgjZNJVGbLCdoFaxE+IKCUc2Jqj6SjUHJh0LyUBdvb3+7BX kp7cUKQ6I1kduK45DUzK

=end parse_key_file

=cut

sub parse_key_file {
    my ( $c, $upload, $dl ) = @_;

    my ($kd, @errs);

    # Parse original file name, syntax "Kexample.com.+003+14527.key"
    #  1: K
    #  2: domain name
    #  3: +
    #  4: algorithm
    #  5: +
    #  6: keyId
    #  7: key or private (should never be private or other)
    #
    my $fn = $upload->filename;

    my ($kf_domname, $kf_alg, $kf_keyid);

    if ($fn =~ m/^K(.+\..+)\.\+(\d+)\+(\d+)\.(.+)$/x) {
	#print STDERR "f1: $1\n";
	#print STDERR "f2: $2\n";
	#print STDERR "f3: $3\n";
	#print STDERR "f4: $4\n";

	if ($4 ne 'key') {
	    push @errs,  "key file ending should be '.key', but is '$4'. Do NOT load private or other keys!";
	    return 0, \@errs;
	}
	$kf_domname = $1;

	# Use += to remove any leading zeroes in the alg and keyid if file name is like 'Kexample.no.+003+00209.key'
	$kf_alg    += $2;
	$kf_keyid  += $3; 

    } else {
	push @errs,  "Key file name syntax error, syntax must be: 'Kexample.com.+003+41209.key'";
	return 0, \@errs;
    }

    return parse_key_text($c, $dl, $kf_domname, $kf_alg, $kf_keyid);
}

=head1 parse_dsset_key_text 

Parse text, DS or DNSKEY data shall be found

=cut

sub parse_dsset_key_text {
    my ( $c, $d ) = @_;

    my @errs;

    # strip any \r
    $d =~ s/\r*//g;
    my @lines = split "\n", $d;
    unless (@lines >= 1) {
	push @errs,  "Illegal input, no lines in DS/DNSKEY text.";
	return 0, \@errs;
    }

   if ($d =~ m/\sDS\s/mx) {
	# dsset file and syntax expected
       return parse_dsset_text($c, \@lines);

   } elsif ($d =~ m/\sDNSKEY\s/mx) {
	# key file and syntax expected
	return parse_key_text($c, \@lines);

   } else {
       # illegal stuff, reject
	push @errs,  "Illegal input, no DS/DNSKEY data found.";
	return 0, \@errs;
   }

}

=begin parse_dsset_text

A dsset text is expected to ontain at least one DS entry.
The file itself may be any text file, and contains one or more DS entries.


Expected syntax:
 <domainname>            IN DS <keyTag> <alg> <digestType> <digest>

Example with two DS records, one calculated with digestType=1 (SHA1), 
the other with digestType=2 (SHA256), and this is the default output from dnssec-signzone -S.

File name:
dsset-example.no.

Contents:
example.no.             IN DS 681 10 1 00D59FA4454D6F7D68B7D9DBAD7767265CCFE0A2
example.no.             IN DS 681 10 2 D7016B9C6F2F35EFE7F1CD1FC046E863363C1C3CB8706D57491AD19E 5932CF83

---

Another example where the dsset file conains several domains, which may come from OpenDNDDSEC.
In that case the file name should not reflect s specific domain name, but have some other name, like ds.txt:

File name:
ds.txt

Contents:
example.no.             IN DS 681 10 1 00D59FA4454D6F7D68B7D9DBAD7767265CCFE0A2
example.no.             IN DS 681 10 2 D7016B9C6F2F35EFE7F1CD1FC046E863363C1C3CB8706D57491AD19E 5932CF83
example.no.             IN DS 48321 10 1 C1EE5292D263C2797FD2722FDABA2F3234DDED36
example.no.             IN DS 48321 10 2 B784E6638D3C832CACB1874FF35CD6BB9247FCEEBC658CA49708D469 1C54AD95
thaugen.no.             IN DS 7726 8 1 FC1ADFA40155B948436C58FD057CC4DD2ACE11C2
thaugen.no.             IN DS 7726 8 2 3050E258B6D923B0246A1DD752A239648DCE192F49269E0145D28EB3 33801F5A
thaugen.no.             IN DS 12859 8 1 E31F024D5A3E8FCEA7E8F0582C9F70D96D4CC73F
thaugen.no.             IN DS 12859 8 2 914E8E7D838A4E76D7ADC522C63C52F919E5F114D07891EE4389D041 CD8E1219
thaugen.no.             IN DS 48105 3 1 643366C7DC773F9C1C058908B750CBA33F9A834D
thaugen.no.             IN DS 48105 3 2 7020D126D774973FB388FC4FAEFBA8BBC345D6C5C9F81FA74C9C5A85 035F66B3

=end parse_dsset_file

=cut

sub parse_dsset_text {
    my ($c, $dl, $df_domname) = @_;

    my ($kd, @errs);

    ##
    # Parse the DS-set text
    #
    # 1: domain name      $1  (shall match $df_domname if set)
    # 2: IN 
    # 3: DS
    # 4: keyTag (keyid)   $2
    # 5: alg              $3
    # 6: digestType       $4
    # 7: digest           $5
    # Multiple DS records are allowed for a domain, typically using different digestTypes

    my $found=0;
    foreach my $l (@$dl) {
	#print STDERR "Parse dsset line: '$l'\n";

	if ($l =~ m/(^.+\..+)\.\s+IN\s+DS\s+(\d+)\s+(\d+)\s+(\d+)\s+(.+)$/x) {
	    #print STDERR "1 domainname: $1\n";
	    #print STDERR "4 keyTag    : $2\n";
	    #print STDERR "5 alg       : $3\n";
	    #print STDERR "6 digestType: $4\n";
	    #print STDERR "7 digest    : $5\n";
	    
	    # Some basic validation
	    if ($df_domname && $1 ne $df_domname) {
		push @errs,  "domain name in dsset file name '$df_domname' does not match domain name for the DS record inside the file '$1'!\n";
		return 0, \@errs;
	    }
	    # Some basic validation
	    if ($df_domname && $1 ne $df_domname) {
		push @errs,  "domain name in dsset file name '$df_domname' does not match domain name for the DS record inside the file '$1'!\n";
		return 0, \@errs;
	    }
	    if (!$df_domname && $c->stash->{pd}->{ace} && $1 ne $c->stash->{pd}->{ace}) {
		# The dsset file does not contain a domain name, thus multiple domain DS records may 
		# be present. Discard those not defined for the ace name we are currently handling.
		next;
	    }

	    my $k;
	    $k->{keyTag}     = $2;
	    $k->{alg}        = $3;
	    $k->{digestType} = $4;
	    $k->{digest}     = $5;

	    # Multiple DS records for each keyTag is possible, so store in an array
	    # Could check for uniqueness as well, but those are stripped by the client anyway
	    $kd->{keyids}->{domainname} = $1;
	    my $a = $2 . '-' . $3 . '-' . $4;
	    %{$kd->{keyids}->{ds}->{$a}} = %{$k};
	    ++$found;
	}
    }
    #print STDERR "Loaded dsset info: ", Dumper $kd;

    unless ($found) {
	push @errs,  "No DS records found!";
	return 0, \@errs;
    }

    return 1, undef, $kd;
}


=begin parse_key_text

Expected key text syntax, possibly just pasted from a key file.
If a comment is present, it is parsed as well.

; This is a key-signing key, keyid 41209, for example.com.
; Created: Fri Aug  9 09:52:38 2013
; Publish: Fri Aug  9 09:52:38 2013
; Activate: Fri Aug  9 09:52:38 2013
example.com. IN DNSKEY 257 3 3 BMQV+cK4gPgJg5ZwFkLusLXWffjXqTFKslbAC+TfEiVSUK9TkEU3ocoC 0A6u9dXuGO/U0dhBVitZsQp2ybYHf/1vXxnZFldO/CXTEQ2pav0aBaTn 7jMpANSiq46PhVtwAkNQyHHAsjHpMCPUF4U62MCGZ8UNA56cBQE37OOO 1VwNeOmLOhmgJrk5S4QFiLy6LvgiU/NP/QeX67osj/zef2pIlqoiFPhz F+yIzL43WEGEsjz1LZqOBm9PDhx3s6hRKGl7fMf3z7zVqwYOp/kHlgLC SpBuJfdYsZphRIeK/tx97HuCXj8pSCSLK+CHFyHvtfApNiLpZQ4aQNK2 Y8e4ibKrsdgjZNJVGbLCdoFaxE+IKCUc2Jqj6SjUHJh0LyUBdvb3+7BX kp7cUKQ6I1kduK45DUzK

=end parse_key_file

=cut

sub parse_key_text {
    my ( $c, $dl,  $kf_domname, $kf_alg, $kf_keyid) = @_;
    
    my ($kd, @errs, $kc_keyid, $kc_domname );

    ##
    # Syntax:
    #
    # 1: domain name                      $1
    # 2: IN 
    # 3: DNSKEY
    # 4: flag bits (always 257 for KSK)   $2
    # 5: protocol, always 3               $3
    # 6: key algorithm                    $4
    # 7: 
    # Only one DNSKEY line is allowed for a key

    my $found = 0;
    foreach my $l (@$dl) {
	#print STDERR "Parse key text line '$l'\n";

	if ($l =~ m/; This is a key-signing key, keyid (\d+), for (.+\..+)\./m) {
 	    #print STDERR "This is keyid: $1 for domain: $2\n";

	    if ($kf_keyid && $1 ne $kf_keyid) {
		push @errs,  "Key file comment says keyid '$1', but it does not match keyid '$kf_keyid' in file name\n";
		return 0, \@errs;
	    }
	    if ($kf_domname && $2 ne $kf_domname) {
		push @errs,  "Key file comment says domain name '$2', but it does not match domain name '$kf_domname' in file name\n";
		return 0, \@errs;
	    }
	    $kc_keyid   = $1 || $kf_keyid;
	    $kc_domname = $2 || $kf_domname;

	} elsif ($l =~ m/(^.+\..+)\.\s+IN\s+DNSKEY\s+(\d+)\s+(\d+)\s+(\d+)\s+(.+)$/x) {
	    #print STDERR "1: $1\n";
	    #print STDERR "2: $2\n";
	    #print STDERR "3: $3\n";
	    #print STDERR "4: $4\n";
	    #print STDERR "5: $5\n";
	    
	    # Some basic validation
	    if ($1 && $kf_domname && $1 ne $kf_domname) {
		push @errs,  "Domain name from file name '$kf_domname' does not match domain name inside the key file '$1'!\n";
		return 0, \@errs;
	    }
	    if ($4 && $kf_alg && $4 != $kf_alg) {
		push @errs,  "Algorithm from file name '$kf_alg' does not match algorithm in the key file '$4'!\n";
		return 0, \@errs;
	    }
	    unless ($2 == 257 ) {
		push @errs,  "Key flags is '$2', must always be '257' for a KSK key (SEP bit must be set for a trusted anchor)!\n";
		return 0, \@errs;
	    }
	    if ($3 != 3) {
		push @errs,  "Protocol is '$3', must always be '3' for DNSSEC!\n";
		return 0, \@errs;
	    }
	    unless (exists($c->config->{dnssec}->{algorithms}->{$4}) && defined($c->config->{dnssec}->{algorithms}->{$4})) {
		push @errs,  "Unsupported key algorithm: $4!\n";
		return 0, \@errs;
	    }

	    my ($k, $keyid);
	    if ($kc_keyid || $kf_keyid) {
		$keyid = $kf_keyid || $kc_keyid;
	    } else {
		push @errs,  "KeyTag could not be determined, set to '99999', please specify manually!\n";
		$keyid = '99999';
	    }
	    $k->{key_keyTag}   = $keyid;
	    $k->{key_flags}    = $2;
	    $k->{key_protocol} = $3;
	    $k->{key_alg}      = $4;
	    $k->{key_pubKey}   = $5;

	    $kd->{keyids}->{domainname} = $1;

	    my $a = $keyid . '-' . $4 . '-' . $2;
	    %{$kd->{keyids}->{key}->{$a}} = %{$k};

	    ++$found;
	}
    }
    unless ($found) {
	push @errs,  "No DNSKEY records found!";
	return 0, \@errs;
    }
    if ($found > 1) {
	push @errs,  "More than one DNSKEY records found!";
	return 0, \@errs;
    }
    #print STDERR "Loaded key info: ", Dumper $kd;
    return 1, \@errs, $kd;

}


=head1 AUTHOR

Trond Haugen

=head1 LICENSE

This library is free software, you can redistribute it and/or modify
it under the same terms as Perl itself.

=cut

1;
