package NORID::EppClient::Controller::DomainBulk;

use strict;
use warnings;

use Storable qw(dclone);
use Net::DRI;
use NORID::EppClient::Model::NORID::DRI;
use NORID::EppClient::Controller::Util;
use base qw(Catalyst::Controller::RequestToken);

# For Debug purposes:
use Data::Dumper;
$Data::Dumper::Indent=1;

my $TT_DOMAIN = 'domainbulk.tt2';
my $TT_LOGIN = 'login.tt2';

my $MIN_NAMESERVERS = 2;

=pod

=head1 NAME

EppClient - Domain object operations

=head1 DESCRIPTION

The module supports creation and maintainance of domain objects for .NO. 

=head1 SYNOPSIS

=head1 DESCRIPTION

package NORID::EppClient::Controller::Domain - Catalyst Contact Controller

=head1 METHODS

=cut

my @allpars = ('bdomainnames', 'bnsset', 'btech', 'bpw', 'bduration', 'bdeletefromdns', 'bdeletefromregistry');

my @vfypars = ('bvfytransferwauthinfo', 'bvfyrenew', 'bvfywithdraw', 'bvfydelete', 'bvfyremalldnssec');


=head2 menu 

menu method.

=cut

sub menu : Local {
    my ( $self, $c ) = @_;

    # Present default bulk values as the last info from the domain tab
    # However, if any bulk values do exist, the tt2 will present them instead
    $c->stash->{pd} = $c->session->{pd} if ($c->session->{pd} && !defined($c->stash->{pd}));

    # Recreate token when menu is pressed, this will cleanup any errors
    $c->session->{'db_token'} = $self->create_token($c);

    $c->stash->{template} = $TT_DOMAIN;
}


=head2 action

action method.

This method catches all the EPP contact actions coming from the contact form 
submission and dispatches the action to the real action method.

=cut

sub action : Local {
    my ( $self, $c) = @_;
    
    $c->stash->{template} = $TT_DOMAIN;

    $c->forward('validate_and_create_csrf_token');

    my $set = $c->forward('set_stash_from_req_param');
    #$c->log->debug("set_stash_from_req_param returned: $set");
    unless ($set) {
	push @{$c->stash->{opsummary}}, "Nothing to do";
	$c->session->{pb} = dclone($c->stash->{pb});
	return;
    }

    my $op = $c->stash->{op};
    my $obj = $c->stash->{obj};
    my $cmd  = $c->stash->{cmd};

    if ($op) {
	if ($op eq 'reset') {
	    $c->forward('reset');
	} else {
	    ####
	    # All parameters are now stashed. Validate them locally from the stash
	    # before any DRI stuff is called.
	    my ($valok, @verr) = validate_from_stash ($c);
	    unless ($valok) {
		#$c->log->debug('Validate failed!');
		# $s should contain an array with some validation error descriptions
		push @{$c->stash->{opsummary}}, @verr;
		$c->stash->{cmd_show_status} = 1;
		$c->stash->{cmd_success} = $valok;
		$c->session->{pb} = dclone($c->stash->{pb});
		return;
	    }
	    #$c->log->debug("Validated OK, valok: $valok !");

	    ###
	    # Syntax OK, do the bulk stuff

	    # facets
	    if ($c->stash->{facets} && $c->stash->{facets}->{vfyfacets} && $c->stash->{facets}->{eppf}) {
		$c->stash->{p}->{facets} = $c->stash->{facets}->{eppf};
	    }

	    # A summary list separator
	    my $SEPARATOR = "------";

	    my $totnames = scalar(@{$c->stash->{pb}->{bdomainnames}});
	    my ($totok, $toterr, $totskipped);


	    # Time the total bukoperation
	    my $t1 = time;
    
	    # Perform a single login so we avoid login/logout for every domain
	    unless ($c->forward('/login/_login_is_ok')) {
		$c->log->info('EppClient: login before EPP operation failed!');
		# If either of above don't work out, send to the login page
		$c->stash->{template} = $TT_LOGIN;
		return;
	    }
	    my $cnt = 1;
	    my $doms = @{$c->stash->{pb}->{bdomainnames}};
	    foreach my $name (@{$c->stash->{pb}->{bdomainnames}}) {

		#$name = &NORID::EppClient::Controller::Util::encode_unicode($name);
		###
		# Setup the domain-params
		$c->req->parameters->{domainop} = 'info';
		$c->req->parameters->{domainname} = $name;

		# perform the info operation
		my $itxt = "Bulk operation start: info on $cnt/$doms, domain: $name..";
		$c->log->info($itxt);
		#print STDERR "$itxt\n";

		$c->forward('/domain/action', [ 1 ]);
		$c->stash->{template} = $TT_DOMAIN;

		# Check that the info was successful
		unless ($c->stash->{cmd_success}) {
		    ++$toterr;
		    $c->log->info($itxt . "FAILED");
		    #print STDERR ($itxt . "FAILED\n");

		    # check if we get a problem due to session limt exceeded, in which case we reconnect and 
		    # try again
		    if ($c->stash->{rcres} && $c->stash->{rcres}->{Message}) {
			if ($c->stash->{rcres}->{Message} =~ 
			    m/ERROR 2502 Session limit exceeded; server closing connection/) {

			    # Perform a new login to reconnect
			    unless ($c->forward('/login/_login_is_ok')) {
				$c->log->info('EppClient: login again due to session limit exceeded FAILED!');
				# If either of above don't work out, send to the login page
				$c->stash->{template} = $TT_LOGIN;
				return;
			    }
			    $c->log->info("EppClient: info failed, login again due to session limit exceeded!");
			    
			    # the domain in question should be reprocessed, push it back on top of the array
			    push @{$c->stash->{opsummary}}, "Session limit exceeded, reconnect and reprocess $name";
			    $c->log->info("EppClient: reprocess domain info on $name due to session limit exceeded!");
			    unshift @{$c->stash->{pb}->{bdomainnames}}, $name;
			    --$toterr;
			}
		    } else {
			push @{$c->stash->{opsummary}}, "Processing $name failed.. ", $SEPARATOR;
		    }
		    #print STDERR "rcres error for info $name: ", Dumper $c->stash->{rcres}, "\n";
		    next;
		}

		# Now build the domain_update operation for this domain
		$c->req->parameters->{domainop} = 'update';
		$c->req->parameters->{domainname} = $name;

		if ($op eq 'setns' || $op eq 'setall') {
		    # insert ns in the pd nsset and perform a domain update
		    $c->log->info("Bulk nsset IN!");
		    # Synthetisize a domain update
		    delete $c->stash->{pd}->{nsset};
		    delete ($c->req->parameters->{nsset}); 
		    push @{$c->req->parameters->{nsset}}, @{$c->stash->{pb}->{bnsset}};

		    # print STDERR "bulk pd nsset: ", Dumper $c->req->parameters, "\n";

		} 
		if ($op eq 'settech' || $op eq 'setall') {
		    # insert techc in the pd nsset and perform a domain update
		    delete $c->stash->{pd}->{coset}->{tech};
		    delete ($c->req->parameters->{tech}); 
		    push @{$c->req->parameters->{tech}}, @{$c->stash->{pb}->{"btech"}};
		    #print STDERR "bulk pd $op: ", Dumper $c->req->parameters, "\n";
		}

		if ($op eq 'setauthinfo') {
		    delete $c->stash->{pd}->{pw};
		    $c->req->parameters->{pw}                   = $c->stash->{pb}->{bpw};

		} elsif ($op eq 'removednssec') {

		    #print STDERR "rem DNSSEC bulk, op: $op, info found: ", Dumper @{$c->stash->{pd}->{secdns_merged}}->[0], "\n";
		    
		    # Only perform remove DNSSEC data is registered,
		    if ( $c->stash->{pd}->{secdns_merged} && @{$c->stash->{pd}->{secdns_merged}} ) {
			$c->req->parameters->{dnssecdeleteall} = $c->stash->{pb}->{bvfyremalldnssec};
			$c->req->parameters->{vfydnssecupdate} = $c->stash->{pb}->{bvfyremalldnssec};
			# We need to specify at least a keyTag to trig the dnssec funstions in Domain.pm
			$c->req->parameters->{keyTag}          = $c->stash->{pd}->{secdns_merged}[0]->{keyTag};

		    } else {
			# tell user that no dnssec data existed
			push @{$c->stash->{opsummary}}, 'No DNSSEC data to remove';
		    }

		} elsif ($op eq 'transfer_wauthinfo') {
		    $c->req->parameters->{domainop}             = $op;
		    $c->req->parameters->{pw}                   = $c->stash->{pb}->{bpw};
		    $c->req->parameters->{duration}             = $c->stash->{pb}->{bduration};
		    $c->req->parameters->{vfyrenew}             = $c->stash->{pb}->{bvfyrenew};
		    $c->req->parameters->{vfytransferwauthinfo} = $c->stash->{pb}->{bvfytransferwauthinfo};

		} elsif ($op eq 'renew') {
		    $c->req->parameters->{domainop} = $op;
		    $c->req->parameters->{duration} = $c->stash->{pb}->{bduration};
		    $c->req->parameters->{vfyrenew} = $c->stash->{pb}->{bvfyrenew};

		} elsif ($op eq 'withdraw') {
		    $c->req->parameters->{domainop}    = $op;
		    $c->req->parameters->{vfywithdraw} = $c->stash->{pb}->{bvfywithdraw};

		} elsif ($op eq 'delete') {
		    $c->req->parameters->{domainop}           = $op;
		    $c->req->parameters->{deletefromdns}      = $c->stash->{pb}->{bdeletefromdns};
		    $c->req->parameters->{deletefromregistry} = $c->stash->{pb}->{bdeletefromregistry};
		    $c->req->parameters->{vfydelete}          = $c->stash->{pb}->{bvfydelete};
		}

		$c->forward('/domain/action', [ 1 ]);
		$c->stash->{template} = $TT_DOMAIN;

		# all results are pushed by /domain - operation
	       
		# Include clTRID and svTRID for reference purposes per operation
		my $rcres = $c->stash->{rcres};
		push @{$c->stash->{opsummary}}, 'clTRID: ' . $rcres->{clTRID} if ($rcres->{clTRID});
		push @{$c->stash->{opsummary}}, 'svTRID: ' . $rcres->{svTRID} if ($rcres->{svTRID});

		# To get a better formatted result block, insert a summary list separator
		push @{$c->stash->{opsummary}}, $SEPARATOR;

		my $ltxt = "Bulk operation end: $op on  $cnt/$doms, domain: $name, result: ";

		if ($c->stash->{cmd_success}) {
		    ++$totok;
		    $c->log->info($ltxt . "SUCCESS");
		    #print STDERR ($ltxt . "SUCCESS\n");
		} else {
		    ++$toterr;
		    $c->log->info($ltxt . "FAILED");
		    #print STDERR ($ltxt . "FAILED\n");

		    # check if we get a problem due to session limt exceeded, in which case we reconnect and 
		    # try again
		    if ($c->stash->{rcres} && $c->stash->{rcres}->{Message}) {
			
			if ($c->stash->{rcres}->{Message} =~ 
			    m/ERROR 2502 Session limit exceeded; server closing connection/) {

			    # Perform a new login to reconnect
			    unless ($c->forward('/login/_login_is_ok')) {
				$c->log->info('EppClient: login again due to session limit exceeded FAILED!');
				# If either of above don't work out, send to the login page
				$c->stash->{template} = $TT_LOGIN;
				return;
			    }
			    $c->log->info("EppClient: $op failed, login again due to session limit exceeded!");
			    
			    # the domain in question should be reprocessed, push it back on top of the array
			    push @{$c->stash->{opsummary}}, "Session limit exceeded, reconnect and reprocess $name";
			    $c->log->info("EppClient: reprocess domain $op on $name due to session limit exceeded!");
			    unshift @{$c->stash->{pb}->{bdomainnames}}, $name;
			    --$toterr;
			}
		    }
		    #print STDERR "rcres error for $name: ", Dumper $c->stash->{rcres}, "\n";
		}
		++$cnt;

	    }
	    if ($toterr) {
		$c->stash->{cmd_success} = 0;
	    }
	    if ($totok && $totok <  $totnames) {
		# Some ok, some skipped or error, present yellow
		$c->stash->{cmd_pending} = 1;
	    }
	    my $t2 = time;

	    push @{$c->stash->{opsummary}}, "In total $totnames bulk domains processed, final result: ", 
	    " - " . ($totskipped or '0') . " skipped",
	    " - " . ($totok or '0') . " succeeded",
	    " - " . ($toterr or '0') . " failed",
	    "Time used: ". ($t2 - $t1) . " seconds.";
	}
	$c->session->{pb} = dclone($c->stash->{pb});
    } else {
	$c->forward('menu');
    }
}


=head2 set_stash_from_req_param

Map map the request parameters to the stash.
Also do the detection of what needs to be updated and set the stash 
according to it.

=cut

sub set_stash_from_req_param: Private {
    my ( $self, $c) = @_;

    my $ret = 1;

    my $op = lc($c->req->param('bdomainop') || "") || lc($c->req->param('reset') || "");

    $c->stash->{op} = $op;
    $c->stash->{cmd} = $op;
    $c->stash->{obj} = 'domain';

    $c->log->debug("op: $op");

    ####
    # Pick the parameters from the request form

    # map all single scalar form parameters
    # srid is the lookup key which is used for everything but the create
    # but we don't use it for a create.
    # trim trimmable values as they are processed.

    foreach my $e (@allpars) {
	if (my $v = $c->req->param($e)) {

	    $v = lc($v) unless ($e eq 'bpw');;

	    if ($e eq 'btech') {
		$v = uc($v);

	    } elsif ($e eq 'bduration' || $e eq 'bdeletefromdns' || $e eq 'bdeletefromregistry' || $e eq 'bpw') {
		$c->stash->{pb}->{$e} = $v;
		next;
	    } 

	    $v = NORID::EppClient::Controller::Util::trim($v);
	
	    # Translate each textarea string into an array
	    # wash the string and remove commas and multiple white spaces
	    $v =~ s/,/ /mxg;

	    # change the parameter into an array and wash away identical names
	    if (my @va = sort(NORID::EppClient::Controller::Util::wash_array(split " ", $v))) {
		# rebuild the value, now as a proper array
		delete $c->stash->{pb}->{$e};
		push @{$c->stash->{pb}->{$e}}, @va;
	    } 
	}
    }
    foreach my $e (@vfypars) {
	if (my $v = $c->req->param($e)) {
	    $c->stash->{pb}->{$e} = $v;
	}
    }

    # Store any facet values
    $c->session->{facetmenu}     = $c->req->param('facetmenu');
    $c->session->{cmddetails}    = $c->req->param('cmddetails');
    &NORID::EppClient::Controller::Util::set_stash_facets_from_req_param($c);

    #print STDERR "ret: $ret, pb: ", Dumper $c->stash->{pb};
    return $ret;
}


=head2 validate_from_stash

Validate the parameters needed for each operation.

Some parameters are also checked locally by the Domain module or the DRI lib.

Finally, the registry server will validate also.

=cut

sub validate_from_stash {
    my ( $c ) = @_;

    my $op   = $c->stash->{op};
    my $nd = @{$c->stash->{pb}->{bdomainnames}} if ($c->stash->{pb}->{bdomainnames});

    my $valok = 1;
    my @e;

    unless ($nd && $nd > 0) {
	$valok = 0;
	push @e, "'$op' requires at least one domain name";
    }

    foreach my $d ( @{$c->stash->{pb}->{bdomainnames}}) {
	unless (NORID::EppClient::Controller::Util::is_a_valid_domainname($d)) {
	    $valok = 0;
	    push @e, "$op: illegal domainname: '$d'";
	}
    }

    foreach my $h ( @{$c->stash->{pb}->{bnsset}}) {
	unless (NORID::EppClient::Controller::Util::is_a_valid_hostname($h)) {
	    $valok = 0;
	    push @e, "$op: illegal host name: '$h'";
	}
    }

    if ($op =~ m/^set(Ns|Tech|All)$/mxi) {
	if (($op eq 'setns' || $op eq 'setall') && !($c->stash->{pb}->{bnsset} && @{$c->stash->{pb}->{bnsset}})) {
	    $valok = 0;
	    push @e, "$op: SetNs or SetAll, but no name servers value(s) specified";
	}
	if ($op eq 'settech' || $op eq 'setall') {
	    unless ($c->stash->{pb}->{btech} && @{$c->stash->{pb}->{btech}}) {
		$valok = 0;
		push @e, "$op: SetTech or SetAll, but no Tech contact(s) specified";
	    }
	    foreach my $ch ( @{$c->stash->{pb}->{btech}}) {
		unless (NORID::EppClient::Controller::Util::isRoleID($ch)) {
		    $valok = 0;
		    push @e, "'$op': Invalid btech contact ID: $ch, must be a role";
		}
	    }	    
	}
    }

    if ($op eq 'removednssec') {
	unless ( $c->stash->{pb}->{bvfyremalldnssec} ) {
	    $valok = 0;
	    push @e, "Removal of all DNSSEC data requires a checked 'verify' field";
	}
    }

    # validation of remaining operations and parameters are done by the Domain-module
    $c->log->debug("Validate returns: valok: $valok, errs: " . join(", ", @e));
    return $valok, @e;
}




=head2 reset

Null all data and also control data.

Reset form to initialized state.

=cut

sub reset : Private {
    my ( $self, $c ) = @_;

    delete $c->stash->{pb};
    delete ($c->session->{cmddetails});
    delete $c->session->{pb};
    delete ($c->session->{facetmenu});
    delete $c->stash->{facets};
    delete $c->session->{facets};
}


=head2 index

Show the domain page.

=cut

sub index : Private {
    my ( $self, $c ) = @_;

    $c->forward('menu');
}

=head2 validate_and_create_csrf_token

CSRF validation.

=cut

sub validate_and_create_csrf_token: Private {
    my ( $self, $c) = @_;

    &NORID::EppClient::Controller::Util::validate_and_create_csrf_token($self, $c, 'db');

}



=head1 AUTHOR

Trond Haugen

=head1 LICENSE

This library is free software, you can redistribute it and/or modify
it under the same terms as Perl itself.

=cut

1;
