package NORID::EppClient::Controller::Host;

use strict;
use warnings;
use Storable qw(dclone);
use NORID::EppClient::Model::NORID::DRI;
use NORID::EppClient::Controller::Util qw /RC_OK RC_ERROR/;
use base qw(Catalyst::Controller::RequestToken);

# For Debug purposes:
use Data::Dumper;
$Data::Dumper::Indent=1;

my $TT_HOST = 'host.tt2';
my $TT_LOGIN = 'login.tt2';

=pod

=head1 NAME

EppClient - Host object operations

=head1 DESCRIPTION

The client supports creation and maintainance of host, contact and domain 
objects for .NO. It supports various transfer operations, as well as poll 
operation for the message queue.

It is intended as a simple and manual web client for low volume registrars.

=head1 SYNOPSIS

=head1 DESCRIPTION

package NORID::EppClient::Controller::Host - Catalyst Host Controller

=head1 METHODS

=head2 menu 

menu method.

=cut

# The host specific menu
sub menu : Local {
    my ( $self, $c ) = @_;

    $c->stash->{ph} = $c->session->{ph} if ($c->session->{ph} && !defined($c->stash->{ph}));

    # Recreate token when menu is pressed, this will cleanup any errors
    $c->session->{'h_token'} = $self->create_token($c);

    $c->stash->{template}  = $TT_HOST;
}

=head2 onemoreip

onemoreip method.

=cut

sub onemoreip : Private {
    my ( $self, $c ) = @_;

    push @{$c->stash->{ph}->{v46}}, "";
}

=head2 action

action method.

This method catches all the EPP host actions coming from the host form 
submission and dispatches the action to the real action method.

=cut

sub action : Local {
    my ( $self, $c) = @_;
    
    $c->session->{lastnamespace} = $c->namespace;

    $c->stash->{template} = $TT_HOST;

    #print STDERR "stash action: ", $c->req->param('h_token'), "\n";
    #print STDERR "session act : ", $c->session->{'__NORID::EppClient::Controller::Host_token'}, "\n";

    $c->forward('validate_and_create_csrf_token');

    unless ($c->forward('set_stash_from_req_param')) {
	push @{$c->stash->{opsummary}}, "Nothing to do";
	my ($valok, @verr) = validate_from_stash ($self, $c);
	unless ($valok) {
	    #$c->log->debug('Validate failed!');
	    push @{$c->stash->{opsummary}}, @verr;
	    $c->stash->{cmd_show_status} = 1;
	    $c->stash->{cmd_success} = $valok;
	}
	return RC_OK;
    }

    my $op  = $c->stash->{op};
    my $cmd = $c->stash->{cmd} || undef;
    my $nm  = $c->stash->{ph}->{name};

    if ($op) {

	if ($op eq '+') {
	    $c->forward('onemoreip');
	} elsif ($op eq 'reset') {
	    $c->forward('reset');
	} elsif ($cmd) {
	    ####
	    # All parameters are now stashed. Validate them locally from the stash
	    # before any DRI stuff is called.
	    my ($valok, @verr) = validate_from_stash ($self, $c);
	    unless ($valok) {
		#$c->log->debug('Validate failed!');
		# $s should contain an array with some validation error descriptions
		push @{$c->stash->{opsummary}}, @verr;
		$c->stash->{cmd_show_status} = 1;
		$c->stash->{cmd_success} = $valok;
		return RC_ERROR;
	    }
	    #$c->log->debug("Validated OK, valok: $valok !");
	    ###
	    # Syntax OK, do the DRI stuff

	    #$c->log->debug('EppClient: login before EPP operation needed!');

	    unless ($c->forward('/login/_login_is_ok')) {
		$c->log->info('EppClient: login before EPP operation failed!');
		# If either of above don't work out, send to the login page
		$c->stash->{template} = $TT_LOGIN;
		return RC_ERROR;
	    }

	    $c->log->info('EppClient: login before EPP operation SUCCEEDED!');

	    ######
	    # Do the EPP operation, send it to Net::DRI
	    $c->stash->{p} = $c->stash->{ph};
	    if ($c->stash->{facets} && $c->stash->{facets}->{vfyfacets} && $c->stash->{facets}->{eppf}) {
		$c->stash->{p}->{facets} = $c->stash->{facets}->{eppf};
	    }
	 
	    my ($resok, $dri, $rc, $rx) = NORID::EppClient::Model::NORID::DRI::do_epp_command($c, 'hash');

	    $c->log->info("do_epp_command: op: $op, name: $nm, res: $resok");

	    $c->stash->{cmd_show_status} = 1;
	    $c->stash->{cmd_success} = $resok;

	    if ($resok) {

		$c->stash->{cmd_pending} = 1 if ($rc->is_pending);

		# Success, map the various operation results to stash, summary and session 
		#$c->log->debug("EppClient: resok: $resok!");
		push @{$c->stash->{opsummary}},  "$op '$nm' succeeded.";

		if ($op eq 'info') {
		    $c->stash->{infoobj} = 
			&NORID::EppClient::Model::NORID::DRI::host_object_as_fmt($dri, 'hash');
		}
		NORID::EppClient::Controller::Util::set_common_results($c, $resok, $dri, $rc, $rx);

		$c->forward("set_".$op."_result");
	    } else {
		# some EPP error, like DRI exception with some error
		$c->log->info('EppClient: NOT resok, some EPP error!');

		#####
		# Collect results from the DRI-operation.
		push @{$c->stash->{opsummary}},  "$op '$nm' failed.";
		NORID::EppClient::Controller::Util::set_common_results($c, $resok, $dri, $rc, $rx);
	    }
	} else {
	    push @{$c->stash->{opsummary}}, "Nothing to do";
	    return RC_OK;
	}
    } else {

	$c->forward('menu');
    }

    # print STDERR "eash action: ", $c->session->{h_token}, "\n";
    # print STDERR "ession act : ", $c->session->{'__NORID::EppClient::Controller::Host_token'}, "\n";

}

=head2 get_info

Do an info on an existing to fetch the current values.

=cut

sub get_host_info : Private {
    my ( $self, $c ) = @_;

    #$c->log->debug("EppClient: get_info: op: " .  $c->stash->{op});

    $c->stash->{cmd} = 'info';

    my ($resok, $dri, $rc, $rx) = NORID::EppClient::Model::NORID::DRI::do_epp_command($c, 'hash');
    
    #$c->log->debug("EppClient: get_info: name: " .  $c->stash->{ph}->{name} . " resok: $resok");
	    
    if ($resok) {
	# Success, map the various results to stash, summary and session 
	# depending of the operation

	$c->stash->{infoobj} = 
	    &NORID::EppClient::Model::NORID::DRI::host_object_as_fmt($dri, 'hash');

	$c->forward("set_info_result");
    } else {
	# some EPP error, like already exists etc.
	#$c->log->debug('EppClient: NOT resok, some EPP error!');
	push @{$c->stash->{opsummary}},  "Info failed";
    }
}


=head2 set_check_result

Set result from check operation.

=cut

sub set_check_result : Private {
    my ( $self, $c ) = @_;

    my $dri = $c->stash->{dri};
    my $nm = $c->stash->{ph}->{name};

    push @{$c->stash->{opsummary}},  $dri->get_info('exist')?"host exists. ":"host does not exist. ";

    $c->forward('null_data');
#    $c->forward('set_session_from_stash');
    $c->stash->{ph}->{name} = $nm;
}

=head2 set_info_result

Set result and object data from info operation.

=cut

sub set_info_result : Private {
    my ( $self, $c ) = @_;

    my $dri = $c->stash->{dri};

    my $nm = $c->stash->{ph}->{name};

    my $s = NORID::EppClient::Model::NORID::DRI::host_object_as_fmt($dri, 'hash');
    
    $c->forward('set_stash_from_host_object', [ $s ]);

#    print STDERR "Host set_info_result, stash ph : ", Dumper $c->stash->{ph};

    $c->forward('set_session_from_stash');


}

=head2 set_create_result

Set result from create operation.

When a create is succesful, the object is created. 
A handle is created as well, and we do an info to fetch it.

=cut

sub set_create_result : Private {
    my ( $self, $c ) = @_;

    my $dri = $c->stash->{dri};
    my $nm = $c->stash->{ph}->{name};

    # Do a get info to fetch the created data and to get the handle.
    $c->forward('get_host_info');
}


=head2 set_update_result

Set result from the update operation.

=cut

sub set_update_result : Private {
    my ( $self, $c ) = @_;

    my $dri = $c->stash->{dri};
    my $rc = $c->stash->{rc};
    my $nm = $c->stash->{ph}->{name};

    unless ($rc && $rc->can('is_pending') && $rc->is_pending) {
	# The new name has been effectuated
	$c->stash->{ph}->{name} = $c->stash->{ph}->{nname} if ($c->stash->{ph}->{nname});
    }
    # Do a get info to fetch the updated data.
    $c->forward('get_host_info');
}

=head2 set_delete_result

Set result from the delete operation.

When a delete is done successfully, all form fields are cleared except the name.
When a delete is tried, but not succeeded, the current values are still valid, thus all are kept.

=cut

sub set_delete_result : Private {
    my ( $self, $c ) = @_;

    my $dri = $c->stash->{dri};
    my $nm = $c->stash->{ph}->{name};

    push @{$c->stash->{opsummary}},  "Delete '$nm': " . $dri->result_message();

    $c->forward('null_data');
#    $c->forward('set_session_from_stash'); # no objects exists 
    $c->stash->{ph}->{name} = $nm;
}

=head2 set_stash_from_req_param

Set stash from the request parameters.

=cut

sub set_stash_from_req_param: Private {
    my ( $self, $c) = @_;

    my $op = lc($c->req->param('hostop') || "") ||
	$c->req->param('increment') || 
	lc($c->req->param('reset') || "");

    $c->stash->{obj} = 'host';
    $c->stash->{op} = $op;
    $c->stash->{cmd} = $op;

    $c->session->{objdetails} = $c->req->param('objdetails');
    $c->session->{cmddetails} = $c->req->param('cmddetails');
    $c->session->{showstatus} = $c->req->param('showstatus');
    $c->session->{facetmenu}  = $c->req->param('facetmenu');

    # Store any facet values
    &NORID::EppClient::Controller::Util::set_stash_facets_from_req_param($c);

    my $oname = $c->session->{ph}->{'name'} || "";

    # hostname and name is the same field, but we want the browser
    # to remember the field-name and need a unique form-name.

    my $name = NORID::EppClient::Controller::Util::trim($c->req->param('hostname'));
    $c->stash->{ph}->{name} = lc($name);

    $c->stash->{ph}->{contact} = NORID::EppClient::Controller::Util::trim(uc($c->req->param('contact') || ""));

    if (my @ips = $c->req->param("v46")) {
	@ips = NORID::EppClient::Controller::Util::trim(@ips);
	@ips = NORID::EppClient::Controller::Util::wash_array(@ips);
  	push @{$c->stash->{ph}->{v46}}, sort @ips;
	foreach my $ip (sort @ips) {
	    # sort the in v4 and v6 also
	    if ($ip =~ m/\./) {
		push @{$c->stash->{ph}->{v4}}, $ip;
	    } else {
		push @{$c->stash->{ph}->{v6}}, $ip;
	    }
	}
    }
    # Sponsoringclientid is only interesting if set to another user.
    # 
    delete ($c->stash->{ph}->{sponsoringclientid});
    if (my $oid = NORID::EppClient::Controller::Util::trim($c->req->param('sponsoringclientid'))) {
	$c->stash->{ph}->{sponsoringclientid} = lc($oid) if ($oid ne $c->session->{clid});
    }

    $c->stash->{ph}->{roid} = $c->session->{ph}->{roid} || undef;

    if ($op eq 'update') {
	return $c->forward('detect_and_stash_host_object_changes');
    }
    #$c->log->debug("EppClient: action: stash_set_params: oname: $oname,  name: $name");
    #print STDERR "ph:",  Dumper $c->stash->{ph};
    1;
}

=head2 validate_from_stash

Validate the parameters not checked locally by the DRI lib.

=cut

sub validate_from_stash {
    my ( $self, $c) = @_;

    my $op   = $c->stash->{op};
    my $cmd  = $c->stash->{cmd};

    my $oname = $c->session->{ph}->{name} || "";
    my $name  = $c->stash->{ph}->{name};
    my $nname = $c->stash->{ph}->{nname} if ($c->stash->{ph}->{nname});

    my $valok = 1;
    my @e;


    unless ($name) {
	$valok = 0;
	push @e, "'$op' requires a host name";
    }
    #print STDERR "validate name: $name\n";
 
    # A valid host name is needed for all commands
    #$valok = 0 unless (NORID::EppClient::Controller::Util::is_a_valid_parameter($c, 'host', 'host name', $name));
    unless (NORID::EppClient::Controller::Util::is_a_valid_hostname($name)) {
	$valok = 0;
	push @e, "Illegal host name '$name'";
    }

    if ($nname) {
	#$valok = 0 unless (NORID::EppClient::Controller::Util::is_a_valid_parameter($c, 'host', 'new host name', $nname));
	unless (NORID::EppClient::Controller::Util::is_a_valid_hostname($nname)) {
	    $valok = 0;
	    push @e, "Illegal host name '$nname'";
	}
    }

    if (my $oid = $c->stash->{ph}->{sponsoringclientid}) {
	unless ($oid =~ m/^reg\d+$/mx) {
	    $valok = 0;
	    push @e, "Illegal Sponsor ID: $oid. Syntax must be a registrar id (like reg123)";
	}
    }

    my ($facok, @facerrs) = &NORID::EppClient::Controller::Util::is_valid_facets($c);
    unless ($facok) {
	$valok = 0;
	push @e, "Invalid facets", @facerrs;
    }


    # The host name is a sufficient parameter except for create and update which need 
    # more parameters and thus further validation.
    unless ($cmd eq 'create' || $cmd eq 'update') {
	return $valok, @e;
    }

    foreach my $ip (@{$c->stash->{ph}->{v4}}) {
	# Check that all the ip-addresses are valid v4
	#$valok = 0 unless (NORID::EppClient::Controller::Util::is_a_valid_parameter($c, 'host', 'IP-address', $ip));
	unless (Net::DRI::Util::is_ipv4($ip, 1)) {
	    $valok = 0;
	    push @e, "Illegal IPV4-address '$ip'";	
	}
    }

    foreach my $ip (@{$c->stash->{ph}->{v6}}) {
	# Check that all the ip-addresses are valid v6
	#$valok = 0 unless (NORID::EppClient::Controller::Util::is_a_valid_parameter($c, 'host', 'IP-address', $ip));
	unless (Net::DRI::Util::is_ipv6($ip, 1)) {
	    $valok = 0;
	    push @e, "Illegal IPV6-address '$ip'";	
	}
    }
    if (my $ct = $c->stash->{ph}->{contact}) {
	# The contact must be a role
	unless (NORID::EppClient::Controller::Util::isRoleID($ct)) {
	    $valok = 0;
	    push @e, "Illegal Contact ID '$ct', must be a role";	
	}
    }

    #$c->log->debug("Validate returns: valok: $valok, errs: " . join(", ", @e));

    #print STDERR "ph: ", Dumper $c->stash->{ph}, "\n";

    return $valok, @e;
}

sub detect_and_stash_host_object_changes: Private {
    my ( $self, $c) = @_;

    my $oname = $c->session->{ph}->{name} || "";
    my $name = $c->stash->{ph}->{name};

#    print STDERR "detect changes: name: $name, oname: $oname\n";
#    print STDERR "detect changes: o: ", Dumper $c->session->{ph}, "\n";
#    print STDERR "detect changes: n: ", Dumper $c->stash->{ph}, "\n";

    my $update = 0;
    # check which parameters are changed

    if ($oname && $name && $name ne $oname) {
	$c->stash->{ph}->{nname} = $name;
	$c->stash->{ph}->{name}  = $oname;
	++$update;
	$c->log->debug("EppClient: action: UPDATE: new name: $oname -> $name");
    }

    my $oct = $c->session->{ph}->{'contact'} || "";
    my $ct = $c->stash->{ph}->{'contact'} || "";
    
    if (($oct || $ct) && $oct ne $ct) {
	# Need to update the contact. Either to another value or to delete it
	$c->log->debug("EppClient: action: UPDATE CONTACT: $update, oct: '$oct', ct: '$ct'");
	$c->stash->{ph}->{ncontact}->{add} = $ct if $ct;
	$c->stash->{ph}->{ncontact}->{del} = $oct if $oct;
	++$update;
    }
    # Ip-address changes need to be done as add/delete arrays.
    foreach my $v ('v4', 'v6') {
	my ($ar, $dr) = NORID::EppClient::Controller::Util::make_add_del_list(
								       $c->stash->{ph}->{$v}, 
								       $c->session->{ph}->{$v});
	if ($ar && @$ar) {
	    $c->stash->{ph}->{ipset}->{add}->{$v} = $ar;
	    ++$update;
	}
	if ($dr && @$dr) {
	    $c->stash->{ph}->{ipset}->{del}->{$v} = $dr;
	    ++$update;
	}
	#$c->log->debug("EppClient: action: update: stash_host: update: $update, oname: $oname, name:$name, oct: $oct, ct: $ct, add: @$ar, del: @$dr"); 
    }
    return $update;
}

=head2 set_session_from_stash

Put the data from the stash into the session data so we can present 
the previous values in the next page.

=cut

sub set_session_from_stash: Private {
    my ( $self, $c ) = @_;

    $c->session->{ph} = dclone($c->stash->{ph}) if ($c->stash->{ph});
}

=head2 set_stash_from_host_object

We have a DRI host object.
Put the interesting data from it into the stash.

=cut

sub set_stash_from_host_object: Private {
    my ( $self, $c, $s) = @_;

    $c->forward('null_data');

    $c->stash->{ph}->{name} = $s->{name} if $s->{name};
    push @{$c->stash->{ph}->{v4}}, sort @{$s->{v4}} if $s->{v4};
    push @{$c->stash->{ph}->{v6}}, sort @{$s->{v6}} if $s->{v6};
    push @{$c->stash->{ph}->{v46}}, sort @{$s->{v4}} if $s->{v4};
    push @{$c->stash->{ph}->{v46}}, sort @{$s->{v6}} if $s->{v6};

    $c->stash->{ph}->{contact} = $s->{contact} if $s->{contact};

    # Sponsoringclientid is only interesting if clID is set to another user.
    # 
    delete ($c->stash->{ph}->{sponsoringclientid});
    if (my $oid = $s->{clID}) {
	$c->stash->{ph}->{sponsoringclientid} = $oid if ($oid ne $c->session->{clid});
    }
    foreach my $e ('status','statuslist', 'roid', 'clID', 'crID', 'upID', 'crDate', 'upDate') {
	$c->stash->{ph}->{$e} = $s->{$e} || $c->session->{ph}->{$e} || "";
    }
}

=head2 reset

Reset form, zap session and stash, and display empty values.

=cut

sub reset : Private {
    my ( $self, $c ) = @_;

    $c->forward('null_data');
    delete ($c->session->{objdetails});
    delete ($c->session->{cmddetails});
    delete ($c->session->{showstatus});
    delete ($c->session->{facetmenu});
    delete $c->stash->{facets};
    delete $c->session->{facets};
}

sub null_data : Private {
    my ( $self, $c ) = @_;

    delete $c->stash->{ph};
    delete $c->session->{ph};

}

=head2 index

Show the host page.

=cut

sub index : Private {
    my ( $self, $c ) = @_;

    $c->forward('menu');
}


=head2 validate_and_create_csrf_token

CSRF validation.

=cut

sub validate_and_create_csrf_token: Private {
    my ( $self, $c) = @_;

    &NORID::EppClient::Controller::Util::validate_and_create_csrf_token($self, $c, 'h');

}



=head1 AUTHOR

Trond Haugen

=head1 LICENSE

This library is free software, you can redistribute it and/or modify
it under the same terms as Perl itself.

=cut

1;
