#!/usr/bin/perl
use strict;
use warnings;
use Test::More "no_plan";
use Test::WWW::Mechanize::Catalyst 'NORID::EppClient';
use HTML::Entities;
use URI::Escape;
use Encode;
use NORID::EppClient::Controller::Util;
use NORID::EppClient::InitTests;

use Data::Dumper;
$Data::Dumper::Indent=1;

my $tc = NORID::EppClient::InitTests::get_test_config;
my $t = $tc->{t};

my $rega = $t->{rega};
my $regb = $t->{regb};
my $pwa  = $t->{pwa};
my $pwb  = $t->{pwb};

my $cf = NORID::EppClient->config();

my $today  = NORID::EppClient::Controller::Util::timestamp_as_date(time);
my $future = NORID::EppClient::Controller::Util::timestamp_as_date(time + (60*60*48));
my $past   = NORID::EppClient::Controller::Util::timestamp_as_date(time - (60*60*48));

# string to make names unique, todays date + process id
my $sfx = "$today-$$";

my $ua1 = Test::WWW::Mechanize::Catalyst->new;
my $ua2 = Test::WWW::Mechanize::Catalyst->new;

$ua1->get_ok("http://localhost/", 'get localhost');

is($ua1->ct, "text/html", 'text/html');

$ua1->title_is("Norid EPP Client", "On the EPP Client's root page");

$ua1->content_contains("You need to log in to use this application", "Yup, on the login page");

### login with correct pw
# login with correct pw
$ua1->get_ok("http://localhost/login?command=Login&clid=$rega&pw=$pwa", 'Login OK');
$ua2->get_ok("http://localhost/login?command=Login&clid=$regb&pw=$pwb", 'Login OK');

is($ua1->ct, "text/html", 'text/html');
$ua1->title_is("Norid EPP Client", "Inside the EPP Client");
$ua1->content_contains("Main menu", "Main menu found");
$ua1->content_contains("Please select one of the above tabs for a specific operation.", "Main menu found, I'm sure!");


############### Contact tab ###############

#####
# ILLEGAL PARS:
# Covers also the Update function
#

my $contactname="Carl Concreate Barks-$sfx";
my $type='person';
my $cc='NO';
my $pc='NO-7005';
my $city='Trondheim';
my $email='carl.barks@data.no';
my $street1 = "Barksroad-1 44A";
my $street2 = "Barksroad-2 44A";
my $street3 = "Barksroad-3 44A";
my $street4 = "Barksroad-4 44B";

### Missing some mandatory pars

# Contact ID/SRID, shall NOT be specified on a create
$ua1->get_ok("http://localhost/contact/action?contactop=Create&srid=OX12O&contactname=$contactname&type=$type&cc=$cc&pc=$pc&city=$city&email=$email", "Create with srid shall fail");
$ua1->content_contains("create: Contact ID must be empty to verify the create", "[023-J-3]  Create with srid failed");

# no name
$ua1->get_ok("http://localhost/contact/action?contactop=Create&contactname=&type=$type&cc=$cc&pc=$pc&city=$city&email=$email", "Create contact without name shall fail");
$ua1->content_contains("create: mandatory parameter 'name' not specified", "[023-J-3] Create without name failed");

# name has illegal character - like a '_' char 
my $contactname_ill="Carl_Barks-$sfx";
$ua1->get_ok("http://localhost/contact/action?contactop=Create&contactname=$contactname_ill&type=$type&cc=$cc&pc=$pc&city=$city&email=$email", "Create contact with illegal character shall fail");
$ua1->content_contains("The contact name [$contactname_ill] is invalid.", "[023-J-3] Create with illegal character in name failed");


# name is too long for EPP, max. length is 255
# this case will be stopped by Net::DRI
my $cnlong = join " ", ( 'aa' .. 'gg');
my $len = length($cnlong);
$ua1->get_ok("http://localhost/contact/action?contactop=Create&contactname=$cnlong&type=$type&cc=$cc&pc=$pc&city=$city&email=$email", "Create contact with too long name ($len characters) shall fail");
#print $ua1->content;
$ua1->content_contains("create '$cnlong' failed", "[023-J-3] Create with too long name ($len characters) failed 1");
$ua1->content_contains("Invalid parameters: Invalid contact information: name", "[023-J-3] Create with too long name ($len characters) failed 2");



# type
$ua1->get_ok("http://localhost/contact/action?contactop=Create&contactname=$contactname&type=&cc=$cc&pc=$pc&city=$city&email=$email", "Create contact without type shall fail");
$ua1->content_contains("create: mandatory parameter 'type' not specified", "[023-J-3] Create without type failed");

# identitytype, for organization or person 
my $identitytype='localIdentity';
# AUTO generate the number
my $orgno = 'N.ORG.AUTO';
$type='organization';
my $identityvalue = $orgno;
$ua1->get_ok("http://localhost/contact/action?contactop=Create&contactname=$contactname&type=$type&cc=$cc&pc=$pc&city=$city&email=$email&identitytype=&identityvalue=$identityvalue", "Create without identitytype shall fail");
$ua1->content_contains("create: mandatory parameter 'identitytype' not specified", "[023-J-3] Create  without identitytype failed");


# identityvalue, for organization or person 
$ua1->get_ok("http://localhost/contact/action?contactop=Create&contactname=$contactname&type=$type&cc=$cc&pc=$pc&city=$city&email=$email&identitytype=&$identitytype&identityvalue=", "Create without identityvalue shall fail");
$ua1->content_contains("create: mandatory parameter 'identitytype' not specified", "[023-J-3] Create  without identityvalue failed");

# with 4 street lines, one too many
$ua1->get_ok("http://localhost/contact/action?contactop=Create&contactname=$contactname&type=$type&cc=$cc&pc=$pc&city=$city&email=$email&identitytype=$identitytype&identityvalue=$identityvalue&street=$street1&street=$street2&street=$street3&street=$street4", "Create with 4 address lines shall fail");
$ua1->content_contains("Too many address lines, EPP has a maximum of three", "[023-J-3] Create with 4 address lines failed");


# cc 
$ua1->get_ok("http://localhost/contact/action?contactop=Create&contactname=$contactname&type=$type&cc=&pc=$pc&city=$city&email=$email&identitytype=&$identitytype&identityvalue=", "Create without cc shall fail");
$ua1->content_contains("create: mandatory parameter 'cc' not specified", "[023-J-3] Create  without cc failed");

# pc
$ua1->get_ok("http://localhost/contact/action?contactop=Create&contactname=$contactname&type=$type&cc$cc=&pc=&city=$city&email=$email&identitytype=&$identitytype&identityvalue=", "Create without pc shall fail");
$ua1->content_contains("create: mandatory parameter 'pc' not specified", "[023-J-3] Create  without pc failed");

# Illegal pc
$pc='NO-7000';
$ua1->get_ok("http://localhost/contact/action?contactop=Create&contactname=$contactname&type=$type&cc=$cc&pc=$pc&city=$city&email=$email&identitytype=$identitytype&identityvalue=$identityvalue", "Create with illegal pc shall fail");
$ua1->content_contains("[$pc] is not the proper zip code for [$city].", "[023-J-3] Create with illegal pc failed");

#print $ua1->content;

# city
$ua1->get_ok("http://localhost/contact/action?contactop=Create&contactname=$contactname&type=$type&cc=$cc&pc=$pc&city=&email=$email&identitytype=$identitytype&identityvalue=$identityvalue", "Create without city shall fail");
$ua1->content_contains("create: mandatory parameter 'city' not specified", "[023-J-3] Create  without city failed");

# voice illegal
my $voice = '+47123456';
$ua1->get_ok("http://localhost/contact/action?contactop=Create&contactname=$contactname&type=$type&cc=$cc=$cc&pc=$pc&city=$city&email=$email&identitytype=$identitytype&identityvalue=$identityvalue&voice=".uri_escape($voice), "Create with illegal voice shall fail");
$ua1->content_contains("Illegal telephone number for 'voice': $voice (syntax: +47.73551000)", "[023-J-3] Create  with illegal voice failed");

# mobile illegal
my $mobilephone = '+47123456';
$ua1->get_ok("http://localhost/contact/action?contactop=Create&contactname=$contactname&type=$type&cc=$cc=$cc&pc=$pc&city=$city&email=$email&identitytype=$identitytype&identityvalue=$identityvalue&mobilephone=".uri_escape($mobilephone), "Create with illegal mobilephone shall fail");
$ua1->content_contains("Illegal telephone number for 'mobilephone': $mobilephone (syntax: +47.73551000)", "[023-J-3] Create  with illegal mobilephone failed");


# primary email illegal
$email='illegal@ille_gal.no';
$ua1->get_ok("http://localhost/contact/action?contactop=Create&contactname=$contactname&type=$type&cc=$cc=$cc&pc=$pc&city=$city&email=$email&identitytype=$identitytype&identityvalue=$identityvalue", "Create with illegal primary email shall fail");
$ua1->content_contains("Illegal primary email address '$email'", "[023-J-3] Create  with illegal primary email failed");


# secondary email illegal
$ua1->get_ok("http://localhost/contact/action?contactop=Create&contactname=$contactname&type=$type&cc=$cc=$cc&pc=$pc&city=$city&xemail=$email&identitytype=$identitytype&identityvalue=$identityvalue", "Create with illegal secondary email shall fail");
$ua1->content_contains("Illegal secondary email address '$email'", "[023-J-3] Create  with illegal secondary email failed");


# related org syntax error
$type='person';
my $relorg='ILL12';
$ua1->get_ok("http://localhost/contact/action?contactop=Create&contactname=$contactname&type=&cc=$cc&pc=$pc&city=$city&organization=$relorg", "Create person contact with illegal related org. ID shall fail");
$ua1->content_contains("'create': Invalid Related org. ID: $relorg", "[023-J-3] Create person contact with illegal related org. ID failed");

# role contact syntax error
$type='role';
$relorg='ILL12O';
my $rolecontact='ILL12';
$ua1->get_ok("http://localhost/contact/action?contactop=Create&contactname=$contactname&type=&cc=$cc&pc=$pc&city=$city&organization=$relorg&rolecontact=$rolecontact", "Create role contact with illegal role contact ID shall fail");
$ua1->content_contains("'create': Invalid Role contact ID: $rolecontact", "[023-J-3] Create role contact with illegal role contact ID failed");


# Unsupported check. Do a contact check, the registry will should because of not implemented operation
my $srid_p = 'SOM12P';
$ua1->get_ok("http://localhost/contact/action?contactop=Check&srid=$srid_p", "Check $srid_p should be rejected");
$ua1->content_contains("Protocol error. Command [check] for object [urn:ietf:params:xml:ns:contact-1.0] is not supported", "[023-J-3] Check $srid_p was rejected");
#print $ua1->content;


#######
# LEGAL parameters
#

$pc='NO-7005';
$email='carl.barks@data.no';

# Create: A LocalIdentifier AUTO organization contact
# Also use special chars &, 
$identitytype='localIdentity';

# AUTO generate the number, shall be rejected
$orgno = 'N.ORG.AUTO';
$type='organization';
$identityvalue = $orgno;

my $contactname_idn="Carl Concreate Barks   osv &amp; Co $sfx";
my $contactname_idnesc="Carl Concreate Barks   osv %26 Co $sfx";
my $contactname_idni=uri_escape($contactname_idn);
my $ci = encode("UTF-8", decode("ISO8859-1", $contactname_idn));
my $cci='AT';
my $pci='AT-7005';
my $city_idni='Trondheim   %26 Co';
my $emaili='carl.barks@data.no';
my $street1i = "Barksroad-1 44A  %26 Co";
my$street2i = "Barksroad-2 44A  %26 Co";

$ua1->get_ok(encode("UTF-8", decode("ISO8859-1","http://localhost/contact/action?contactop=Create&contactname=$contactname_idnesc&type=$type&cc=$cci&pc=$pci&city=$city_idni&street=$street1i&street=$street2i&email=$emaili&identitytype=$identitytype&identityvalue=$identityvalue")), encode("UTF-8", decode("ISO8859-1","Create organization with LocalID number, IDN  chars and '&'")));

$ua1->content_contains("create '$contactname_idn' failed", "[023-J-3] Create Local IDN ID '$ci' failed, N.ORG.AUTO not permitted");
$ua1->content_contains("The value [localIdentity:N.ORG.AUTO] of type [Registry::NORID::Person::Identity::LocalId] is not well-formed", "[023-J-3] The local ID value 'N.ORG.AUTO' is not permitted");


###
# Create a real organization and first doing a local Brreg check.
# Create an organization using a real Brreg name and number with  in the name
#    Using: 913681800	KVRNER EIENDOM KLFTA AS
# Now, also use a wrong name to trig the Brreg correction function which should insert the 
# correct name looked up.
#
$type='organization';
$identitytype='organizationNumber';

$orgno= "913681800";
my $ctname_ok  = "KVRNER EIENDOM KLFTA AS";
my $ctname_err = "KVRNER xxxxxxx KLOFTA AS";
my $contactname_1 = decode("ISO8859-1", $ctname_err);
my $econtactname_1=encode_entities($contactname_1);

$identityvalue = $orgno;

$ua1->get_ok("http://localhost/contact/action?contactop=Create&contactname=$contactname_1&type=$type&cc=$cc&pc=$pc&city=$city&email=$email&identitytype=$identitytype&identityvalue=$identityvalue&brchk=on", "Create a real Brreg organization with IDN characters ...");

####
# The create may be ok or fail due to already existing object in the registry.
# We're not interested in that now, on to verify that the local brreg check did work.
# Check the response that should state that the org number was foung, that the name has been corrected.
my $result  = "Brnnysund result:";
my $dresult = decode("ISO8859-1", $result);
my $eresult = encode_entities($dresult);
$ua1->content_contains($result, "Content contains $eresult - OK");
$ua1->content_contains('- The organization number was found.', "Organization number was found in Brreg - OK");
$ua1->content_contains('- Mismatch on organization name', "Mismatch on organization name was detected - OK");
$ua1->content_contains('- Name corrected to ', "Name was corrected to correct name in Brreg - OK");

###
# Now a create with fake organization data from test config. No brreg check on these because
# they are synthetic and do not exist in Brreg.
#
# Either it works or we get a reject due to already registered,
# but in any case the proper create syntax is verified.

$orgno = $t->{orgnoa1};
$contactname_1 = $t->{orgnamea1};
$econtactname_1=encode_entities($contactname_1);
my $lcname_1  = $t->{legalcname1};
$identityvalue = $orgno;

#
# 2017 new and old datamodel.

my $srid_o;
my $regexp = qr/(Contact with identity \[$identitytype:$identityvalue\] already exists as \[(\w+)\])/;

# Use org name and legalc name in an org
$ua1->get_ok("http://localhost/contact/action?contactop=Create&contactorg=$contactname_1&contactname=$lcname_1&type=$type&cc=$cc&pc=$pc&city=$city&email=$email&identitytype=$identitytype&identityvalue=$identityvalue", "Create Brreg organization new model...");
    
$ua1->content_contains("create '$lcname_1' succeeded", "[023-J-3] Created fake Brreg '$lcname_1 / contactname_1'");
$srid_o = $ua1->form_number('1')->value('srid');
    

# Create: An OK person contact with ok parameters and only mandatory parameters
$type='person';
my $PID      = $t->{pidno1};
my $pid_name = $t->{pidname1};
my $identitytype_p='anonymousPersonIdentifier';

$ua1->get_ok("http://localhost/contact/action?contactop=Create&contactname=$pid_name&type=$type&identitytype=$identitytype_p&identityvalue=$PID&cc=$cc&pc=$pc&city=$city&email=$email", "Create person contact with ok and only mandatory parameters");
my $srid_pm = $ua1->form_number('1')->value('srid');
$ua1->content_contains("create '$pid_name' succeeded", "[023-J-3] Create person contact with ok and only mandatory parameters, id: '$srid_pm'");


# Create: A person contact with ok parameters and all person parameters in use and related org as the above
$voice       = uri_escape('+47.12345677');
$mobilephone = uri_escape('+47.12345679');

$ua1->get_ok("http://localhost/contact/action?contactop=Create&contactname=$pid_name&type=$type&identitytype=$identitytype_p&identityvalue=$PID&cc=$cc&pc=$pc&city=$city&street=$street1&street=$street2&street=$street3&email=$email&voice=$voice&mobilephone=$mobilephone", "Create person contact with all possible parameters filled in");
$srid_p = $ua1->form_number('1')->value('srid');
$ua1->content_contains("create '$pid_name' succeeded", "[023-J-3] Created person contact '$pid_name' as id '$srid_p' with all possible parameters filled in");

###
# Create: A private person contact 2.
$PID      = $t->{pidno2};
$pid_name = $t->{pidname2};

$ua1->get_ok("http://localhost/contact/action?contactop=Create&contactname=$pid_name&type=$type&identitytype=$identitytype_p&identityvalue=$PID&cc=$cc&pc=$pc&city=$city&email=$email", "Create anonymeous private person under $rega");
my $srid_priv = $ua1->form_number('1')->value('srid');
$ua1->content_contains("create '$pid_name' succeeded", "[023-J-?] Created private '$pid_name' ($srid_priv) of type $identitytype_p $rega OK");


# Create a $regb person for other sponsorship tests
$ua2->get_ok("http://localhost/contact/action?contactop=Create&contactname=$pid_name&type=$type&identitytype=$identitytype_p&identityvalue=$PID&cc=$cc&pc=$pc&city=$city&street=$street1&street=$street2&email=$email", "Create person contact under $regb for sponsorship test");
my $srid_sp = $ua2->form_number('1')->value('srid');
$ua2->content_contains("create '$pid_name' succeeded", "[023-J-3] Created person contact '$pid_name' as id '$srid_sp' under $regb for sponsorship test");


###
# Operations
my @editops = ('Info', 'Update', 'Delete');

# Operations without contact ID
foreach my $op (@editops) {
    my $lop = lc($op);
    $ua1->get_ok("http://localhost/contact/action?contactop=$op&srid=", "$op contact without ID/Handle ...");
    $ua1->content_contains("'$lop' requires a handle", "[023-I-3] $op contact without ID/Handle failed");
}

# Operations with illegal contact ID
my $illid="ILL12";
foreach my $op (@editops) {
    my $lop = lc($op);
    $ua1->get_ok("http://localhost/contact/action?contactop=$op&srid=$illid", "$op contact with illegal ID/Handle ...");
    $ua1->content_contains("'$lop': Invalid Contact ID: $illid", "[023-I-3] $op contact with illegal ID/Handle failed");
}

# Operations on valid but non-existing contact ID
$illid="ILLX12P";
foreach my $op (@editops) {
    my $lop = lc($op);
    next if ($op eq 'Update'); # because it requires a lot of other parameters, and is assume covered
    $ua1->get_ok("http://localhost/contact/action?contactop=$op&srid=$illid", "$op contact with valid but non-existing ID/Handle ...");
    $ua1->content_contains("$lop '$illid' failed", "[023-I-3] $op contact with non-existing ID/Handle failed");
    $ua1->content_contains(" Contact with handle [$illid] not found", "[023-I-3] $op contact with non-existing ID/Handle failed - not found");
 }

# Look up the existing contact
$ua1->get_ok("http://localhost/contact/action?contactop=Info&srid=$srid_p", "Lookup contact...");
$ua1->content_contains("info '$srid_p' succeeded.", "[023-I-3] Contact lookup of $srid_p");

$type='role';
my $rname = "Carl Concreate Roles Ltd $sfx";
my $srid_r;

# Create: A role contact with ok parameters
$ua1->get_ok("http://localhost/contact/action?contactop=Create&contactname=$rname&type=$type&cc=$cc&pc=$pc&city=$city&email=$email", "Create role contact");
$srid_r = $ua1->form_number('1')->value('srid');
$ua1->content_contains("create '$rname' succeeded", "[023-J-3] Created role '$rname' ($srid_r)");


# Info: Look up the existing contact
$ua1->get_ok("http://localhost/contact/action?contactop=Info&srid=$srid_r&objdetails=on", "Lookup contact...");
$ua1->content_contains("info '$srid_r' succeeded.", "[023-I-3] Contact lookup of $srid_r");

# verify proper sponsorship
my $c = $ua1->content(format => 'text');
$regexp = qr/(Client ID:)(reg\d+)/;
my $regid;
if ($c =~ $regexp) {
    $regid = $2;
}
ok ($regid && $regid eq $rega, "[023-B-1] Correct contact object sponsorship verified");



# Info: Look up the existing private contact
$ua1->get_ok("http://localhost/contact/action?contactop=Info&srid=$srid_priv&objdetails=on", "Lookup contact...");
$ua1->content_contains("info '$srid_priv' succeeded.", "[023-I-?] Contact lookup of private contact $srid_priv");

# verify proper sponsorship
$c = $ua1->content(format => 'text');
$regexp = qr/(Client ID:)(reg\d+)/;
$regid=undef;
if ($c =~ $regexp) {
    $regid = $2;
}
ok ($regid && $regid eq $rega, "[023-B-1] Correct contact object sponsorship verified");



##
# Info: Look up the existing contact under $regb
$ua1->get_ok("http://localhost/contact/action?contactop=Info&srid=$srid_sp&objdetails=on", "Lookup contact...");
$ua1->content_contains("info '$srid_sp' succeeded.", "[023-I-3] Contact lookup of $srid_sp under $regb");

# verify proper sponsorship
$c = $ua1->content(format => 'text');
$regexp = qr/(Client ID:)(reg\d+)/;
$regid=undef;
if ($c =~ $regexp) {
    $regid = $2;
}
ok ($regid && $regid eq $regb, "[023-B-1] Correct contact object sponsorship under $regb verified");


##

# Info without ID
$ua1->get_ok("http://localhost/contact/action?contactop=Info&srid=", "Lookup empty contact ID");
$ua1->content_contains("'info' requires a handle", "[023-I-3] Contact Info om empty contact ID failed");

# Look up a non-existing contact
my $nx = "NONEXIST12P";
$ua1->get_ok("http://localhost/contact/action?contactop=Info&srid=$nx", "Lookup non-existing contact...");
$ua1->content_contains("info '$nx' failed.", "[023-I-3] Contact lookup of non-existing $nx failed");



### Delete the Contacts created above
foreach my $c ($srid_pm, $srid_r, $srid_p, $srid_priv) {
    $ua1->get_ok("http://localhost/contact/action?contactop=Delete&srid=$c", "Delete '$c'...");
    $ua1->content_contains("delete '$c' succeeded", "[023-J-13] Delete '$c' contact");
}
$ua2->get_ok("http://localhost/contact/action?contactop=Delete&srid=$srid_sp", "Delete '$srid_sp'...");
$ua2->content_contains("delete '$srid_sp' succeeded", "[023-J-13] Delete '$srid_sp' contact");


### Delete the Org. contact from above
$ua1->get_ok("http://localhost/contact/action?contactop=Delete&srid=$srid_o", "Delete '$srid_o'...");
$regexp = qr/(delete '$srid_o' failed.)/;
if ($ua1->content =~ $regexp) {
    $ua1->content_contains("Object [$srid_o] is in use and cannot be deleted.", "[023-J-13] Delete '$srid_o' contact not done, in use");
} else {
    $ua1->content_contains("delete '$srid_o' succeeded", "[023-J-13] Delete '$srid_o' contact done");
}


